(*-----------#pc-version-----------------------------------------------------*)
(*
implementation
uses rsutil,io,util,sys;
*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "calctet.h"
#include "calctet1.h"
#include "util.h"
#include "rsutil.h"
#include "sys.h"
#include "io.h"
#include "mon.h"
#include "part.h"
(*-----------#unix-version-----------------------------------------------------*)

(* 1.  Put in partialimp which doesn't calculate signs.  The version that does 
calculate signs is incorrect because for ij.k, instead of accepting only 
i(ik) intersect j(jk) it also accepts k(ik) intersect j(jk) or i(ik) 
intersects k(jk) and j(jk) intersect k(ik.)  In followbranch don't score 
model if it has submodel with same score. 
   2.  June 7,1989.  Changed restorecorr and restoreimp so check that 
changedcorr and changed are in range. Change insertmodel so puts 200 models
in list instead of depending on eqsnumber.  
   3.  August 23, 1989.  Made getprob, plus external 
   4.  Jan. 10, 1990.  Fixed bug in tetimp. 
   5.  Feb. 10, 1990.  Changed printout of tetrads. 
   6.  March 24, 1990.  changed calctrek to calculate differences in
       implied implications *)

function plus;
(* this function performs the '+' operation upon two paths and 
   returns the result.  for example, if path1 contains the paths ab and bc
   (representing ab + bc), and path2 contains the paths ef and gh 
   (representing ef + gh), then the function returns the list
   with member ab, cd, ef, and gh (representing ab + cd + ef + gh).  it is a
   destructive function. *)

var curpos: pathsptr;

begin (* plus *)
   if aaabort or ((path1 = nil) and (path2 = nil)) then plus := nil
   else if (path1 = nil) and (path2 <> nil) then
	 plus := path2
   else if (path1 <> nil) and (path2 = nil) then
	 plus := path1
      else
	 begin
	    curpos := path1;
	    while curpos^.next <> nil do curpos := curpos^.next;
	    curpos^.next := path2;
	    plus := path1;
	 end; (* else *)
end; (* plus *)

function isin(paths1,paths2: pathsptr; depth: integer): pathsptr;
(* this function determines whether the list of edges of paths2 occurs
   anywhere in the list headed by paths1 *)
var stop: boolean;
begin
   isin := nil;
   stop := false;
   while (paths1 <> nil) and (not stop) do
    if (paths1^.level <> 0) and (paths1^.level < depth) then stop := true
    else
      begin
         stop := paths1^.vertices = paths2^.vertices;
	 if (not stop) then paths1 := paths1^.next
         else isin := paths1;
      end;
end;

function maketrek(path: pathsptr; i,j,level: integer): treksptr;
label 999;
var temp,curpos: treksptr;
begin
 if aaabort or (path = nil) then maketrek := nil
 else if (path^.level < level) then maketrek := nil
 else
  begin
   new(temp);
   curpos := temp;
   while (path <> nil) do
     begin
       if (path^.level < level) and (path^.level <> 0) then goto 999;
       curpos^.level := 0;
       curpos^.next := nil;
       curpos^.source := i;
       if i <= j then
          begin
	     curpos^.trek2 := path;
	     curpos^.trek1 := nil;
	  end
       else
          begin
	     curpos^.trek2 := nil;
	     curpos^.trek1 := path;
	  end;
       path := path^.next;
       if (path <> nil) then
        if (path^.level >= level) or (path^.level = 0) then
          begin
	     new(curpos^.next);
             curpos := curpos^.next;
	  end;
      end;
    999: maketrek := temp;
   end;
end;


function plustrek(trek1,trek2: treksptr): treksptr;
(* this function implements the '+' operation with the restriction that
   each path can occur in the list at most once.  *)

var theend: treksptr;
(* theend represents the end of the list being constructed *)

begin (* plustrek *)
   if aaabort or ((trek1 = nil) and (trek2 = nil)) then plustrek := nil
   else if (trek1 = nil) and (trek2 <> nil) then plustrek := trek2
   else if (trek1 <> nil) and (trek2 = nil) then plustrek := trek1
   else
	 begin
	    theend := trek1;
	    while theend^.next <> nil do theend := theend^.next;
	    theend^.next := trek2;
	    plustrek := trek1;
	 end; (* else *)
end; (* plustrek *)

function nointersect(path1,path2: pathsptr; vert: vertextype): boolean;
begin 
 if (path1 = nil) or (path2 = nil) then nointersect := true
 else nointersect := (((path1^.vertices - [vert]) *  
                  (path2^.vertices - [vert])) = []);
end; 

function formtrek(path1,path2: pathsptr; source,sink1,sink2,depth: integer):
                  treksptr;
var temp: treksptr;
(* should check that not both path1 and path2 are nil and nocycle *)
begin
     new(temp);
     temp^.level := depth;
     if sink1 < sink2 then 
        begin
	   temp^.trek1 := path1;
           temp^.trek2 := path2;
        end
     else
        begin
           temp^.trek1 := path2;
           temp^.trek2 := path1;
        end;
     temp^.next := nil;
     temp^.source := source;
     formtrek := temp;
end;

function timestrek(paths1,paths2: pathsptr; level,source,
                   sink1,sink2: integer): treksptr;
(* this function takes the product of two paths.  it allows only one
   occurence of a given path in the final list.  for example, 
   (ab + cb) * (cd + ad) would return (abcd + abad + cbcd).  it does not
   allow any circular paths to be formed. *)

label 999;
var curpos1,curpos2,pathsecond: pathsptr; new1,switched: boolean;
    temp,curposnew,newhead: treksptr; 

begin (* timestrek *)
   newhead := nil;
   if aaabort or (paths1 = nil) or (paths2 = nil) then timestrek := nil
   else if (paths1^.level < level) and (paths2^.level < level) then
       timestrek := nil
   else 
(* make curpos1 one with highest level so know when to stop: procedure
   assumes that paths1 is to lower index and paths2 to higher: trek1 is
   always path to lowre index so switched records whether curpos1 is to
   lower terminus or higher terminus *)
   
        begin
            if paths1^.level >= paths2^.level then
               begin
                  curpos1 := paths1;
                  pathsecond := paths2;
                  switched := false;
               end
            else 
               begin
                  curpos1 := paths2;
                  pathsecond := paths1;
                  switched := true;
               end;
	    curposnew := nil;
	    new1 := true;
	    while curpos1 <> nil do
	       begin
	          if (curpos1^.level < level) and (curpos1^.level <> 0) then
	              new1 := false;
		  curpos2 := pathsecond;
		  while curpos2 <> nil do
		     begin
		        if (not new1) and (curpos2^.level < level) and
			   (curpos2^.level <> 0) then goto 999
			else if nointersect(curpos1,curpos2,source) then
		      (* if no cycle *)
			      begin
			       if not switched then
                                  temp := formtrek(curpos1,curpos2,source,
                                                sink1,sink2,level)
                               else temp := formtrek(curpos2,curpos1,source,
                                                sink1,sink2,level);
			       if curposnew = nil then
				    begin
				     curposnew := temp;
				     newhead := curposnew;
				    end
				else
				    begin
				     curposnew^.next := temp;
				     curposnew := curposnew^.next;
				    end;
				 end; (* if *)
			   curpos2 := curpos2^.next;
		     end; (* while *)
		  curpos1 := curpos1^.next;
	       end; (* while *)
	  999:   if curposnew <> nil then
		  begin
		     curposnew^.next := nil;
		     timestrek := newhead;
		  end (* if *)
	       else timestrek := nil;
	 end; (* else *)
end; (* timestrek *)

function concatpaths(paths1,paths2: pathsptr; neededges: boolean): pathsptr;
(* concatenates two paths and returns result.  it is a non-destructive
   function. *)
label 999;
var pathsnew: pathsptr;
begin
   new(pathsnew);
   pathsnew^.level := 0;
   pathsnew^.next := nil;
   pathsnew^.vertices := paths1^.vertices + paths2^.vertices;
   if neededges then
        pathsnew^.strength := paths1^.strength * paths2^.strength;
   concatpaths := pathsnew;
999: end;

function timesc(* (paths1,paths2: pathsptr; depth: integer;
                    vert: vertextype; neededges: boolean): pathsptr;  *);
(* this function takes the product of two paths.  it allows only one
   occurence of a given path in the final list.  for example, 
   (ab + cb) * (cd + ad) would return (abcd + abad + cbcd).  it does not
   allow any circular paths to be formed. *)

var temp,curpos1,curpos2,curposnew,newhead: pathsptr;

begin (* concat *)
   newhead := nil;
   if aaabort or (paths1 = nil)  or (paths2 = nil) then timesc := nil
   else
        begin
	    curpos1 := paths1;
	    curposnew := nil;
	    while curpos1 <> nil do
	       begin
		  curpos2 := paths2;
		  while curpos2 <> nil do
		     begin
                        if nointersect(curpos1,curpos2,vert) then
		      (* if no cycle *)
			      begin
			       temp := concatpaths(curpos1,curpos2,neededges);
			       if (isin(newhead,temp,depth) <> nil) then
			   (* if already in list don't add it *)
				 begin
				  disposepathslist(temp,0);
				  temp := nil;
				 end;
			       if temp <>  nil then
				 begin
				  temp^.level := depth;
				  if curposnew = nil then
				    begin
				     curposnew := temp;
				     newhead := curposnew;
				    end
				   else
				    begin
				     curposnew^.next := temp;
				     curposnew := curposnew^.next;
				    end;
				 end (* if *)
			      end; (* if *)
			curpos2 := curpos2^.next;
		     end; (* while *)
		  curpos1 := curpos1^.next;
	       end; (* while *)
	    if curposnew <> nil then
		  begin
		     curposnew^.next := nil;
		     timesc := newhead;
		  end (* if *)
	       else timesc := nil;
	 end; (* else *)
end; (* concat *)

procedure updatecorrscore(i,j: integer; prob: real; var held,nheld: real);
begin
   if (i <= numob) and (j <= numob) then
    if prob >= signif then held := held + prob
    else nheld := nheld + (1 - prob);
end;

procedure updatecorr(var corrlast: integer; i,j,level: integer; 
                     var trek: trekstype; var chimpheld,chimpnheld: real; 
                     newtrek: treksptr);
begin
   if (newtrek <> nil) and (trek^[i,j] = nil) then
      begin 
        if (corrimpl^[i,j] = notthere) then corrimpl^[i,j] := level + 1; 
        if (corrimpl^[j,i] = notthere) then corrimpl^[j,i] := level + 1;
        corrlast := corrlast + 1;
        updatecorrscore(i,j,corrproba^[i,j],chimpheld,chimpnheld);
      end;
    trek^[i,j] := newtrek;
end;

procedure calctrek1(var trek: trekstype; var paths: pathstype; level: integer; 
                    var corrlast: integer; var chimpheld,chimpnheld: real);
(* this procedure computes all of the treks between every pair of variables
   i and j and stores the result in trek[i,j]. *)

label 999;
var i,j,k: integer;

begin (* calctrek *)
   for i:= 1 to max do
     for j:=1 to max do
       if i = j then trek^[i,j] := nil 
       else if i < j then
	 begin
            if level = 1 then trek^[i,j] := nil;
	    checkinterrupt;
	    if aaabort then goto 999;
            if paths^[i,j] <> nil then
              if paths^[i,j]^.level >= level then
                 begin
      	           updatecorr(corrlast,i,j,level,trek,chimpheld,chimpnheld,
                           plustrek(maketrek(paths^[i,j],i,j,level),trek^[i,j]));
                   trek^[i,j]^.level := paths^[i,j]^.level;
                 end;
            if paths^[j,i] <> nil then
              if paths^[j,i]^.level >= level then
                 begin
      	           updatecorr(corrlast,i,j,level,trek,chimpheld,chimpnheld,
                           plustrek(maketrek(paths^[j,i],j,i,level),trek^[i,j]));
                   trek^[i,j]^.level := paths^[j,i]^.level
                 end;
	    for k:=1 to max do
             if (paths^[k,i] <> nil) and (paths^[k,j] <> nil) then
      	           updatecorr(corrlast,i,j,level,trek,chimpheld,chimpnheld,
   	                 plustrek(timestrek(paths^[k,i],paths^[k,j],level,k,
				     i,j),trek^[i,j]));
	 end (* for *)
       else trek^[i,j] := trek^[j,i];
999: end; (* calctrek *)

procedure calctrek(* var trek: trekstype; var paths: pathstype;
			   level: integer *);
(* this procedure computes all of the treks between every pair of variables
   i and j and stores the result in trek[i,j]. *)

label 999;
var i,j,k: integer;

begin (* calctrek *)
   for i:= 1 to max do
     for j:=1 to max do
       if i = j then trek^[i,j] := nil 
       else if i < j then
	 begin
            if level = 1 then trek^[i,j] := nil;
	    checkinterrupt;
	    if aaabort then goto 999;
            if paths^[i,j] <> nil then
              if paths^[i,j]^.level >= level then
                 begin
   	           trek^[i,j] := plustrek(maketrek(paths^[i,j],i,j,level),trek^[i,j]);
                   trek^[i,j]^.level := paths^[i,j]^.level;
                 end;
            if paths^[j,i] <> nil then
              if paths^[j,i]^.level >= level then
                 begin
      	           trek^[i,j] := plustrek(maketrek(paths^[j,i],j,i,level),trek^[i,j]);
                   trek^[i,j]^.level := paths^[j,i]^.level
                 end;
	    for k:=1 to max do
             if (paths^[k,i] <> nil) and (paths^[k,j] <> nil) then
   	       trek^[i,j] := plustrek(timestrek(paths^[k,i],paths^[k,j],level,k,
				     i,j),trek^[i,j]);
	 end (* for *)
       else trek^[i,j] := trek^[j,i];
999: end; (* calctrek *)

function plusc(* paths1,paths2: pathsptr; level: integer): pathsptr; *);
(* this function implements the '+' operation with the restriction that
   each path can occur in the list at most once.  *)

var curpos2,theend,temp: pathsptr;
(* theend represents the end of the list being constructed *)

begin (* plusc *)
   if aaabort or ((paths1 = nil) and (paths2 = nil)) then plusc := nil
   else if (paths1 = nil) and (paths2 <> nil) then plusc := paths2
   else if (paths1 <> nil) and (paths2 = nil) then plusc := paths1
      else
	 begin
	    theend := paths1;
	    while theend^.next <> nil do theend := theend^.next;
	    curpos2 := paths2;
	    while curpos2 <> nil do
	       begin
		  if (isin(paths1,curpos2,level) <> nil) then
			begin
			   temp := curpos2^.next;
			   dispose(curpos2);
			   curpos2 := temp;
			end
		     else
			begin
			   theend^.next := curpos2;
			   theend := theend^.next;
			   temp := curpos2^.next;
			   curpos2^.next := nil;
			   curpos2 := temp;
			end;
	       end;
	    plusc := paths1;
	 end; (* else *)
end; (* plusc *)

procedure addedge(* head,tail: edge; var treks: trekstype; 
	          var paths: pathstype; depth: integer;
                  var corrlast: integer; var chimpheld,chimpnheld: real *);
(* an edge is added in 3 steps.  first the edge itself is a 
   path between head and tail.  then all new paths that are paths
   starting at the head are formed by concatenating the new edge to the
   old paths starting at the tail.  finally, all other new paths
   are created by conatenating old paths that ended at
   the head with the new paths starting at the head *)

label 999;
var newpath,curpos2,pathhead,pathi,pathj,newedge,newuedge: pathsptr; 
    i,j,source,index: integer; curpos1,newtrek: treksptr;
    newheadpaths,newuheadpaths: array[1..maxvertex] of pathsptr;

begin
(* create path consisting of new edge *)
   newcycle := false;
   new(newedge);
   newedge^.next := nil;
   newedge^.vertices := [head,tail]; 
   newedge^.level := depth;

(* create all new paths starting at head *) (* stopped modifying here *)
  for i := 1 to max do
      begin
        if (paths^[tail,i] = nil) then newheadpaths[i] := nil
        else newheadpaths[i] := timesc(newedge,paths^[tail,i],depth,tail,false);
(*         newuheadpaths[i] := utimesc(newedge,upaths[tail,i],depth,tail);  *)
      end;
  newheadpaths[tail] := newedge;
(*  newuheadpaths[tail] := newuedge; *)
  
(* create all new paths formed by concatenating old paths to head with new
   paths from head; create all new treks formed by concatenating old treks
   to head with new treks to head *)
   for i := 1 to max do
    for j := 1 to max do 
     if (i <> j) and (j <> head) and (i <> head) then
      if (paths^[i,head] <> nil) and (newheadpaths[j] <> nil) then
        paths^[i,j] := plus(timesc(paths^[i,head],newheadpaths[j],depth,head,
                                  false),paths^[i,j]);

(* newheadpaths are paths from head to i *)
   for i := 1 to max do
      if (newheadpaths[i] <> nil) then
         paths^[head,i] := plusc(newheadpaths[i],paths^[head,i],depth);
   calctrek1(treks,paths,depth,corrlast,chimpheld,chimpnheld);
999: end;


(**************************************************************************)

procedure tetimp(i,j,k,l,level: integer; pattern1: boolean; 
                 var implied: boolean);
label 999;
var curpos1,curpos2: treksptr; 
    treki,trekj,trekk,trekl,trekil,trekjk: vertexset; 
    level1low,level2low: boolean;
begin   
  implied := true;
  if (treks1^[i,j] = nil) or (treks1^[k,l] = nil) then goto 999;
  curpos1 := treks1^[i,j];
  level1low := false;
  while (curpos1 <> nil) and (implied) do
     begin
       if (curpos1^.level < level) and (curpos1^.level <> 0) then 
         level1low := true;
       curpos2 := treks1^[k,l];
       if curpos2 <> nil then
          level2low := (curpos2^.level < level) and (curpos2^.level <> 0);
       if level1low and level2low then goto 999;
       while (curpos2 <> nil) and (implied) and
             (not (level1low and level2low)) do
         begin  
          treki := [curpos1^.source]; trekj := [curpos1^.source]; 
          trekk := [curpos2^.source]; trekl := [curpos2^.source];
          trekil := []; trekjk := [];
          if curpos1^.trek1 <> nil then treki := curpos1^.trek1^.vertices;
          if curpos1^.trek2 <> nil then trekj := curpos1^.trek2^.vertices;
          if curpos2^.trek1 <> nil then
            if pattern1 then trekk := curpos2^.trek1^.vertices
            else trekl := curpos2^.trek1^.vertices;
          if curpos2^.trek2 <> nil then
             if pattern1 then trekl := curpos2^.trek2^.vertices
             else trekk := curpos2^.trek2^.vertices;
          implied := not (((treki * trekk) = []) and ((trekj * trekl) = []));
          curpos2 := curpos2^.next;
          if curpos2 <> nil then
             if (curpos2^.level < level) and (curpos2^.level <> 0) then 
                level2low := true;
         end;  
       curpos1 := curpos1^.next;
     end;  
999: end;  

procedure comparepair(* i,j,k,l,pattern,level: integer; var now: boolean *);

var pattern1,pattern2: boolean;
begin
case pattern of 
   1: begin
        tetimp(i,j,k,l,level,false,now);
        if now then tetimp(i,k,j,l,level,false,now);
      end;
   2: begin
        tetimp(i,l,j,k,level,false,now);
        if now then tetimp(i,j,k,l,level,true,now);
      end;
   3: begin
        tetimp(i,k,j,l,level,true,now);
        if now then tetimp(i,l,j,k,level,true,now);
      end;
 end;
end;


(**************************************************************************)

procedure outputres(* a1,b1,c1,d1,a2,b2,c2,d2: integer; prob,locresid: real; 
	 	   isimplied,isheld: boolean *);
	 (* this outputs the equation, its residual, whether it is implied,
	    held by delta, and the prob. of wishart's difference statistic.*)

var impldcheck,heldcheck: char;
    i,linelength: integer;
begin
  setcheck(isimplied,impldcheck);
  setcheck(isheld,heldcheck);
  if ((corrimpl^[a1,b1] = notthere) or (corrimpl^[c1,d1] = notthere)) and
     ((corrimpl^[a2,b2] = notthere) or (corrimpl^[c2,d2] = notthere)) then impldcheck := 'T';
  if (isheld) and
    (((corrproba^[a1,b1] >= signif) or (corrproba^[c1,d1] >= signif)) and
     ((corrproba^[a2,b2] >= signif) or (corrproba^[c2,d2] >= signif))) then heldcheck := 'T';
  linelength:= 0;
  rswritename(a1,ofilvar,linelength); twrite(ofilvar,' |'); 
  rswritename(b1,ofilvar,linelength); twrite(ofilvar,', |');
  rswritename(c1,ofilvar,linelength); twrite(ofilvar,' |');
  rswritename(d1,ofilvar,linelength); twrite(ofilvar,' - |');
  rswritename(a2,ofilvar,linelength); twrite(ofilvar,' |');
  rswritename(b2,ofilvar,linelength); twrite(ofilvar,', |');
  rswritename(c2,ofilvar,linelength); twrite(ofilvar,' |');
  rswritename(d2,ofilvar,linelength); twrite(ofilvar,' |');
  for i:= 1 to (41 - linelength) do twrite(ofilvar,' |');
  tpwritereal(ofilvar,locresid,7,4,false);twrite(ofilvar,'  |');
  twritechar(ofilvar,impldcheck);twrite(ofilvar,'  |');
  twritechar(ofilvar,heldcheck);twrite(ofilvar,'  |');
  tpwritereal(ofilvar,prob,7,4,false);
  twriteeoln(ofilvar);
end;

procedure getprob(* var a: real; gamma: real *);

function interpolate(highprob, lowprob, hgamma, interval: real): real;
begin
        interpolate := highprob -
		  (gamma - hgamma) * (highprob - lowprob)/interval
end;

begin
   if (gamma < 0.01) then a := 0.992
   else if (gamma <= 0.05) then a := interpolate(0.992, 0.960, 0.01, 0.04)
   else if (gamma <= 0.10) then a := interpolate(0.960, 0.920, 0.05, 0.05)
   else if (gamma <= 0.20) then a := interpolate(0.920, 0.841, 0.10, 0.10)
   else if (gamma <= 0.30) then a := interpolate(0.841 ,0.764, 0.20, 0.10)
   else if (gamma <= 0.40) then a := interpolate(0.764 ,0.689, 0.30, 0.10)
   else if (gamma <= 0.50) then a := interpolate(0.689 ,0.617, 0.40, 0.10)
   else if (gamma <= 0.60) then a := interpolate(0.617 ,0.549, 0.50, 0.10)
   else if (gamma <= 0.70) then a := interpolate(0.549 ,0.484, 0.60, 0.10)
   else if (gamma <= 0.80) then a := interpolate(0.484 ,0.424, 0.70, 0.10)
   else if (gamma <= 0.90) then a := interpolate(0.424 ,0.368, 0.80, 0.10)
   else if (gamma <= 1.00) then a := interpolate(0.368 ,0.317, 0.90, 0.10)
   else if (gamma <= 1.10) then a := interpolate(0.317 ,0.271, 1.00, 0.10)
   else if (gamma <= 1.20) then a := interpolate(0.271 ,0.230, 1.10, 0.10)
   else if (gamma <= 1.30) then a := interpolate(0.230 ,0.194, 1.20, 0.10)
   else if (gamma <= 1.40) then a := interpolate(0.194 ,0.162, 1.30, 0.10)
   else if (gamma <= 1.50) then a := interpolate(0.162 ,0.134, 1.40, 0.10)
   else if (gamma <= 1.60) then a := interpolate(0.134 ,0.110, 1.50, 0.10)
   else if (gamma <= 1.70) then a := interpolate(0.110 ,0.089, 1.60, 0.10)
   else if (gamma <= 1.80) then a := interpolate(0.089 ,0.072, 1.70, 0.10)
   else if (gamma <= 1.90) then a := interpolate(0.072 ,0.057, 1.80, 0.10)
   else if (gamma <= 2.00) then a := interpolate(0.057 ,0.046, 1.90, 0.10)
   else if (gamma <= 2.10) then a := interpolate(0.046 ,0.036, 2.00, 0.10)
   else if (gamma <= 2.20) then a := interpolate(0.036 ,0.028, 2.10, 0.10)
   else if (gamma <= 2.30) then a := interpolate(0.028 ,0.021, 2.20, 0.10)
   else if (gamma <= 2.40) then a := interpolate(0.021 ,0.016, 2.30, 0.10)
   else if (gamma <= 2.50) then a := interpolate(0.016 ,0.012, 2.40, 0.10)
   else if (gamma <= 2.60) then a := interpolate(0.012 ,0.009, 2.50, 0.10)
   else if (gamma <= 2.70) then a := interpolate(0.009 ,0.007, 2.60, 0.10)
   else if (gamma <= 2.80) then a := interpolate(0.007 ,0.005, 2.70, 0.10)
   else if (gamma <= 2.90) then a := interpolate(0.005 ,0.004, 2.80, 0.10)
   else if (gamma <= 3.00) then a := interpolate(0.004 ,0.003, 2.90, 0.10)
   else if (gamma <= 3.10) then a := interpolate(0.003 ,0.002, 3.00, 0.10)
   else if (gamma <= 3.20) then a := interpolate(0.002 ,0.0014, 3.10, 0.10)
   else if (gamma <= 3.30) then a := interpolate(0.0014 ,0.0010, 3.20, 0.10)
   else if (gamma <= 3.40) then a := interpolate(0.0010 ,0.0006, 3.30, 0.10)
   else if (gamma <= 3.50) then a := interpolate(0.0006 ,0.000466, 3.40, 0.10)
   else if (gamma <= 3.60) then a := interpolate(0.000466 ,0.000262, 3.50, 0.10)
   else if (gamma <= 3.70) then a := interpolate(0.000262 ,0.000216, 3.60, 0.10)
   else if (gamma <= 3.80) then a := interpolate(0.000216 ,0.000144, 3.70, 0.10)
   else if (gamma <= 3.90) then a := interpolate(0.000144 ,0.000096, 3.80, 0.10)
   else if (gamma <= 4.00) then a := interpolate(0.000096 ,0.000064, 3.90, 0.10)
   else if (gamma <= 4.10) then a := interpolate(0.000064 ,0.000042, 4.00, 0.10)
   else if (gamma <= 4.20) then a := interpolate(0.000042 ,0.00026, 4.10, 0.10)
   else if (gamma <= 4.30) then a := interpolate(0.000026 ,0.000018, 4.20, 0.10)
   else if (gamma <= 4.40) then a := interpolate(0.000018 ,0.000010, 4.30, 0.10)
   else if (gamma <= 4.50) then a := interpolate(0.000010 ,0.000006, 4.40, 0.10)
   else if (gamma <= 4.60) then a := interpolate(0.000006 ,0.000004, 4.50, 0.10)
   else if (gamma <= 4.70) then a := interpolate(0.000004 ,0.000002, 4.60, 0.10)
   else if (gamma <= 4.80) then a := interpolate(0.000004 ,0.000004, 4.70, 0.10)
   else a := 0.000;
end; (*getprob*)

procedure calcstat(* a,b,c,d: integer; locresid: real; var prob: real;
	           pattern: integer *);
	  (* this calculates the wishart statistic for the
	   variance of the sample tetrad difference *)

var gamma, stdev, d1, d2, d3, d4: real;

begin
   stdev := 0;
   gamma := 0;
   prob := 0;

   d1 := rho^[b,b] * (rho^[c,c] * rho^[d,d] - rho^[d,c] * rho^[d,c]) -
	 rho^[b,c] * (rho^[c,b] * rho^[d,d] - rho^[d,c] * rho^[d,b]) +
	 rho^[b,d] * (rho^[c,b] * rho^[d,c] - rho^[c,c] * rho^[d,b]);

   d2 :=  rho^[b,a] * (rho^[c,c] * rho^[d,d] - rho^[d,c] * rho^[d,c]) -
	 rho^[b,c] * (rho^[c,a] * rho^[d,d] - rho^[d,c] * rho^[d,a]) +
	 rho^[b,d] * (rho^[c,a] * rho^[d,c] - rho^[c,c] * rho^[d,a]);

   d3 := rho^[b,a] * (rho^[c,b] * rho^[d,d] - rho^[d,c] * rho^[d,b]) -
	 rho^[b,b] * (rho^[c,a] * rho^[d,d] - rho^[d,c] * rho^[d,a]) +
	 rho^[b,d] * (rho^[c,a] * rho^[d,b] - rho^[c,b] * rho^[d,a]);

   d4 :=  rho^[b,a] * (rho^[c,b] * rho^[d,c] - rho^[c,c] * rho^[d,b]) -
	 rho^[b,b] * (rho^[c,a] * rho^[d,c] - rho^[c,c] * rho^[d,a]) +
	 rho^[b,c] * (rho^[c,a] * rho^[d,b] - rho^[c,b] * rho^[d,a]);

   stdev := ((sampsize+1)/ ((sampsize-1) * (sampsize-2)) *
	    (rho^[a,a] * rho^[b,b] - rho^[a,b] * rho^[a,b]) *
	    (rho^[c,c] * rho^[d,d] - rho^[c,d] * rho^[c,d])) -
	    ((rho^[a,a]*d1 - rho^[a,b]*d2 + rho^[a,c]*d3 -
	    rho^[a,d]*d4)/(sampsize - 2));
   if stdev < 0 then myerror('Standard deviation less than 0          ',fatal,
   			       false)
   else stdev := sqrt(stdev);
   if stdev = 0.0 then stdev := 0.00000001;
   gamma := locresid/stdev;
   getprob(prob, gamma);
end;  (* calcstat *)

procedure update(* localimp: implist; singleedge: boolean; locresid: real;
                 var chimpheld,chimpnheld: real; *);
var prob: real;
begin
  if (localimp^.prob > signif) then chimpheld := chimpheld + localimp^.prob
  else chimpnheld := chimpnheld + (1 - localimp^.prob);
  if singleedge then
   with localimp^ do
    begin
      case localimp^.pattern of
       1: outputres(i,j,k,l,i,k,j,l,prob,locresid,false,prob > signif);
       2: outputres(i,l,j,k,i,j,k,l,prob,locresid,false,prob > signif);
       3: outputres(i,k,j,l,i,l,j,k,prob,locresid,false,prob > signif);
      end;
   end;
end;

(*-----------#pc-version-----------------------------------------------------*)
(*
end.
*)
(*-----------#pc-version-----------------------------------------------------*)