(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "part.h"
#include "calctet.h"
#include "sys.h"
#include "indep1.h"
#include "up.h"
#include "util.h"
#include "mon.h"
#include "cross.h"
#include "rsutil.h"
#include "io.h"
(*-----------#unix-version-----------------------------------------------------*)


(* to do
   generalize calcstats to work for classvar's that are not BINARY
*)


(*-----------------------------------------------------------------------------*)
(* SPLIT SAMPLE *)

(* this function sets the last TESTPER percent of the samples to be test samples
   for crossvalidation *)

procedure setupsamparray(*var samparray:sampletypearray;  testlow, testhigh, estlow, esthigh, trainlow, trainhigh: integer *);
var 
  i:integer;
begin
  for i := 1 to tround(sampsize) do samparray[i] := ignore;
  for i := testlow to testhigh do samparray[i]:=test;
  if not trainandest then
     begin
       for i := trainlow to trainhigh do samparray[i] := train;
       for i := estlow to esthigh do samparray[i] := est;
     end
  else for i := trainlow to trainhigh do samparray[i] := trainest;
end;

procedure initsamparray(* var samparray:sampletypearray *); 
var i: integer;
begin
    if trainandest then 
       for i:=1 to round(sampsize) do samparray[i]:=trainest
    else for i:=1 to round(sampsize) do samparray[i]:=train;
end;


(*
procedure checksamparray(samparray:sampletypearray);
var
 i,ctest,cest,ctrain,cte:integer;
begin
    cte:=0;
    ctest:=0;
    cest:=0;
    ctrain:=0;
    writeln(output,round(sampsize));
    for i:=1 to round(sampsize) do
	begin
  	if samparray[i]=trainest then cte:=cte+1;
  	if samparray[i]=est then cest:=cest+1;
  	if samparray[i]=train then ctrain:=ctrain+1;
  	if samparray[i]=test then ctest:=ctest+1;
        end;
    writeln(output,'cte = ',cte);
    writeln(output,'cest = ' ,cest);
    writeln(output,'ctrain = ' , ctrain);
    writeln(output,'ctest = ' , ctest);
end;
*)

(* must have RAW DISCRETE data (can it be cellcount?) *)
procedure dosplit;
begin
  twrite(output,'Do you want to split the sample into two parts ? [y]|');
  trainandest:=compareto(yes);
  twriteeoln(output);
  twrite(output,'|');
  trainlow := tround(getnumber(0,15000,false,'Starting point for training? |',
                        trainlow));
  trainhigh := tround(getnumber(0,15000,false,'End point for training? |',
                        trainhigh));
  testlow := tround(getnumber(0,15000,false,'Starting point for testing? |',
                        testlow));
  testhigh := tround(getnumber(0,15000,false,'End point for testing? |',
                        testhigh));
  if not trainandest then
      begin
         estlow := tround(getnumber(0,15000,false,
                                   'Starting point for estimation? |',estlow));
         esthigh := tround(getnumber(0,15000,false,'End  point for estimation? |',
                              esthigh));
      end;
  setupsamparray(samparray,testlow,testhigh,estlow,esthigh,trainlow,trainhigh);
(*  checksamparray(samparray);*)
end;

(* END SPLIT SAMPLE *)
(*-----------------------------------------------------------------------------*)



(* function finds the value which has the highest condistional probability after
   running update *)
function findbestcat:categorynum;
var
  bestscore:real;
  bestcat,j:categorynum;

begin
  bestscore:=-10000;
  if updatemethod=analytic
  then
   begin
	for j := 1 to cats[classvar].num do
	    if exactfreqcard[classvar,j] >= bestscore 
	    then begin bestscore:=exactfreqcard[classvar,j];bestcat:=j; end;
   end
  else
   begin
	for j := 1 to cats[classvar].num do
	    if lwtc[classvar,j] >= bestscore 
	    then begin bestscore:=lwtc[classvar,j];bestcat:=j; end;
   end;
  findbestcat:=bestcat;
end;


(* function finds the the conditional probability for the positive value associated with the classvar *)
function findpvalueforpositive(*classvar,poscat:integer):real;*);
var
  bestscore:real;
  bestcat,j:categorynum;

begin
  bestscore:=-10000;
  if updatemethod=analytic
  then
     findpvalueforpositive:=exactfreqcard[classvar,poscat]
  else
     findpvalueforpositive:=(lwtc[classvar,poscat]/likelihoodcount);
end;


procedure getclass(var classname:vertexnametype;var classvar:vertextype);
var boo:boolean;
begin
  boo:=false;
  twrite(output,'Classification variable: |');
  readttyline(buf);
  if not(myeoln) then
  begin
    readname(classname,ignoreall);
    if (ignoreall in errors) then
      begin
	errors:=errors-[ignoreall];
      end;
    classvar:=listposition(classname,boo,vernames,boo);
  end;
end;

(* doesn't check for input's correctness *)
(* procedure used to get the category associated with the positive value of the
   classification variable *)
procedure getposcat(* classvar:integer; var poscat:integer *);
label 999;
var j,rounded: integer;
    realnum:real;
begin
999:
  twrite(output,'Positive value for classification variable: |');
  readttyline(buf);
  if myeoln then goto 999;
  realnum:=readnum(true,warning);
  rounded := tround(realnum);
  poscat := 0;
  for j := 1 to maxval[classvar] do 
    if varcatvalue[classvar,j] = rounded then poscat := j;
end;

procedure writerocheader(var rocfile:text;poscat,classvar:integer);
begin

  twriteln(rocfile,'{This file contains the sample value|');
  twriteln(rocfile,'for the classification variable and the conditional|');
  twriteln(rocfile,'probability of the positive value occurring|');
  twrite(rocfile,'Classification variable: |');
  writevername(vernames[classvar],rocfile);
  twriteeoln(rocfile);
  twrite(rocfile,'Positive value: |');
  twriteint(rocfile,round(cats[classvar].value[poscat]));
  twriteeoln(rocfile);
  twriteln(rocfile,'Value  Prob        }|');
  twriteeoln(rocfile);
end;

(* calcstats calculates the number of false positives, false negatives, 
   true positives and true negatives -- this assumes that the variable
   we are classifying is boolean (e.g. diagnostic test) and the value associated with
   positive or true is 1  *)
procedure calcstats(samparray:sampletypearray;datamat:dataarr;var rocfile:text;roc:boolean);
var
  bestcat:integer;
  classcat,j:categorynum;
  i,poscat,negcat:integer;
  pvalue,pcutoff:real;
begin
  testsize:=0;
(*
  twriteln(output,'Assuming that 1 is positive value.|');
  twriteln(output,'Assuming that 0 is negative value.|');
  poscat:=getnumcat(classvar,1,cats);
  negcat:=getnumcat(classvar,0,cats);
*)
  getposcat(classvar,poscat);

  twrite(output,'P-value cutoff for positive classification? : |');
  readttyline(buf);
  pcutoff:=readnum(true,warning);
  twriteeoln(output);

  if roc then writerocheader(rocfile,poscat,classvar);
(*         pcutoff:=0.500;*)
  for i:=1 to round(sampsize) do
     if (samparray[i]=test) then
	begin
	 testsize:=testsize+1;
	 classcat:=getbigdata(i,classvar);
	 for j:=1 to max do
	    if j<>classvar then
		addevidence(evidencelist,evidenceset,vernames[j],j,cats[j].value[getbigdata(i,j)]);
	 updatebayes;


	 pvalue:= findpvalueforpositive(classvar,poscat);
         if roc then 
	 begin
		twriteint(rocfile,round(cats[classvar].value[classcat]));
		twrite(rocfile,'    |');
		tpwritereal(rocfile,pvalue,8,6,true);
		twriteeoln(rocfile);
	 end;
         if pvalue>=pcutoff then bestcat:=poscat else bestcat:=poscat+1;
	 if (bestcat=poscat) and (classcat=poscat) then truepos:=truepos+1;
	 if (bestcat<>poscat) and (classcat<>poscat) then trueneg:=trueneg+1;
	 if (bestcat<>poscat) and (classcat=poscat) then falseneg:=falseneg+1;
	 if (bestcat=poscat) and (classcat<>poscat) then falsepos:=falsepos+1;
	end;
end;


(* procedure getstats assumes that truepos,trueneg,falsepos and falseneg 
   are set correctly*)
procedure getstats(var curmodel: modeltypeptr; numleft,testsize: integer;
                   pospercent,percentpop: real; poscat: vertextype);
begin
  with curmodel^ do
    begin
      acc:=(truepos + trueneg)/testsize;
      if (falsepos+truepos)=0 then pospred := -1 else
         pospred:=(truepos/(falsepos + truepos));
      if (falseneg + trueneg)=0 then negpred := -1 else
         negpred:=(trueneg/(falseneg + trueneg)); 
      if (falseneg+truepos)=0 then sens:=-1 else
         sens:=(1 - (falseneg/(falseneg+truepos)));
      if (falsepos+trueneg)=0 then spec:=-1 else
         spec:=(1 - (falsepos/(falsepos+trueneg)));
      if (pospred >= pospercent) and (mostpos = 0) then 
           mostpos := 1 - (numleft/testsize);
      if ((1 - numleft/testsize) >= percentpop) then mostneg := pospred;
   end;
end;

procedure writeesttype(var ofile: text; estimate: esttype);
begin
    twrite(ofile,'Estimate Type : |');
    case estimate of
         MLinternal: twriteln(ofile,'MLinternal |');
         MLexternal: twriteln(ofile,'MLexternal |');
         uniform: twriteln(ofile,'Uniform |');
         noisyormeth: twriteln(ofile,'Noisy Or |');
    end;
end;

procedure writemodel(*var ofile: text; amodel: modelt*);
var i,j: vertextype;
begin
    for i := 1 to max do
       if i in graphvertices then
          for j := 1 to max do
             if (j in graphvertices) and (j in amodel[i]) then
                begin
                   writename(j,ofile);
                   twrite(ofile,' -> |');
                   writename(i,ofile);
                   twriteeoln(ofile);
                end;
    twriteeoln(ofile);
end;

procedure printstats(* var ofile:text; var curmodel: modeltypeptr *);
var
  temp:real;
begin
  with curmodel^ do
    begin
       twriteln(ofile,'The statistics for this run.|');
       twriteeoln(ofile);
       twrite(ofile,'Test sample size : |');
       twriteint(ofile,testsize);
       twriteeoln(ofile);
       twrite(ofile,'Specificity : |');
       twritereal(ofile,spec);
       twriteeoln(ofile);
       twrite(ofile,'Sensitivity : |');
       twritereal(ofile,sens);
       twriteeoln(ofile);
       twrite(ofile,'Positive predictive value : |');
       twritereal(ofile,pospred);
       twriteeoln(ofile);
       twrite(ofile,'Negative predictive value : |');
       twritereal(ofile,negpred);
       twriteeoln(ofile);
       twrite(ofile,'Accuracy : |');
       twritereal(ofile,acc);
       twriteeoln(ofile);
       twrite(ofile,'Significance : |');
       twritereal(ofile,signif);
       twriteeoln(ofile);
       twrite(ofile,'Precision : |');
       twriteint(ofile,precision);
       twriteeoln(ofile);
       twrite(ofile,'Mostpos : |');
       twritereal(ofile,mostpos);
       twriteeoln(ofile);
       twrite(ofile,'Mostneg : |');
       twritereal(ofile,mostneg);
       twriteeoln(ofile);
       twrite(ofile,'Acccut : |');
       twritereal(ofile,acccut);
       twriteeoln(ofile);
       writeesttype(ofile,estimate);
       writemodel(ofile,amodel);
       twriteeoln(ofile);
  end;
end;

procedure printshortstats(var ofile:text;curmodel: modeltypeptr);
var
  temp:real;
begin
  with curmodel^ do
     begin
       twrite(ofile,'  Spec:|');
       twritereal(ofile,spec);
       twrite(ofile,' Sens:|');
       twritereal(ofile,sens);
       twrite(ofile,' PPV:|');
       twritereal(ofile,pospred);
       twrite(ofile,' NPV:|');
       twritereal(ofile,negpred);
       twrite(ofile,' ACC:|');
       twritereal(ofile,acc);
       twriteeoln(ofile);
       twrite(ofile,' MOSTPOS:|');
       twritereal(ofile,mostpos);
       twriteeoln(ofile);
       twrite(ofile,' MOSTNEG:|');
       twritereal(ofile,mostneg);
       twriteeoln(ofile);
    end;
end;

procedure initcross;
begin
  falsepos:=0;
  falseneg:=0;
  truepos:=0;
  trueneg:=0;
end;

procedure docrossvalidate;
label 999;
var
  rocfile:text; curmodel: modeltypeptr;
  mybuf:inbuf; 
  roc,boo,boo2:boolean;
  acc,ppv,npv,spec,sens:real;
  poscat:integer; maxtype: infotype; pospercent,percentpop: real;
begin

  new(curmodel);
  curmodel^.estimate := uniform;
  curmodel^.precision := 1;

  boo:=false;
  mybuf:='';
999:
  twrite(output,'Classification variable: |');
  readttyline(buf);
  if not(myeoln) then
  begin
    readname(classname,ignoreall);
    if (ignoreall in errors) then
      begin
	errors:=errors-[ignoreall];
	goto 999;
      end;
    gettypetomaximize(maxtype,pospercent,percentpop);
    classvar:=listposition(classname,boo,vernames,boo);
    if classvar=0 then goto 999;
    if cats[classvar].num>2 then
	begin
	  twriteln(output,'Cross-validation halted: current implementation only for binary classification variables.|');
	  goto 999;
	end;
    getposcat(classvar,poscat);
    twrite(output,'Do you want to create an ROC file? [y] |');
    roc:=compareto(yes);
    if roc then getfile(rocfile,'File name for ROC information: |',outputf,mybuf,boo2);

  initmod(curmodel);

(* only matters that it is not noisyor *)
  doautocrossvalidate(poscat,curmodel,pospercent,percentpop,uniform,maxtype);
  printstats(output,curmodel);
  disposebayesnet(cats);
  dispose(curmodel);
 end;
end;

(*++++++++++++++++++++++++++++++++++++++++++++++++++++++*)
(* autocrossvalidate *)

procedure setuppredarray(var predarray:predarraytype;classvar:vertextype;poscat:integer; estimate: esttype);
var i:integer; j: vertextype;
  classcat:categorynum;
  pvalue:real;

begin
    testsize:=0;
    for i:=1 to round(sampsize) do
     if (samparray[i]=test) then
	begin
	 testsize:=testsize+1;
	 classcat:=getbigdata(i,classvar);
         if estimate <> noisyormeth then
            begin
              for j:=1 to max do
                 if ((j<>classvar) and (j in graphvertices)) then
            	     addevidence(evidencelist,evidenceset,vernames[j],j,cats[j].value[getbigdata(i,j)]);
               updatebayes;
               pvalue:= findpvalueforpositive(classvar,poscat);
              end
          else
            begin
              pvalue := 1;
              for j:=1 to max do
                 if ((j<>classvar) and (j in graphvertices)) then
                       pvalue := pvalue * (1 - noisyor[j,getbigdata(i,j)].pos);
              pvalue := 1 - pvalue; 
            end;
(*         if (pvalue>0.99999) then 
		writeln(output,'probability greater than .99999 for sample ',i);*)
	 predarray[testsize].cat:=classcat;
	 predarray[testsize].prob:=pvalue;
        end;
end;

procedure splitpredarray(var predarray:predarraytype;poscat:categorynum;var pospoint,negpoint:integer);
var i:integer;
    prob:real;
    cat:categorynum;
begin
  pospoint:=1;
  negpoint:=1;
  for i:=1 to testsize do
   begin
	prob:=predarray[i].prob;
	cat:=predarray[i].cat;
        if (poscat=cat) then
	   begin
		predarray[pospoint].prob:=prob;
		predarray[pospoint].cat:=cat;
		pospoint:=pospoint+1;
	   end
	else
	   begin
		predarray[maxsamp+negpoint].prob:=prob;
		predarray[maxsamp+negpoint].cat:=cat;
		negpoint:=negpoint+1;
	   end;
    end;
  pospoint:=pospoint-1;
  negpoint:=negpoint-1;
end;


procedure swappred(i,j:integer);
var tempcat:integer;
    tempprob:real;
begin
  	   tempprob:=predarray[i].prob;
	   tempcat:=predarray[i].cat;
	   predarray[i].prob:=predarray[j].prob;
	   predarray[i].cat:=predarray[j].cat;
	   predarray[j].prob:=tempprob;
	   predarray[j].cat:=tempcat;
end;

function sortedpred(lower,upper:integer):boolean;
var boo:boolean;
    i:integer;
begin
   boo:=true;
   for i:=lower to upper-1 do
     if predarray[i].prob>predarray[i+1].prob then
	boo:=false;
   sortedpred:=boo;
end;

procedure quicksortpred(lower,upper:integer);
var splitval:real;
    bottom,top,split:integer;
begin
  if lower< upper then
  begin
  bottom:=lower+1;
  top:=upper;
  split:=lower;
  splitval:=predarray[split].prob;
  repeat
     if splitval>=predarray[bottom].prob then
      begin
	swappred(split,bottom);
	split:=bottom;
        bottom:=bottom+1;
       end
     else
      begin
	swappred(bottom,top);
	top:=top-1;
      end;
   until (split=top);
  if not(sortedpred(lower,split)) then
        quicksortpred(lower,split);
  if (split<upper) then
    if not(sortedpred(split+1,upper)) then
        quicksortpred(split+1,upper);
  end;
end;

procedure bublesortpred(lower,upper:integer);
var i,j,tempcat:integer;
    tempprob:real;
begin
  for i:= lower to (upper - 1) do
  for j:= (i+1) to upper do
       if (predarray[i].prob>predarray[j].prob) then
           swappred(i,j);
end;

procedure checksort(lower,upper:integer);
var i:integer;
begin
for i:=lower to (upper-1) do
   if (predarray[i].prob>predarray[i+1].prob) then writeln(output,'Error in array sort ',i);
end;

procedure setcounts(cutoff:real;numpos,numneg:integer);
var i:integer;
begin
  i:=0;
  repeat
	i:=i+1
  until ((predarray[i].prob>cutoff) or (i=numpos+1));
  i:=i-1;
  falseneg:=i;
  truepos:=numpos-i;
  i:=maxsamp;
  repeat
	i:=i+1
  until ((predarray[i].prob>cutoff) or (i=numneg+maxsamp+1));
  i:=i-1-maxsamp;
  trueneg:=i;
  falsepos:=numneg-i;
end;

procedure calcnumposandneg(var numpos,numneg,testsize,poscat:integer);
var i:integer;
begin
  numpos:=0;numneg:=0;
  for i:=1 to testsize do
    if predarray[i].cat=poscat then numpos:=numpos+1
    else numneg:=numneg+1;
end;

procedure getnextcutoff(var numleft,numpos,numneg,poscat:integer);
var i:integer;
begin
  numleft:=numleft+1;
  while ((predarray[numleft].prob>=predarray[numleft+1].prob) and (numleft<testsize)) do
    begin
      if predarray[numleft].cat=poscat then falseneg:=falseneg+1
      else trueneg:=trueneg+1;
      numleft:=numleft+1;
    end;
   if predarray[numleft].cat=poscat then falseneg:=falseneg+1
   else trueneg:=trueneg+1;
   falsepos:=numneg-trueneg;
   truepos:=numpos-falseneg;
   crosscutoff:=predarray[numleft].prob;
end;

procedure copymodel(*var frommodel,tomodel: modeltypeptr*);
var i: vertextype;
begin
   if (frommodel <> nil) and (tomodel <> nil) then
      with tomodel^ do
         begin
           precision := frommodel^.precision;
           estimate := frommodel^.estimate;
           signif := frommodel^.signif;
           acc := frommodel^.acc;
           sens := frommodel^.sens;
           spec := frommodel^.spec;
           pospred := frommodel^.pospred;
           negpred := frommodel^.negpred;
           mostpos := frommodel^.mostpos;
           mostneg := frommodel^.mostneg;
           mostposcut := frommodel^.mostposcut;
           mostnegcut := frommodel^.mostnegcut;
           acccut:= frommodel^.acccut;
           senscut := frommodel^.senscut;
           speccut := frommodel^.speccut;
           pospredcut := frommodel^.pospredcut;
           negpredcut := frommodel^.negpredcut;
           for i := 1 to max do amodel[i] := frommodel^.amodel[i];
        end;
end;

procedure comparemodel(*var bestmodel,curmodel: modeltypeptr; maxtype: infotype;printmod:boolean*);
var switch: boolean;
begin
   case maxtype of
      accuracy: switch := bestmodel^.acc < curmodel^.acc;
      sensitivity: switch := bestmodel^.sens < curmodel^.sens;
      specificity: switch := bestmodel^.spec < curmodel^.spec;
      positivepredvalue: switch := bestmodel^.pospred < curmodel^.pospred;
      negativepredvalue: switch := bestmodel^.negpred < curmodel^.negpred;
      mostpos: switch := bestmodel^.mostpos < curmodel^.mostpos;
      mostneg: switch := bestmodel^.mostneg < curmodel^.mostneg;
   end;
   if switch then
        begin
            copymodel(curmodel,bestmodel);
	    if printmod then printstats(output,curmodel);
        end;
end;

procedure maxmodel(var bestmodel,curmodel: modeltypeptr);
var switch: boolean; typeofinfo: infotype;
begin
 for typeofinfo := accuracy to mostneg do
   case typeofinfo of 
      accuracy: if bestmodel^.acc < curmodel^.acc then
                   begin
                      bestmodel^.acc := curmodel^.acc;
                      bestmodel^.acccut := crosscutoff;
                   end;
      sensitivity: if bestmodel^.sens < curmodel^.sens then
                     begin
                        bestmodel^.sens := curmodel^.sens;
                        bestmodel^.senscut := crosscutoff;
                      end;
      specificity: if bestmodel^.spec < curmodel^.spec then
                      begin
                         bestmodel^.spec := curmodel^.spec;
                         bestmodel^.speccut := crosscutoff;
                      end;                     
      positivepredvalue: if bestmodel^.pospred < curmodel^.pospred then
                             begin
                               bestmodel^.pospred := curmodel^.pospred;
                               bestmodel^.pospredcut := crosscutoff;
                             end;                     
      negativepredvalue: if bestmodel^.negpred < curmodel^.negpred then
                            begin
                                bestmodel^.negpred := curmodel^.negpred;
                                bestmodel^.negpredcut := crosscutoff;
                             end;                     
      mostpos: if bestmodel^.mostpos < curmodel^.mostpos then
                    begin
                        bestmodel^.mostpos := curmodel^.mostpos;
                        bestmodel^.mostposcut := crosscutoff;
                     end;                     
      mostneg: if bestmodel^.mostneg < curmodel^.mostneg then
                     begin
                        bestmodel^.mostneg := curmodel^.mostneg;
                        bestmodel^.negpredcut := crosscutoff;
                     end;                     
   end;
end;

procedure writepredarray;
var i,testsize: integer;
begin
   testsize:=0;
   for i:=1 to round(sampsize) do
    if (samparray[i]=test) then
     begin
      writeln(ofilvar,predarray[testsize].cat,' ',predarray[testsize].prob:6:3);
      testsize := testsize + 1;
     end;
end;

procedure doautocrossvalidate(* poscat: integer; var curmodel: modelptrtype;
                                pospercent,percentpop: real; 
                                estimate: esttype; maxtype: infotype *);
var
  rocfile:text;
  mybuf:inbuf;
  roc,boo,boo2:boolean;
  i,numleft,numpos,numneg:integer;
  maxacc,acccut,maxppv,ppvcut,maxnpv,npvcut:real;
  tempmodel: modeltypeptr; debug: boolean;

begin
  debug := false;
  new(tempmodel);
  falsepos:=0;  falseneg:=0;  truepos:=0;  trueneg:=0;
  setuppredarray(predarray,classvar,poscat,estimate);
  calcnumposandneg(numpos,numneg,testsize,poscat);
  bublesortpred(1,numpos+numneg);
  maxacc:=0;acccut:=0;maxppv:=0;ppvcut:=0;maxnpv:=0;npvcut:=0;
  numleft:=0;
  falseneg:=0;truepos:=numpos;trueneg:=0;falsepos:=numneg;
  while numleft<testsize do
   begin
     getnextcutoff(numleft,numpos,numneg,poscat);
     getstats(tempmodel,numleft,testsize,pospercent,percentpop,poscat);
     tempmodel^.acccut := crosscutoff;
     tempmodel^.precision := curmodel^.precision;
     tempmodel^.signif := curmodel^.signif;
     tempmodel^.amodel := curmodel^.amodel;
     comparemodel(curmodel,tempmodel,maxtype,false);
     if debug then writepredarray;
   end;
  dispose(tempmodel);
end;

procedure getesttype(var estimate: esttype);
var num: integer;
begin
   twriteln(output,'1: Maximum likelihood |');
   twriteln(output,'2: Uniform |');
   twriteln(output,'3: Noisy Or |');
   twrite(output,'Type in numbers of estimate type: |');
   num := tround(getnumber(1,3,true,'Type in number of estimate type: |',1));
   if num = 1 then estimate := MLinternal 
   else if num = 2 then estimate := uniform
   else if num = 3 then estimate := noisyormeth;
end;

procedure makerocfile;
label 999;
var
  boo:boolean;
  pvalue:real;
  classcat,j:categorynum;
  i,poscat,numpos,numneg,precnum:integer;
  mybuf:inbuf; estimate: esttype; num: real;
begin
  boo:=false;
  mybuf:='';
999:
  twrite(output,'Classification variable: |');
  readttyline(buf);
  if not(myeoln) then
  begin
    readname(classname,ignoreall);
    if (ignoreall in errors) then
      begin
	errors:=errors-[ignoreall];
	goto 999;
      end;
    classvar:=listposition(classname,boo,vernames,boo);
    if classvar=0 then goto 999;
    if cats[classvar].num>2 then
	begin
	  twriteln(output,'Cross-validation halted: current implementation only for binary classification variables.|');
	  goto 999;
	end;
    getposcat(classvar,poscat);
    getesttype(estimate);
    precnum := tround(getnumber(0.0,50.0,true,'Precision: |',0.0)); 
    doestimate(adjacency,output,true,estimate,precnum);

(* anything except noisyor ok *)
    setuppredarray(predarray,classvar,poscat,uniform);
    calcnumposandneg(numpos,numneg,testsize,poscat);
    bublesortpred(1,numpos+numneg);
    for i:=1 to testsize do
	begin
		classcat:=predarray[i].cat;
		pvalue:=predarray[i].prob;
		twriteint(ofilvar,round(cats[classvar].value[classcat]));
		twrite(ofilvar,'    |');
                if classcat=poscat then
		      tpwritereal(ofilvar,pvalue,8,6,true)
		else  tpwritereal(ofilvar,(1-pvalue),8,6,true);
		twriteeoln(ofilvar);
        end;
        flush(ofilvar);
	twriteeoln(ofilvar);
 end;
end;

(* end of autocrossvalidate *)
(*++++++++++++++++++++++++++++++++++++++++++++++++++++++*)


(* temp stuff to read in data and get statistics *)
(* File layout :
      First row must have the number of rows of data (n)
      then n rows of <truevalue> and P(vitstat=<truevalue>|<allvarsinmodel>)
*)


procedure dogetbestacc;
var
	numleft,i,numpoints,posval,numpos,numneg,classval:integer;
        mybuf:inbuf; 
	pvalue:real;
	predfile:text;
	tempmodel,bestmodel: modeltypeptr;
	ofilename:inbuf;
	changed:boolean;

begin
   new(tempmodel);  mybuf:=' |';
   new(bestmodel);
   initmod(tempmodel);
   initmod(bestmodel);

   getfile(predfile,'File name for prediction information: |',inp,ofilename,changed);
   
  twrite(output,'Positive value for classification variable: |');
  readttyline(buf);
  posval:=round(readnum(true,warning));

   readfileline(predfile,buf);
   numpoints:=round(readnum1(predfile,false,ignoreall));
   (*the following loop sets up the predarray*)
   for i:=1 to numpoints do
     begin
        readfileline(predfile,buf);
	classval:=round(readnum1(predfile,false,ignoreall));
	pvalue:=readnum1(predfile,false,ignoreall);
	predarray[i].cat:=classval;
	if (classval=posval) then
		predarray[i].prob:=pvalue
	else
		predarray[i].prob:=1-pvalue;
     end;
  testsize:=numpoints;
  calcnumposandneg(numpos,numneg,numpoints,posval);
  bublesortpred(1,numpoints);
  numleft:=0;
  falseneg:=0;truepos:=numpos;trueneg:=0;falsepos:=numneg;

  while numleft<numpoints do
   begin
     getnextcutoff(numleft,numpos,numneg,posval);
     getstats(tempmodel,0,0,0,0,1);
     tempmodel^.acccut:=predarray[numleft].prob;
     comparemodel(bestmodel,tempmodel,accuracy,false);
   end;
  printstats(output,bestmodel);
  dispose(bestmodel);
  dispose(tempmodel);
end;
