(*-----------#pc-version-----------------------------------------------------*)
(*
implementation
uses sys;
*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "mon.h"
#include "util.h"
#include "sys.h"
#include "io.h"
#include "eqs.h"
#include "scal.h"
#include "rsutil.h"
#include "calctet.h"
#include "calctet1.h"
(*-----------#unix-version-----------------------------------------------------*)


(* June 7, 1989.  Makelisfile takes fname as parameter.  *)
(** Functions that bear checking - due to dependencies on other stuff -
1) setsign - uses adjaceny[i,j] = -1,  but our conventions for adjacency have changed. **)

(****************************)
(*   EQS.P   *)

(* functions devoted to writing an input file for the statistical package eqs 
   from the current model being analyzed in TETRAD *)
(**************************************)
(* Utilities *)

 (* this checks to see wether the user has purposely
    used a certain name for a latent common cause to 
    indicate that the two effects of the latent are to
    have correlated errors and the latent is to be ignored. *)
function correrrsrc(*vert: vertextype): boolean;*);
  begin
    correrrsrc:= (vernames[vert][1] = correrrchar);
  end;    

(* to fix the scale of each latent var L, an edge from
   L to one of its indicators should be fixed.  This
   procedure finds the appropriate edges to fix. *)
procedure findscalefixers(var adj: adjtype; var scalefixers: fixers);
var   lat,meas:integer;
      found:boolean;
  begin
    for lat:= numob+1 to max do
      begin
        found:= false;  meas:= 1;
	(* check to see if latent is a correlated error latent
	   a vertex whose name begins with a capital Z *)
        if not(correrrsrc(lat)) then
          while (not found) and (meas <= max) do
            begin
              if (adj^[lat,meas] <> notthere) and (meas <> lat) then
                begin
                  scalefixers[lat] := meas;
                  found:= true;
                end
              else meas:= meas+1;
            end;
      end;
  end;


(* this returns true if i and j are supposed to have 
   correlated errors *)
function correrrs(var adj: adjtype; i,j: vertextype): boolean;
 var src: vertextype;
     answer: boolean;
  begin
    if (i <> j) and (adj^[i,j] = 3) and (adj^[j,i] = 3) then correrrs := true;
    if (i = j) or correrrsrc(i) or 
    correrrsrc(j) then correrrs:= false
    else
     begin
      src:= numob+1; 
      answer:= false;
      while not(answer) and (src <= max) do
       begin
         answer:= (correrrsrc(src) and
                   (adj^[src,i] <> notthere) and
                   (adj^[src,j] <> notthere));
         src:= src+1;
       end;
      correrrs:= answer;
    end;
  end;               


procedure setsign(i,j: integer; var adjacency: adjtype; var parms: covarray);
(* makes sure that adjacency matrix and parameter matrix agree in sign *)
begin
    if ((adjacency^[i,j] = -1) and (parms^[i,j] >= 0)) or
       ((adjacency^[i,j] <> - 1) and (parms^[i,j] < 0)) then 
            parms^[i,j] := -1 * parms^[i,j];
end;

(* End of Utilities *)
(********************)

(******************************************)
(* Writing the EQS input files -  *)


procedure eqsinit(var scalefixers: fixers; var eqsnames: eqnamearray);
 var i,j: integer;
 begin
  for i:= 1 to maxvertex do scalefixers[i]:= 0;
  for i:= 1 to maxvertex do
    for j:= 1 to buflength do
      eqsnames[i][j]:= chr(0);
 end;

procedure geteqsnames(var eqsnames: eqnamearray);
  var i,fcount: integer;
  begin
    fcount:= 0;
    for i:= 1 to max do
      if i <= numob then
       getname(i,'V',eqsnames[i])
      else if not(correrrsrc(i)) then
        begin
           fcount:= fcount + 1;
           getname(fcount,'F',eqsnames[i]);
        end;
  end;


procedure writeqstitle(var comment: inbuf);
  begin
   twriteln(eqsinp,'/TITLE|');
   twrite(eqsinp,comment);
   twriteeoln(eqsinp);
  end;

procedure writspecs;
begin
   twriteln(eqsinp,'/SPECIFICATIONS|');
   twrite(eqsinp,'CAS = |');
   twriteint(eqsinp,tround(sampsize));
   twriteln(eqsinp,';|');
   twrite(eqsinp,'VAR = |');
   twriteint(eqsinp,numob);
   twriteln(eqsinp,';|');
   twriteeoln(eqsinp);
end;

procedure writlabls(var eqsnames: eqnamearray);
   var  i: vertextype;
   begin
     twriteln(eqsinp,'/LABELS|');

     for i:=1 to max do
       if not(correrrsrc(i)) then
         begin
           writebuf(eqsnames[i],eqsinp);
           twrite(eqsinp,' = |'); pshort(i, eqsinp); twrite(eqsinp,'; |');
           if (i > 5) and ((i mod 6) = 0) then twriteeoln(eqsinp);
         end;
     twriteeoln(eqsinp);
     twriteeoln(eqsinp);
   end;


procedure writequations(var adj: adjtype; var parm: covarray; 
                        var eqsnames: eqnamearray; var scalefixers: fixers);
   var  effect,cause: vertextype; fcount,charcount:integer;
        name: inbuf;
   begin
      charcount:=0; fcount:= 0;
      twriteln(eqsinp,'/EQUATIONS|');
      for effect:=1 to max do
       if (effect in graphvertices) and (not correrrsrc(effect)) then      
        begin
          writebuf(eqsnames[effect],eqsinp);
          twrite(eqsinp,' = |');
          charcount:=charcount+6;
          for cause:= 1 to max do
            if (cause <> effect) then
               if (not (correrrsrc(cause))) then
                 if (adj^[cause,effect] <> notthere) then
              begin
                if scalefixers[cause] = effect then
                   twrite(eqsinp,' 1.0   |')
                else 
		begin
		  tpwritereal(eqsinp,parm^[cause,effect],
                              decpoints+3,decpoints,false);
		  twrite(eqsinp,'*|');
		end;
                writebuf(eqsnames[cause],eqsinp);
                charcount:=charcount+8;
                if (charcount <= 50) then
                  begin
                    twrite(eqsinp,' + |');
                    charcount:=charcount+3;
                  end
                else
                  begin
                    twriteln(eqsinp,' + |');
                    twrite(eqsinp,'   |');
                    charcount:=3;
                  end;
              end;
          if effect <= numob then 
	     begin
		twrite(eqsinp,' |');
	        getname(effect,'E',name);
	     end
          else 
             begin
               fcount:= fcount+1;
	       twrite(eqsinp,' |');
               getname(fcount,'D',name);
             end;
          writebuf(name,eqsinp);
          twrite(eqsinp,';|');
          charcount:=0;
          twriteeoln(eqsinp);
        end; (* for effect = 1 to max*)
      twriteeoln(eqsinp);
   end;

procedure writvariances;
  var  dcount,effect: integer;
  begin
    twriteln(eqsinp,'/VARIANCES|');
    for effect := 1 to numob do
       if effect in graphvertices then
          begin
            twrite(eqsinp,'E|');    
            twriteint(eqsinp,effect);
            twriteln(eqsinp,' = 0.5*;|');
          end;
     for dcount:=(numob+1) to max do
       if (not correrrsrc(dcount)) and (dcount in graphvertices) then
          begin
            twrite(eqsinp,'D|');
            twriteint(eqsinp,dcount-numob);
            twriteln(eqsinp,' = 0.5*;|');
          end;
     twriteeoln(eqsinp);
  end;

procedure writcovariances(var adj: adjtype);
  var i,j: integer;
      first: boolean;
  begin  
    first:= true;    
    for i := 1 to (max-1) do
      for j:= i+1 to max do
         if (([i,j] <= graphvertices) and correrrs(adj,i,j)) then
           begin
             if first then begin twrite(eqsinp,'/COVARIANCES|');
				 twriteeoln(eqsinp); first:= false; end;
             if i <= numob then
		begin
	           twrite(eqsinp,'E|');
		   twriteint(eqsinp,i);
		   twrite(eqsinp,',|');
		end
             else 
		begin
	           twrite(eqsinp,'D|');
		   twriteint(eqsinp,i-numob);
		   twrite(eqsinp,',|');
		end;
             if j <= numob then
		begin
	           twrite(eqsinp,'E|');
		   twriteint(eqsinp,j);
		end
             else 
		begin
	           twrite(eqsinp,'D|');
		   twriteint(eqsinp,j-numob);
		end;
             twrite(eqsinp,' = 0.5*;|');
             twriteeoln(eqsinp);
           end; (* if correrrs *)
    if not first then twriteeoln(eqsinp);
  end; (* writcovariances *)

procedure writmatrix(var cov: covarray);
var count,countx:integer;
begin
      twriteln(eqsinp,'/MATRIX|');
      for count:=1 to numob do
        begin
          for countx:=1 to count do
            begin 
              if (countx > 9) and ((countx mod 10) = 0) then 
                 twriteeoln(eqsinp);
              twrite(eqsinp,' |');
	      tpwritereal(eqsinp,cov^[count,countx],10,decpoints,false);
            end;
          twriteeoln(eqsinp);
        end;
    end;

(* END OF Writing the EQS input files -  *)
(******************************************)

(**********************************)
(* Main routine for EQS writer *)

procedure makeeqsfile(* var adj: adjtype; var parm: covarray; 
                        var comment,fname: inbuf *);
   var scalefixers : fixers;
       eqsnames: eqnamearray;
   begin
      eqsinit(scalefixers, eqsnames);
      geteqsnames(eqsnames);
      writeqstitle(comment);
      writspecs;
      writlabls(eqsnames);
      findscalefixers(adj,scalefixers);
      writequations(adj, parm, eqsnames,scalefixers);
      writvariances;
      writcovariances(adj);
      writmatrix(cov);
      twriteln(eqsinp,'/END|');
      twriteeoln(eqsinp);
  end;

procedure readalphastring(var instring: inbuf;length:integer; var ok:boolean;err: errortype);
var c: char; notwarned,foundwhitespace: boolean; i: integer;
begin
   ok:=true;
   if length > buflength then length := buflength;
   for i := 1 to length do instring[i] := ' ';
   foundwhitespace := false;
   notwarned := true;
   i := 1;
   curpos:=1;
   while (not myeoln) and (not foundwhitespace) do
      begin
        readit(c);
	if i = 1 then
           if not alphabetic(c) then
	     begin
                myerror('Non-alphabetic character beginning name. |',err,true);
		ok:=false;
	     end
           else instring[i] := uppercase(c)
	else if (i <= length) then
          begin
           if (not alphanumeric(c)) and (not whitespace(c)) then
	     begin
	       myerror('Non-alphanumeric character in name.|     ',err,true);
		ok:=false;
	     end
           else if (not whitespace(c)) then instring[i] := uppercase(c)
           else foundwhitespace := true;
          end
        else 
            begin
               if notwarned and (not whitespace(c)) then
                   myerror('Truncating name to 8 characters.|',warning,true);
               notwarned := false;
               if whitespace(c) then foundwhitespace := true;
            end;
        if (not foundwhitespace) then i := i + 1;
      end;
      instring[i]:='|';
end;

procedure getalphastring(prompt:inbuf;var instring: inbuf);
var ok:boolean;
 begin
   ok:= false;
   while not(ok) do
    begin
      twrite(output,prompt);
      readttyline(buf);
      readalphastring(instring,8,ok,ignoreall);
    end;
 end;

procedure caliswritedata(cov:covarray;datasetname,comment:inbuf;eqsnames: eqnamearray);
   var  i,j: vertextype;
	charcount :integer;
   begin
	twrite(eqsinp,'DATA |'); twrite(eqsinp,datasetname); twriteln(eqsinp,'(TYPE=COV);|');
	twrite(eqsinp,'TITLE "|'); twrite(eqsinp,comment); twriteln(eqsinp,'";|');
	twrite(eqsinp,'_TYPE_ = ''COV''; INPUT _NAME_ $V1-V|');
	twriteint(eqsinp,numob);twriteln(eqsinp,';|');

	(* WRITE OUT THE LABELS *)
	twrite(eqsinp,'LABEL |');
	for i:=1 to max do
	       if not(correrrsrc(i)) then
	         begin
        	   writebuf(eqsnames[i],eqsinp);
	           twrite(eqsinp,' = ''|'); pshort(i, eqsinp); twrite(eqsinp,''' |');
	           if (i > 5) and ((i mod 6) = 0) then twriteeoln(eqsinp);
	         end;
	twriteln(eqsinp,';|');
	twriteln(eqsinp,'CARDS;|');

	(* WRITE OUT THE COVARIANCE MATRIX *)
	for i := 1 to numob do
	  begin
       	   writebuf(eqsnames[i],eqsinp);
	   twrite(eqsinp,'   |');
           charcount := 8;
		for j := 1 to numob do
		     begin
			if charcount > 70 then 
			begin 
			  twriteeoln(eqsinp);
			  charcount:=0;
			end;
			if j > i then 
			begin
				twrite(eqsinp,'    .      |');
				charcount:=charcount + 11;
			end
			else 
			   begin
				tpwritereal(eqsinp,cov^[i,j],10,decpoints,true);
				twrite(eqsinp,'  |');
				charcount:= charcount + 10;
			   end;
			end;
	   twriteeoln(eqsinp);
	  end;
	twriteln(eqsinp,';|'); twriteeoln(eqsinp);
   end;

procedure calislineqs(var adj: adjtype; var parm: covarray; 
                        var eqsnames: eqnamearray; var scalefixers: fixers);
   var  effect,cause: vertextype; fcount,charcount:integer;
	pcount:integer;
        name: inbuf;
	first : boolean;
   begin
	first := true;
	charcount:=0; fcount:= 0; pcount:=0;
	twriteln(eqsinp,'LINEQS|');

      for effect:=1 to max do
       if (effect in graphvertices) and (not correrrsrc(effect)) then      
        begin
	  if not first then twriteln(eqsinp,',|');
	  first := false;
          writebuf(eqsnames[effect],eqsinp);
          twrite(eqsinp,' = |');
          charcount:=charcount+6;
          for cause:= 1 to max do
            if (cause <> effect) then
               if (not (correrrsrc(cause))) then
                 if (adj^[cause,effect] <> notthere) then
              begin
                if scalefixers[cause] = effect then
                   twrite(eqsinp,'|')
                else 
		begin
		  pcount := pcount + 1;
		  getname(pcount,'X',name);
		  twrite(eqsinp,name);
		  twrite(eqsinp,' (|');
		  tpwritereal(eqsinp,parm^[cause,effect],10,decpoints,false);
		  twrite(eqsinp,') |');
		end;
                writebuf(eqsnames[cause],eqsinp);
                charcount:=charcount+8;
                if (charcount <= 50) then
                  begin
                    twrite(eqsinp,' + |');
                    charcount:=charcount+3;
                  end
                else
                  begin
                    twriteln(eqsinp,' + |');
                    twrite(eqsinp,'   |');
                    charcount:=3;
                  end;
              end;
          if effect <= numob then 
             getname(effect,'E',name)
          else 
             begin
               fcount:= fcount+1;
               getname(fcount,'D',name);
             end;
          writebuf(name,eqsinp);
          charcount:=0;
        end; (* for effect = 1 to max*)
      if not first then twriteln(eqsinp,';|');
      twriteeoln(eqsinp);
   end;

procedure calisstd;
  var  dcount,effect: integer;
	firstinrow:boolean;
	thecount:integer;
  begin
    thecount:=0;
    twriteln(eqsinp,'STD|');
    firstinrow := true;
    for effect := 1 to numob do
       if effect in graphvertices then
          begin
	    thecount:=thecount+1;
	    if not(firstinrow) then twriteln(eqsinp,',|')
		else firstinrow := false;
            twrite(eqsinp,'E|');
            twriteint(eqsinp,effect);
            twrite(eqsinp,' = THE|');
	    twriteint(eqsinp,thecount); 
	    twrite(eqsinp,' (0.5)|');
          end;
    for dcount:=(numob+1) to max do
       if (not correrrsrc(dcount)) and (dcount in graphvertices) then
          begin
	    thecount:=thecount+1;
	    if not(firstinrow) then twriteln(eqsinp,',|')
		else firstinrow := false;
            twrite(eqsinp,'D|');
            twriteint(eqsinp,dcount-numob);
            twrite(eqsinp,' = THE|');
	    twriteint(eqsinp,thecount); 
	    twrite(eqsinp,' (0.5)|');
          end;
     if not firstinrow then 
	begin
	  twriteln(eqsinp,';|');
	  twriteeoln(eqsinp);
	end;
  end;



procedure caliscov(adj:adjtype);
  var i,j: integer;
      first: boolean;
        name: inbuf;
	phicount:integer;
  begin
    phicount:=0;
    first:= true;    
    for i := 1 to (max-1) do
      for j:= i+1 to max do
         if (([i,j] <= graphvertices) and correrrs(adj,i,j)) then
           begin
	     phicount:=phicount + 1;
             if first then
		begin twriteln(eqsinp,'COV|'); first:= false; end
	     else  twriteln(eqsinp,', |');
             if i <= numob then
		begin
	           twrite(eqsinp,'E|');
		   twriteint(eqsinp,i);
		   twrite(eqsinp,' |');
		end
             else 
		begin
	           twrite(eqsinp,'D|');
		   twriteint(eqsinp,i-numob);
		   twrite(eqsinp,' |');
		end;
             if j <= numob then
		begin
	           twrite(eqsinp,'E|');
		   twriteint(eqsinp,j);
		end
             else 
		begin
	           twrite(eqsinp,'D|');
		   twriteint(eqsinp,j-numob);
		end;
             twrite(eqsinp,' = PHI|');
	     twriteint(eqsinp,phicount);
             twrite(eqsinp,' (0.5)|');
           end; (* if correrrs *)
    if not first then begin twriteln(eqsinp,';|'); twriteeoln(eqsinp); end;
  end; (* writcovariances *)


	
procedure caliswriteproc(cov:covarray;adj:adjtype;parm:covarray;
			 eqsnames:eqnamearray;scalefixers:fixers;
			 datasetname:inbuf;comment:inbuf);
   begin
	twrite(eqsinp,'PROC CALIS DATA=|');
	twrite(eqsinp,datasetname);
	twrite(eqsinp,' COV EDF=|');
	twriteint(eqsinp,tround(sampsize)-1);
	twriteln(eqsinp,';|');

	twrite(eqsinp,'TITLE2 "|');
	twrite(eqsinp,comment);
	twriteln(eqsinp,'";|');
	findscalefixers(adj,scalefixers);
	calislineqs(adj,parm,eqsnames,scalefixers);	(* WRITE OUT THE LINEAR EQUATIONS*)
	calisstd;			(* WRITE OUT THE STD section*)
	caliscov(adj);			(* WRITE OUT THE COV section*)

	twrite(eqsinp,'RUN;|');
   end;

(**********************************)
(* Main routine for CALIS writer  *)

procedure makecalisfile(* var adj: adjtype; var parm: covarray; 
                        var comment,fname: inbuf *);
   var scalefixers : fixers;
       eqsnames: eqnamearray;
       datasetname:inbuf;
   begin
      eqsinit(scalefixers, eqsnames);
      geteqsnames(eqsnames);
        datasetname:='TETDATA|';
(*	getalphastring('Name for SAS data set: |',datasetname);	*)
	caliswritedata(cov,datasetname,comment,eqsnames);
	caliswriteproc(cov,adj,parm,eqsnames,scalefixers,datasetname,comment);
      twriteeoln(eqsinp);
   end;



(*********************************************)
(* Support routine for LISREL/SIMPLIS writer *)

procedure lisrelwritecov(cov:covarray);
  var i,j:vertextype;
	cc : integer;
  begin
   (* WRITE OUT THE COVARIANCE MATRIX *)
   cc := 0;
   twriteln(eqsinp,'Covariance Matrix|');
   for i := 1 to numob do
     begin
      for j := 1 to i do
        begin
         if cc > 65 then begin twriteeoln(eqsinp);cc:=0; end;
           tpwritereal(eqsinp,cov^[i,j],10,decpoints,true);
           twrite(eqsinp,'  |'); cc:= cc+11;
        end;
      cc:=0;
      twriteeoln(eqsinp);
     end;
   twriteln(eqsinp,' |');
   twrite(eqsinp,'Sample Size |');
   twriteint(eqsinp,tround(sampsize));
   twriteln(eqsinp,' |');
  end;

function getksinum(cause: vertextype; 
                   var lisindep: vertexset): vertextype;
  var i,count: vertextype;
      found: boolean;
  begin
    i:= 1; found:= false; count:= 0;
    while (not found) and (i <= maxvertex) do
      begin
        if i in lisindep then 
          begin
            found:= (i = cause);
            count:= count+1;
          end;
        if i < maxvertex then i:= i+1
        else found := true;
      end;
    getksinum:= count;
  end;

function measureofksi(i: vertextype; var lisindep: vertexset): vertextype;
  var hits,v: vertextype;
      found: boolean;
  begin
    v:= 0; hits:= 0;
    repeat
      v:=v+1;
      if v in lisindep then hits:= hits+1;
    until i=hits;
    measureofksi:= v;
  end;
     
   (* Returns true if v has no indegree except for "Z" variables *)
function lisexog(v: vertextype; var adj: adjtype): boolean;
  var found: boolean;
      i: vertextype;
  begin
    found:= false;
    i:= 1;
    while (not found) and (i <= max) do
      begin
        if (adj^[i,v] <> notthere) and (not correrrsrc(i)) then
          found:= true;
        i:= i+1;
      end;
    lisexog:= (not found); 
  end; (* lisexog *)

procedure lisrelwriterelationships(cov:covarray;adj:adjtype;
				parm:covarray;var scalefixers:fixers;
                                lisdefault: boolean; 
                                lisindep: vertexset);
                      (* ksinum: vertextype; ksi: parray);  *)
   var  i,j,effect,cause: vertextype; charcount,n:integer;
        name: inbuf;
	first : boolean;
  begin
    charcount:=0; 
    twriteln(eqsinp,'Relationships |');
    findscalefixers(adj,scalefixers);
    for effect:=1 to max do
     if (effect in graphvertices) and (not correrrsrc(effect)) then      
      begin
        first:=true;
        for cause:= 1 to max do
          if (cause <> effect) then
            if (not (correrrsrc(cause))) then
              if (adj^[cause,effect] <> notthere) then
	       begin
		if first then
		 begin
	          twriteeoln(eqsinp);
	          writename(effect,eqsinp);
	          twrite(eqsinp,' = |');
        	  charcount:=charcount+6;
		  first:=false;
		 end;
                if (scalefixers[cause] = effect) then
			twrite(eqsinp,' 1*|')
                else 
                  if parm^[cause,effect] <> 0 then
                    begin
                      twrite(eqsinp,' (|');
   	    	      tpwritereal(eqsinp,parm^[cause,effect],
                                  decpoints+3,decpoints,false);
 		      twrite(eqsinp,')*|');
		    end;
		if not(cause in lisindep) or (not lisdefault)
                   or istheoretical(vernames[cause])
                  then writename(cause,eqsinp)
                else 
                  begin
                    twrite(eqsinp,'Ksi|');
                    twriteint(eqsinp,getksinum(cause,lisindep));
                    twrite(eqsinp,' |');
                  end;
                charcount:=charcount+8;
                if (charcount <= 50) then
                  begin
                    twrite(eqsinp,' |');
                    charcount:=charcount+3;
                  end
                else
                  begin
                    twriteln(eqsinp,' |');
                    twrite(eqsinp,'     |');
                    charcount:=3;
                  end;
              end;(*for cause:= 1 to max do*)
          charcount:=0;
        end; (* for effect = 1 to max*)
      if lisdefault then 
       begin
         for i:= 1 to sizeofset(lisindep) do
           begin
            twriteeoln(eqsinp);
	    writename(measureofksi(i,lisindep),eqsinp);
	    twrite(eqsinp,' = 1*Ksi|');
            twriteint(eqsinp,i);
           end;
         twriteeoln(eqsinp);
         for i:= 1 to sizeofset(lisindep) do
           begin
            twriteeoln(eqsinp);
            twrite(eqsinp,'Set the Error Variance of |');
            writename(measureofksi(i,lisindep),eqsinp);
	    twrite(eqsinp,' to 0|');
           end;
         twriteeoln(eqsinp);
    for i:= 1 to max-1 do
      for j:= i+1 to max do
        if (treks1^[i,j] = nil) and lisexog(i,adj)
           and lisexog(j, adj) and (not correrrsrc(i))
           and (not correrrsrc(j)) then 
          begin 
              twrite(eqsinp, 'Set the Covariances of |');
              if i <= numob then
                begin
                  twrite(eqsinp, ' Ksi|');
                  twriteint(eqsinp,getksinum(i,lisindep));
                end
              else writename(i,eqsinp); 
              twrite(eqsinp, ' and |');
              if j <= numob then
                begin
                  twrite(eqsinp, ' Ksi|');
                  twriteint(eqsinp,getksinum(j,lisindep));
                end
              else writename(j,eqsinp); 
              twrite(eqsinp, ' to 0|');
              twriteeoln(eqsinp);
            end;
       end; (* not lisdefault *)
      twriteeoln(eqsinp);
  end;

procedure lisrelcov(adj:adjtype);
  var i,j: integer;
      first: boolean;
        name: inbuf;
	phicount:integer;
  begin
    phicount:=0;
    first:= true;    
    for i := 1 to (max-1) do
      for j:= i+1 to max do
         if (([i,j] <= graphvertices) and correrrs(adj,i,j)) then
           begin
	     phicount:=phicount + 1;
		twrite(eqsinp,'Let the errors between |');
		writename(i,eqsinp);
		twrite(eqsinp,' and |');
		writename(j,eqsinp);
		twriteln(eqsinp,' correlate|');
	    end;
	twriteeoln(eqsinp);
  end; (* writcovariances *)

procedure lisrelwritelatents(var lisindep: vertexset; lisdefault:boolean);
  var  first : boolean;
       i,ksinum: vertextype;
       cc : integer;
  begin
   (* check if all latents are correrrs /and don't print out if is *)
    first := true; ksinum:= 0;
    cc := 0;
    if max > numob then 
     for i := numob + 1 to max do
      if (not correrrsrc(i)) then
       begin
	if first then
	  begin
            twrite(eqsinp,'Latent Variables |'); 
            first:=false; cc :=17; 
          end;
        if cc > 60 then 
          begin 
            twriteeoln(eqsinp); 
            cc:=0; 
          end;
	writename(i,eqsinp); 
        twrite(eqsinp,' |'); 
        cc:=cc+6;
       end;
    if lisdefault then 
     for i:= 1 to numob do
      if i in lisindep then 
        begin
          ksinum:= ksinum+1;    
          if first then 
            begin
              twriteeoln(eqsinp);
              twrite(eqsinp,'Latent Variables |'); 
              first:=false; cc :=17; 
            end;
          twrite(eqsinp,'Ksi|');
          twriteint(eqsinp,ksinum);
          twrite(eqsinp,' |');
        end;
    twriteeoln(eqsinp);
    twriteeoln(eqsinp);
  end;



(* This forms a sets lisindep which contains exogenous 
   measured variables that are independent of at least one 
   other exogenous variable *)
procedure getlisindep(var lisindep: vertexset; var foundindeppair: boolean;
                      var adj: adjtype);
    var i,j: vertextype;
  begin
    lisindep:= [];     
    foundindeppair := false;
    if ((not rantet) and (not ranpart)) then
      begin
	inittreks;
        calcpath(adj,paths1,1);
	calctrek(treks1,paths1,1);
      end;
    for i:= 1 to max-1 do
      for j:= i+1 to max do
        if (treks1^[i,j] = nil) and lisexog(i,adj)
           and lisexog(j, adj) and (not correrrsrc(i))
           and (not correrrsrc(j)) then 
          begin 
            if i <= numob then lisindep:= lisindep + [i];
            if j <= numob then lisindep:= lisindep + [j];
            foundindeppair := true;
          end;
   end; (* getlisindep *)        

(**************************************************)



(* Main routine for LISREL/SIMPLIS writer 
   The only way, in Simplis, to make exogenous vars
   independent is to make them identical to latents and to then 
   to specify that the correlations among these latents is 0.
   Thus ksinum keeps count of the number of non-latent but
   independent vars. *)
procedure makelisrelfile(* var adj: adjtype; var parm: covarray; 
                        var comment,fname: inbuf *);
   var scalefixers : fixers;
       eqsnames: eqnamearray;
       datasetname:inbuf;
       i:vertextype;
       cc : integer;
       ksinum: vertextype;
       ksi: parray;
       lisdefault: boolean;
       lisindep: vertexset;
       foundindeppair: boolean;
   begin
      eqsinit(scalefixers, eqsnames);
      twriteln(eqsinp,comment);
      twriteeoln(eqsinp);
      twrite(eqsinp,'Observed Variables |');
      cc:= 19;
      for i := 1 to numob do
	begin
	if cc > 60 then begin twriteeoln(eqsinp); cc:=0; end;
	  writename(i,eqsinp); twrite(eqsinp,' |'); cc := cc + 6;
 	end;
      twriteeoln(output);
      getlisindep(lisindep,foundindeppair,adj);
      if foundindeppair then 
        begin
          twrite(output,'Exogenous variables uncorrelated? [y]|');
          lisdefault:= compareto('y');
        end
      else lisdefault:= false;
      twriteeoln(eqsinp);
      lisrelwritecov(cov);
      twriteeoln(eqsinp);
      lisrelwritelatents(lisindep, lisdefault);
      lisrelwriterelationships(cov,adj,parm,scalefixers,lisdefault,lisindep);
    (*  lisrelcov(adj); *)
      twriteln(eqsinp,'End of Problem|');
      twriteeoln(eqsinp);
   end;


(************************)
(* new for LISREL *)

procedure makeonecalisfile;
 var comment, fname: inbuf;
     gotparams: char;
     changed: boolean;
  begin
    fname:= 'calis.in';
    fname[9] := chr(0);
    twriteeoln(output);
    twriteln(output,'Name of SAS/CALIS input file that |');
    getfile(eqsinp,'TETRAD II will create: |',outputf,
                              fname,changed);
    if (fname <> nullfilename) and (fname <> blankfilename) then
       begin
         getcomment(comment);
         makecalisfile(adjacency, parm, comment,fname);
	 flush(eqsinp);
       end
     else myerror('No output file.|',fatal,false);
  end;

procedure makeonelisrelfile;
 var comment, fname: inbuf;
     gotparams: char;
     changed: boolean;
  begin
    fname:= 'lisrel.in';
    fname[10] := chr(0);
    twriteeoln(output);
    twriteln(output,'Name of LISREL input file that |');
    getfile(eqsinp,'TETRAD II will create: |',outputf,
                              fname,changed);
    if (fname <> nullfilename) and (fname <> blankfilename) then
       begin
         getcomment(comment);
         makelisrelfile(adjacency, parm, comment,fname);
	 flush(eqsinp);
       end
     else myerror('No output file.|',fatal,false);
  end;

(************************)
(*EQS WRITER ************)
procedure makeoneeqsfile;
 var comment, fname: inbuf;
     gotparams: char;
     changed: boolean;
  begin
    fname:= 'eqs.in';
    fname[7] := chr(0);
    twriteeoln(output);
    twriteln(output,'Name of EQS input file that |');
    getfile(eqsinp,'TETRAD II will create: |',outputf,
                              fname,changed);
    if (fname <> nullfilename) and (fname <> blankfilename) then
       begin
         getcomment(comment);
         makeeqsfile(adjacency, parm, comment,fname);
	 flush(eqsinp);
       end
     else myerror('No output file.|',fatal,false);
  end;

procedure newmakeoneeqsfile(* var localadj: adjtype *);
 var comment, fname: inbuf;
     gotparams: char;
     changed: boolean;
  begin
    fname:= '/usr/user/peter/blub.in';
    fname[24] := chr(0);
    comment[1] := chr(0);
    rewritefile(eqsinp,fname);
    makeeqsfile(localadj, parm, comment,fname);
    flush(eqsinp);
  end;

(*-----------#pc-version-----------------------------------------------------*)
(*
end.
*)
(*-----------#pc-version-----------------------------------------------------*)