(*-----------#pc-version-----------------------------------------------------*)
(*
implementation
uses rsutil,util,io,calctet,calctet1,sys;
*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version---------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "mon.h"
#include "calctet.h"
#include "calctet1.h"
#include "io.h"
#include "sys.h"
#include "util.h"
#include "rsutil.h"
#include "eqs.h"
#include "cspecial.h"
#include "scal.h"
#include "undir.h"
(*-----------#unix-version---------------------------------------------------*)

(****************************)
(*   SCAL.NEW   *)

(* functions devoted to identifying the causal structure among latent
   variables in a linear multiple indicator model *)

(******************************)

(***************************************)
(* Utilities *)

procedure getsigcorr;
var i,j: vertextype; z,prob: real;
begin
  for i := 1 to numob - 1 do
    for j := i + 1 to numob do 
     if ([i,j] <= graphvertices) then
      begin
         z := (1/2) * (sqrt(sampsize - 3)) * ln((1 + abs(rho^[i,j])) / 
              (1 - abs(rho^[i,j])));
         getprob(prob,abs(z));
         insigcorr[i,j]:= (prob > signif);
         insigcorr[j,i]:= insigcorr[i,j];
      end;
end;

procedure printcluster(latvar: vertextype; clustervars: vertexset;
                       var outfile: text);
  var i: integer;
  begin
    twrite(outfile,'Indicators of |');
    pshort(latvar,outfile); twrite(outfile,': [ |');
    for i:= 1 to numob do
      if i in clustervars then begin pshort(i,outfile); twrite(outfile,' |'); end;
    twrite(outfile,'] |');
    twriteeoln(outfile);
  end;

 (* this returns true just in case  both sides of a 
    tetrad equation have insignificant correlations *)
function trivial(i,j,k,l,eqindex: integer): boolean;
   var side1,side2: boolean;
   begin
     case eqindex of 
      1: begin
           side1:= insigcorr[i,j] or insigcorr[k,l];
           side2:= insigcorr[i,k] or insigcorr[j,l];
         end;
      2: begin
           side1:= insigcorr[i,j] or insigcorr[k,l];
           side2:= insigcorr[i,l] or insigcorr[j,k];
         end;
      3: begin
           side1:= insigcorr[i,l] or insigcorr[j,k];
           side2:= insigcorr[i,k] or insigcorr[j,l];
         end;
     end;
     trivial:= (side1 and side2);
   end;

    (* convert pattern to graph *)
procedure patterntograph(var pattern,graph: adjtype);
  var i,j: vertextype;
  begin
    for i:= 1 to max-1 do
      for j:= i+1 to max do
        begin
          graph^[i,j]:= notthere; 
          graph^[j,i]:= notthere;
          if (pattern^[j,i] = outof2) then graph^[i,j]:= there
          else if (pattern^[i,j] = outof2) then graph^[j,i]:= there; 
        end;
  end;

(* End of Utilities *)
(***************************************)

(**************************)
(*  Initialize   *)

(* This checks to see if the graph file input to scales is a pure 
   measurement in which each latent has at least two indicators, 
   each indicator measures exactly one parent.  *)
function checkgivenclusters(origmeasures: modasclusters): boolean;
 label 999;
 var i,j,parent: vertextype;
     ok: boolean;
 begin
  tnewadj(pattern,fatal);   (* check if enough space *)
  if fatal in errors then goto 999;
  ok:= true;
  twriteeoln(output); 
  twriteln(output,'Reading initial measurement model.|');  twriteeoln(output); 
  for i:= 1 to numlatents do
    if (sizeofset(measures[i]) < 2) then
       begin
         twriteln(output,'Error in initial measurement model: |'); 
         pshort(numob+i,output); 
         twrite(output,' must have at least 2 indicators, but has only |');
 	 twriteint(output, sizeofset(measures[i]));twrite(output,'.|');
 	 twriteeoln(output);   twriteeoln(output); 
         ok:= false;
       end;
  for i:= 1 to numob do 
   if (i in graphvertices) then
    begin
      parent:= parentof(i, origmeasures);
      if parent > numlatents then
        begin
           twriteln(output,'Error in initial measurement model: |'); 
           pshort(i,output); twrite(output,' measures no latent.|'); 
 	   twriteeoln(output);   twriteeoln(output); 
           ok:= false;
        end
      else
        for j:= parent+1 to numlatents do
          if i in origmeasures[j] then
           begin
             twriteln(output,'Error in initial measurement model: |'); 
             pshort(i,output); twrite(output,' measures more than one latent: |'); 
             pshort(numob+parent,output); twrite(output,' and |');
 	     pshort(numob+j,output); 
 	     twriteeoln(output);   twriteeoln(output); 
             ok:= false;
           end;
    end;
  checkgivenclusters:= ok;
999:  end; (* checkgivenclusters *)
   
 (* this reports on the edges MIMBuild is ignoring in what it takes
    to be the initial measurement model *)
procedure reportignore(adjacency: adjtype);
 var i,j: vertextype;
 begin
   for i:= 1 to numob do
     for j:= 1 to numob do 
      if ([i,j] <= graphvertices) then
       if (i <> j) and (adjacency^[i,j] <> notthere) then
         begin
           twrite(output,' - Ignoring measured-measured edge: |');
           pshort(i,output); 
           twrite(output,' -> |'); 
           pshort(j, output); 
 	   twriteeoln(output);
         end;
   for i:= 1 to numob do
     for j:= numob+1 to max do
      if ([i,j] <= graphvertices) then
       if (adjacency^[i,j] <> notthere) then
         begin
           twrite(output,' - Ignoring measured-latent edge: |'); 
           pshort(i, output); 
           twrite(output,' -> |'); 
           pshort(j, output); 
           twriteeoln(output);
         end;
   for i:= numob+1 to max do
     for j:= numob+1 to max do
      if ([i,j] <= graphvertices) then
       if (i <> j) and (adjacency^[i,j] <> notthere) then
         begin
           twrite(output,' - Ignoring latent-latent edge: |'); 
           pshort(i, output); 
           twrite(output,' -> |'); 
           pshort(j, output); 
           twriteeoln(output);
         end;
     twriteeoln(output);
 end;

function scaleinit(var adjacency: adjtype): boolean;
 label 999;
 var i,j: vertextype;
     ok: boolean;
 begin
   yes:= 'y'; no:= 'n';
   sdebug:= false;
   printon:= true;
   getsigcorr;
   totty:= false;
   for i:= 1 to maxtheos do
     begin
       measures[i]:= [];
       for j:= 1 to maxtheos do
           treksep[i,j]:= [i,j];
     end;
   numlatents:= origmax - numob;
   for i:= 1 to numlatents do
    begin
     for j:= 1 to numob do
      if (j in graphvertices) then
      begin
        if adjacency^[numob+i,j] <> notthere then measures[i]:= measures[i] + [j];
        origmeasures[i]:= measures[i];
      end;
    end;
   ok:= checkgivenclusters(origmeasures);
   if fatal in errors then goto 999;
   if ok then reportignore(adjacency);
   (* testinit; *)
   scaleinit:= ok;
999:  end;

(*  End of Initialize *)
(***************************************)

(********************************)
(* Intra-Construct :  4x0s  *)

procedure get4x0criterion(indicators: vertexset; numtets:integer; 
                         var bonfail: integer; var criterion: real);
  var i,j,k,l,n: integer;
      prob,score: real;
  begin
    score:= 0.0;
    bonfail:= 0;
    for i:= 1 to numob - 3 do
     if i in indicators then
      for j:= i+1 to numob - 2 do
        if j in indicators then
          for k:= j+1 to numob - 1 do
            if k in indicators then
              for l:= k+1 to numob do
                if l in indicators then
                  for n:= 1 to 3 do
                     begin  
                       prob:= get1prob(i,j,k,l,n);
                       if prob < signif then 
                           score:= score - (weight * (1 - prob))
                       else score:= score + prob;
                       if (prob < (signif/numtets)) then bonfail:= bonfail+1;
                     end;
    criterion:= score/numtets;
  end;  


  (******* messages ******)

procedure message0(latent: vertextype; var outfile: text);
  begin
    twriteeoln(outfile); twriteeoln(outfile);
    pshort(latent, outfile); 
    twrite(outfile,' --  Original Status: |');
  end;

procedure message1(bonfail: integer; maxsofar: real; 
                   numtets: integer; var outfile: text);
begin
     if bonfail = 0 then twriteln(outfile,' Needs NO pruning.|');
     twrite(outfile,'  |');      
     tpwriteint(outfile,bonfail,4,false);  
     twrite(outfile,' of |');      
     tpwriteint(outfile,numtets,5,false);
     twrite(outfile,' tetrads fail the Bonferroni test.|');
     twriteeoln(outfile);
end;

procedure message2(drop: vertextype; numtets:integer; 
                   bonfail:integer;  var outfile: text);
begin
     twrite(outfile,'    w/o |');pshort(drop, outfile); twrite(outfile,': |');
     tpwriteint(outfile,bonfail,4,false);
     twrite(outfile,'out of |'); 
     tpwriteint(outfile,numtets,4,false);
     twrite(outfile,'tetrads fail the Bonferroni test.|');
end;



procedure message3(vartodrop: vertextype; var measures: vertexset;
                   bonfail,numtets: integer; latent: vertextype; var outfile: text);
  begin 
    twrite(outfile,' Dropped |'); 
    pshort(vartodrop,outfile); 
    twrite(outfile,'  Without it, |');
    tpwriteint(outfile,bonfail,4,false);
    twrite(outfile,' of |');
    tpwriteint(outfile,numtets,4,false);
    twrite(outfile,' tetrads fail the Bonferroni test.|');
    if (sizeofset(measures) < 5) and (bonfail > 0) then
       begin
          twriteeoln(outfile);
          twrite(outfile, ' No indicators of |'); pshort(latent,outfile);
          twrite(outfile,' left to drop|'); 
          twriteeoln(outfile);    
       end;
    twriteeoln(outfile);
  end; (* message 3 *)

procedure message4(maxsofar: real; bonfail,setsize: integer; var outfile: text);
  begin
    if setsize = 4 then
       begin
         if bonfail > 0 then twriteln(outfile,' Only 4 indicators, CANNOT BE PRUNED by 4x0.|')
         else twriteln(outfile,' Only 4 indicators, needs NO pruning.|');
         twrite(outfile,'     |'); 
         tpwriteint(outfile,bonfail,2,false);
         twrite(outfile,' of 3 tetrads fail the Bonferroni test.|');
         twriteeoln(outfile); 
       end
    else
     begin
       twrite(outfile,' Has LESS THAN 4 variables.|');
       twriteeoln(outfile);
     end;
  end; (* message4 *)
  (******* messages ******)

  (* this calculates a score for each subset of a factor's measures
     that doesn't include the variable we are considering dropping.
     We drop the indicator whose complement subset has the highest score.
     We terminate either when we run out of indicators to drop or
     there is some subset that passes the amazing bonferroni test. *)
procedure do4x0(latent: vertextype; var measures: vertexset);
  var drop,vartodrop: vertextype;
      criterion, maxsofar: real;
      setsize,numtets: integer;
      minbonfail,bonfail: integer;
      continue: boolean;
  begin
   continue:= true;
   setsize:= sizeofset(measures);
   message0(latent,ofilvar);
   if interactive then message0(latent,output);
   if setsize >= 4 then
     begin
       numtets:= 3 * choosenum(setsize,4);
       get4x0criterion(measures,numtets,bonfail,maxsofar);      
       minbonfail:= bonfail;
     end;
   if setsize > 4 then
      begin
       message1(bonfail, maxsofar,numtets,ofilvar);
       if interactive then message1(bonfail, maxsofar,numtets,output);
       if bonfail > 0 then   
       repeat
        if interactive then 
          begin
            twrite(output,'Continue? (y/n)  [y]  |'); 
            continue:= compareto('y');  writeln;
          end;
        if continue then
          begin
           vartodrop:= numob+1;
           setsize:= setsize-1;
           numtets:= 3 * choosenum(setsize,4);
           for drop:= 1 to numob do
             if drop in measures then 
                begin
                  get4x0criterion(measures - [drop],numtets,bonfail,criterion);
                  if sdebug then begin
                     message2(drop, numtets, bonfail,output);
                     message2(drop, numtets, bonfail,ofilvar); end;
                  if criterion > maxsofar then
                    begin
                      minbonfail:= bonfail;
                      maxsofar:= criterion;
                      vartodrop:= drop;
                    end
                  else if (bonfail <= minbonfail) and (criterion > maxsofar) then
                    begin
                      vartodrop:= drop;
                      maxsofar:= criterion;
                    end;
                end; (* the subset without drop *)
             measures:= measures - [vartodrop];
             message3(vartodrop, measures, minbonfail, numtets, latent, ofilvar);
             if interactive then 
                 message3(vartodrop, measures, minbonfail, numtets, latent,output);
           end; (* continue *)
       until ((not continue) or (sizeofset(measures) < 5) or (minbonfail = 0));
     end 
   else if (setsize <= 4) then 
           begin  
              message4(maxsofar,bonfail,setsize, ofilvar); 
              if interactive then message4(maxsofar,bonfail,setsize, output);
           end;
  end;  (* do4x0 *)

function num4x0s(var measures: modasclusters): integer;
 var num,i: integer;
 begin
  num:= 0;
  for i:= 1 to numlatents do
    if (sizeofset(measures[i]) >= 4) then 
     num:= num + (3 * choosenum(sizeofset(measures[i]), 4));
  num4x0s:= num;
 end;
 
  (* this returns the number of 4x0 equations that fail by the bonferroni
     significance level. *)
function bonferoninum4x0(var measures: modasclusters; 
                         var numtets: integer): integer;
  var i,j,k,l,parent: vertextype;
      tetnum,bonfail: integer;
      cutoff: real;
  begin
    numtets:= num4x0s(measures);
    bonfail:= 0;
    if numtets > 0 then
     begin
      cutoff:= (signif/numtets);
      for parent:= 1 to numlatents do
       if (sizeofset(measures[parent]) >= 4) then
         for i:=1 to numob-3 do 
          if i in measures[parent] then
           for j:= i+1 to numob-2 do
             if (j in measures[parent]) then
              for k:= j+1 to numob-1 do
               if k in measures[parent] then
                for l:= k+1 to numob do 
                 if (l in measures[parent]) then
                   for tetnum:= 1 to 3 do
                    if (get1prob(i,j,k,l,tetnum) < cutoff) then 
                      bonfail:= bonfail+1;
     end;
    bonferoninum4x0:= bonfail;
  end;

(* End of 4x0 *)
(**************************************)

(**************************************)
(* Cross-construct *)


(******************)
(**  2x2  *)

function num2x2s(var measures: modasclusters):integer;
  var count: integer;
      parent1, parent2: vertextype;
  begin
    count:= 0;
    for parent1:= 1 to numlatents-1 do
        for parent2:= parent1+1 to numlatents do 
          if (sizeofset(measures[parent1]) >= 2) or
             (sizeofset(measures[parent2]) >= 2) then
                count:= count + (choosenum(sizeofset(measures[parent1]),2) * 
                                 choosenum(sizeofset(measures[parent2]),2));
    num2x2s:= count;
  end;

  (* this returns the number of equations that fail by the bonferroni
     significance level. *)
function bonferoninum2x2(var measures: modasclusters; var numtets: integer): integer;
  var i,j,k,l,parent1,parent2: vertextype;
      bonfail: integer;
      cutoff: real;
  begin
    numtets:= num2x2s(measures);
    bonfail:= 0;
    if numtets > 0 then
     begin
      cutoff:= (signif/numtets);
      for parent1:= 1 to numlatents-1 do
       for parent2:= parent1+1 to numlatents do
        if (sizeofset(measures[parent1]) >= 2) or
          (sizeofset(measures[parent2]) >= 2) then
         for i:=1 to numob-1 do 
           if i in measures[parent1] then
            for j:= i+1 to numob do
             if (j in measures[parent1]) then
              for k:= 1 to numob-1 do
               if k in measures[parent2] then
                for l:= k+1 to numob do 
                 if (l in measures[parent2]) then
                   if (get1prob(i,j,k,l,3) < cutoff) then bonfail:= bonfail+1;
     end; 
    bonferoninum2x2:= bonfail;
  end;

  (* this calculates the number of 2x2 tetrad equations that involve a 
     given variable, whose parent is passed in as a parameter. *)
function num2x2swithv(parent: vertextype; var measures: modasclusters):integer;
  var i,count,numcousins: integer;
  begin
   numcousins:= sizeofset(measures[parent]) - 1;
   count:= 0;
   for i:= 1 to numlatents do
     if (i <> parent) then
       count:= count + choosenum(sizeofset(measures[i]),2);
   num2x2swithv:= numcousins * count;
  end;
        
  (* this is only called if the number of vars in the
     thisvarparent cluster is > 2 *)        
function get2x2criterion(thisvar,thisvarparent: vertextype;
                         var measures: modasclusters): real;
  var numtets: integer;
      cousin,k,l: vertextype;
      prob: real;
      cousinset: vertexset;
  begin
    cousinset:= measures[thisvarparent] - [thisvar];
    score:= 0.0;
    numtets:= num2x2swithv(thisvarparent,measures);
    for cousin:= 1 to numob do
     if cousin in cousinset then
      for k:= 1 to (numob - 1) do
       if (k in graphvertices) then          
        if not (k in measures[thisvarparent]) then
          for l:= k+1 to numob do 
           if (l in measures[parentof(k,measures)]) then          
            begin
              prob:= get1prob(thisvar,cousin,k,l,3);
              if prob < signif 
              then score:= score - (weight * (1 - prob))
              else score:= score + prob;
            end; 
    get2x2criterion:= score/numtets;
  end;  
(* End of 2x2 *)
(*****************)

(******************)
(**  3x1  *)

function num3x1s(var measures: modasclusters):integer;
  var count: integer;
      parent1, parent2: vertextype;
  begin
    count:= 0;
    for parent1:= 1 to numlatents do
     if sizeofset(measures[parent1]) >= 3 then
      for parent2:= 1 to numlatents do
        if parent2 <> parent1 then
          count:= count + (choosenum(sizeofset(measures[parent1]),3) * 
	                  sizeofset(measures[parent2]));
    num3x1s:= 3*count;
  end;

function bonferoninum3x1(var measures: modasclusters; 
                         var numtets: integer): integer;
  var i,j,k,l,parent1,parent2,n: vertextype;
      bonfail: integer;
      cutoff: real;
  begin
    numtets:= num3x1s(measures);
    bonfail:= 0;
    if numtets > 0 then
     begin
       bonfail:= 0;
       cutoff:= (signif/numtets);
       for parent1:= 1 to numlatents do
        if sizeofset(measures[parent1]) >= 3 then
         for parent2:= 1 to numlatents do
          if parent1 <> parent2 then
            for i:=1 to numob-2 do 
              if i in measures[parent1] then
                for j:= i+1 to numob-1 do
                  if (j in measures[parent1]) then
                    for k:= j+1 to numob do
                      if (k in measures[parent1]) then
                        for l:= 1 to numob do 
                          if (l in measures[parent2]) then
                            for n:= 1 to 3 do
                              if (get1prob(i,j,k,l,n) < cutoff) then 
                                      bonfail:= bonfail+1;
     end; (* if numtets *)
    bonferoninum3x1:= bonfail;
  end; (* bonferoninum3x1 *)


function get3x1criterion(thisvar,thisvarparent: vertextype; 
                         var measures: modasclusters): real;
  var numtets: integer;
      i,j,k,n,latent: integer;
      prob,score: real;
      cousinset,allmeasures: vertexset;
  begin
    cousinset:= measures[thisvarparent] - [thisvar];
    score:= 0.0;
    numtets:= 0;
    allmeasures:= [];
    for i:= 1 to numlatents do
      allmeasures := allmeasures + measures[i];
    (* all the 1 by 3s s.t. thisvar is the singelton *)
    for latent:= 1 to numlatents do
      if (latent <> thisvarparent) and (sizeofset(measures[latent]) >= 3) then
        for i:= 1 to numob-2 do
          if i in measures[latent] then
            for j:= i+1 to numob-1 do
              if j in measures[latent] then 
                for k:= j+1 to numob do
                  if k in measures[latent] then 
                    for n:= 1 to 3 do
                      begin
                         numtets:= numtets+1;
                         prob:= get1prob(thisvar,i,j,k,n);
                         if prob < signif 
                         then score:= score - (weight * (1 - prob))
                         else score:= score + prob;
                      end; 
    (* all 3 x 1s s.t. thisvar is in the 3some *)
    if (sizeofset(measures[thisvarparent]) >= 3) then
     for i:= 1 to numob-1 do 
      if i in cousinset then
        for j:= i+1 to numob do
         if j in cousinset then
           for k:= 1 to numob do
             if (not (k in measures[thisvarparent])) and (k in allmeasures) then
                for n:= 1 to 3 do
                  begin
                     numtets:= numtets+1;
                     prob:= get1prob(thisvar,i,j,k,n);
                     if prob < signif 
                     then score:= score - (weight * (1 - prob))
                     else score:= score + prob;
                  end; 
    get3x1criterion:= score/numtets;
  end;  


(* End of 3x1 *)
(************************************)

(** messages for crossconstruct *)

procedure crossm0(crosstype: integer);
  begin
    twriteeoln(ofilvar);    
    if crosstype = 3 then twriteln(ofilvar,'3x1 Phase |')
    else twriteln(ofilvar,'2x2 Phase |');
    twriteln(ofilvar,'--------|');    
  end;

procedure crossm1(numit, bonfaildrop,numtets: integer; var outfile: text);
  begin
    twriteeoln(outfile);    
    twrite(outfile,'Iteration |');
    tpwriteint(outfile,numit,2,false);
    twrite(outfile,'  |');
    tpwriteint(outfile,bonfaildrop,3,false);
    twrite(outfile,' of |');
    tpwriteint(outfile,numtets,5,false);
    twrite(outfile,' tetrads fail the Bonferroni test.|');
    twriteeoln(outfile);    
  end;

procedure crossmess1(numit, bonfaildrop,numtets: integer);
  begin
    if interactive then crossm1(numit,bonfaildrop,numtets,output);
    crossm1(numit,bonfaildrop,numtets,ofilvar);
  end;

procedure crossm4(var vartodrop, fromlatent: vertextype; 
                  var bonfaildrop,numtets: integer;
                  var minsofar: real; var outfile: text);
  begin
    twrite(outfile,' Dropped |'); 
    pshort(vartodrop,outfile);
    twrite(outfile,'  Without it, |');
    if bonfaildrop >= 0 then
      begin
        tpwriteint(outfile,bonfaildrop,4,false);
        twrite(outfile,' of |');
        tpwriteint(outfile,numtets,4,false);
        twrite(outfile,' tetrads fail the Bonferroni test.|');
      end
    else twrite(outfile,' no 3x1 tetrads left to test.|');
    twriteeoln(outfile);
  end; (* crossm4 *)

procedure mess4(var vartodrop, fromlatent: vertextype; 
                var bonfaildrop,numtets: integer;
                var minsofar: real);
  begin
    crossm4(vartodrop,fromlatent,bonfaildrop,numtets,minsofar,ofilvar);
    if interactive then 
         crossm4(vartodrop,fromlatent,bonfaildrop,numtets,minsofar,output);
  end;

procedure message5;
   begin 
    if interactive then twriteln(output,' No foursomes left|');
    twriteln(ofilvar,' No foursomes left|'); 
   end;


function continuequery: boolean;
  begin
    twrite(output, 'Continue? (y/n)  [y]  |'); 
    continuequery:= compareto('y'); 
  end; 

        (* we update the vartodrop if either 
           1) its criterion is within 1% of the previous criterion and its
              cluster is strictly bigger, or
           2) its criterion is strictly smaller than the previous best. *)
procedure dothisvar(thisvar,latent: vertextype; var vartodrop,fromlatent: vertextype;
                    var measures: modasclusters; var criterion,minsofar: real;
                    crosstype: integer);
  begin
    if (sizeofset(measures[latent]) > crosstype) then 
      begin
       if crosstype = 2 then criterion:= get2x2criterion(thisvar,latent,measures)
       else criterion:= get3x1criterion(thisvar,latent,measures);
       if ((criterion < (minsofar + (0.01*minsofar))) and
           (sizeofset(measures[latent]) > sizeofset(measures[fromlatent]))) or
          (criterion < minsofar)  then
         begin
           minsofar:= criterion;
           vartodrop:= thisvar;
           fromlatent:= latent;
         end;
      end;
  end; (* dothisvar *)

(* This goes through all variables v in measures, and for each calculates
   a criterion that involves all 3x1 tetrads v is involved in,
   It drops the v that has the lowest criterion and terminates when the 
   subset of 3x1 equations that remain pass a bonferroni test. *)
procedure docross(var measures: modasclusters; crosstype: integer);
  var thisvar,vartodrop,latent,fromlatent: vertextype;
      criterion, minsofar: real;
      bonfaildrop,numtets,numit: integer;
      foursomesleft, continue: boolean;
  begin
    crossm0(crosstype);
    numit:=1; continue:= true; foursomesleft:= true;
    if crosstype = 2 then bonfaildrop:= bonferoninum2x2(measures,numtets)
    else bonfaildrop:= bonferoninum3x1(measures,numtets);
    while continue and (bonfaildrop > 0) and foursomesleft do
     begin 
       crossmess1(numit,bonfaildrop,numtets); 
       if interactive then continue:= continuequery;
       if continue then
         begin
           minsofar:= 999.0;
           vartodrop:= numob+1;
           fromlatent:= 1;
           for latent:= 1 to numlatents do
            for thisvar:= 1 to numob do
             if (thisvar in measures[latent]) then
                  dothisvar(thisvar,latent,vartodrop,fromlatent,
                            measures,criterion,minsofar,crosstype);
           (* finished all vars once*)
           if vartodrop <> (numob+1) then 
             begin
               measures[fromlatent]:= measures[fromlatent] - [vartodrop];
               if crosstype=2 then 
                     bonfaildrop:= bonferoninum2x2(measures,numtets)
               else bonfaildrop:= bonferoninum3x1(measures,numtets);
               mess4(vartodrop,fromlatent,bonfaildrop,numtets,minsofar);
             end  (* vartodrop *)
           else foursomesleft:= false;
           numit:= numit+1;
         end; (* continue *)
     end; (* while *)
     if (not foursomesleft) then message5;
     if interactive then twriteeoln(output);
  end; (* docross *)

(** End of crosscontruct  *)
(********************************)

(**********************************)
(** 0-order independence **)

procedure latmess1(var lat1,lat2,numindep,numcorr:integer; var outfile: text);
  begin
    twriteeoln(outfile);
    pshort(numob+lat1,outfile);
    twrite(outfile,', |');
    pshort(numob+lat2, outfile);
    twrite(outfile,': numind: |');
    tpwriteint(outfile,numindep,4,false);
    twrite(outfile,'   numcorr: |');
    tpwriteint(outfile,numcorr,4,false);
    twriteeoln(outfile);
  end;

(*
function latpairindep(lat1,lat2: integer; var measures: modasclusters): boolean;
   var i,j,numcorr,numindep: integer;
   begin
     numcorr:= 0; numindep:= 0;
     for i:= 1 to numob do
      if i in measures[lat1] then
        for j:= 1 to numob do
          if (i <> j) and (j in measures[lat2]) 
          then if insigcorr[i,j] 
                 then numindep:= numindep + 1
                 else numcorr:= numcorr + 1;
     latpairindep:= numindep >= numcorr;
   end;
*)

function latpairindep(lat1,lat2: integer; var measures: modasclusters): boolean;
   var i,j,numeq,bonfail: integer; z,prob: real; 
   begin
     bonfail := 0;
     numeq := 0;
     for i:= 1 to numob do
      if i in measures[lat1] then
        for j:= 1 to numob do
          if (i <> j) and (j in measures[lat2]) then numeq := numeq + 1;
     for i:= 1 to numob do
      if i in measures[lat1] then
        for j:= 1 to numob do
          if (i <> j) and (j in measures[lat2]) then 
            begin
               z := (1/2) * (sqrt(sampsize - 3)) * ln((1 + abs(rho^[i,j])) / 
                    (1 - abs(rho^[i,j])));
               getprob(prob,abs(z));
               if prob <= (signif/numeq) then bonfail := bonfail + 1;
            end;
       latpairindep := (bonfail = 0);
   end;

(****** indep messages ****)
procedure indepmessage1(var outfile: text);

  begin
    twriteeoln(outfile);
    twriteln(outfile,'--------------------------------------------------|');
    twriteln(outfile,'Statistical conclusions about the structural model|');
    twriteeoln(outfile);
    twriteln(outfile,'Edge             (Partial)|');  
    twriteln(outfile,'Removed          Correlation|');
    twriteln(outfile,'-------          -----------|');
  end;

procedure indepmessage2(i,j,k: vertextype; var outfile: text; order: integer);
  var n,totalchars: integer; 
  begin
   totalchars:=0;
   rswritename(numob+i,outfile,totalchars);
   twrite(outfile,' -- |'); totalchars:= totalchars+4;
   rswritename(numob+j,outfile,totalchars);
   for n:= 1 to (17 - totalchars) do twrite(outfile,' |');   
   totalchars:= 17;
   twrite(outfile,'rho(|'); totalchars:= totalchars+4;
   rswritename(numob+i,outfile,totalchars);
   twrite(outfile,' |'); totalchars:= totalchars+1;
   rswritename(numob+j,outfile,totalchars);
   if (order > 0) then 
     begin
       twrite(outfile,' . |'); 
       totalchars:= totalchars+3;
       rswritename(numob+k,outfile,totalchars);
     end;
   twrite(outfile,') = 0|'); 
   twriteeoln(outfile);
  end;

(****** indep messages ****)

procedure indeplats(var treksep: trekseptype; var measures: modasclusters);
 var i,j: vertextype;
  begin
     indepmessage1(ofilvar);
     for i:= 1 to numlatents-1 do
      for j:= i+1 to numlatents do
       if latpairindep(i,j,measures) then 
         begin
          treksep[i,j]:= [];
          treksep[j,i]:= [];
          indepmessage2(i,j,j,ofilvar,0);
        end;
  end;
 
(** End of 0-0rder trek-separation *)
(***********************************)

(***********************************)
(* 1st order trek-separation *)

function getnum1x2x1tets(end1,middle,end2: vertextype; 
                         var measures: modasclusters): integer;
  begin
    getnum1x2x1tets:= sizeofset(measures[end1]) * sizeofset(measures[end2]) *
                      choosenum(sizeofset(measures[middle]),2);
  end;

procedure classify(i,j,k,l,end1,middle,end2: vertextype; measures: modasclusters; 
                   var a,b,c,d: vertextype; var found: boolean);
begin
         if (([i,j] <= measures[middle]) and (k in measures[end1]) and (l in measures[end2]) or
             ([i,j] <= measures[middle]) and (k in measures[end2]) and (l in measures[end1]))
         then begin found:= true; a:= i; b:=j; c:= k; d:=l; end
         else 
         if (([i,k] <= measures[middle]) and (j in measures[end1]) and (l in measures[end2]) or
             ([i,k] <= measures[middle]) and (j in measures[end2]) and (l in measures[end1]))
         then begin found:= true; a:= i; b:=k; c:= j; d:=l; end
         else 
         if (([i,l] <= measures[middle]) and (k in measures[end1]) and (j in measures[end2]) or
             ([i,l] <= measures[middle]) and (k in measures[end2]) and (j in measures[end1]))
         then begin found:= true; a:= i; b:=l; c:= j; d:=k; end
         else 
         if (([j,k] <= measures[middle]) and (i in measures[end1]) and (l in measures[end2]) or
             ([j,k] <= measures[middle]) and (i in measures[end2]) and (l in measures[end1]))
         then begin found:= true; a:= j; b:=k; c:= i; d:=l; end
         else 
         if (([j,l] <= measures[middle]) and (k in measures[end1]) and (i in measures[end2]) or
             ([j,l] <= measures[middle]) and (k in measures[end2]) and (i in measures[end1]))
         then begin found:= true; a:= j; b:=l; c:= i; d:=k; end
         else 
         if (([k,l] <= measures[middle]) and (i in measures[end1]) and (j in measures[end2]) or
             ([k,l] <= measures[middle]) and (i in measures[end2]) and (j in measures[end1]))
         then begin found:= true; a:= k; b:=l; c:= i; d:=j; end;
 end;

procedure sepmess1(end1, middle, end2: vertextype);
  begin
    twriteeoln(output);
    pshort(numob+end1, output);
    twrite(output,' |');
    pshort(numob+end2, output);
    twrite(output,' by |');
    pshort(numob+middle, output);
    twriteeoln(output);
  end;


procedure sepmess2(a,b,c,d: vertextype; p1,p2: real);
  begin
    twriteeoln(output);
    pshort(a, output);
    twrite(output,' |');
    pshort(b, output);
    twrite(output,' * |');
    pshort(c, output);
    twrite(output,' |');
    pshort(d, output);
    twrite(output,' - |');
    pshort(a, output);
    twrite(output,' |');
    pshort(c, output);
    twrite(output,' * |');
    pshort(b, output);
    twrite(output,' |');
    pshort(d, output);
    twrite(output,' = |');
    tpwritereal(output,p1,3,2,false);
    twriteeoln(output);
    pshort(a, output);
    twrite(output,' |');
    pshort(b, output);
    twrite(output,' * |');
    pshort(c, output);
    twrite(output,' |');
    pshort(d, output);
    twrite(output,' - |');
    pshort(a, output);
    twrite(output,' |');
    pshort(d, output);
    twrite(output,' * |');
    pshort(b, output);
    twrite(output,' |');
    pshort(c, output);
    twrite(output,' = |');
    tpwritereal(output,p2,3,2,false);
  end;

procedure sepmess3(bonsucc,numtets: integer; score: real);
  begin
    twriteeoln(output);
    twrite(output,'Bonsucc = |');
    tpwriteint(output,bonsucc,3,false);
    twrite(output,'  Numtets = |');
    tpwriteint(output,numtets,3,false);
    twrite(output,'   Score = |');
    tpwritereal(output,score,3,2,false);
    twriteeoln(output);
  end;

(*********************)

procedure separated(end1, middle, end2: vertextype; var measures: modasclusters;
                   var bonsucc,numtets: integer; var score: real);
  var i,j,k,l,a,b,c,d: vertextype;
      allmeasures: vertexset;
      p1,p2: real;
      found: boolean;
  begin
    if sdebug then sepmess1(end1,middle,end2);
    numtets := 0;
    score:= 0.0;
    bonsucc:= 0;
    allmeasures:= [];
    for i:= 1 to numlatents do
      if i in [end1,middle,end2] then 
        allmeasures:= allmeasures + measures[i];
    for i:= 1 to numob-3 do if (i in allmeasures) then
     for j:= i+1 to numob-2 do if (j in allmeasures) then
      for k:= j+1 to numob-1 do if (k in allmeasures) then
       for l:= k+1 to numob do if (l in allmeasures) then 
        begin
         found:= false;
         classify(i,j,k,l,end1,middle,end2,measures,a,b,c,d,found);
         if found then 
            begin
             if (not trivial(a,b,c,d,1)) then numtets := numtets + 1;
             if (not trivial(a,b,c,d,2)) then numtets := numtets + 1;
            end;
        end;
    for i:= 1 to numob-3 do if (i in allmeasures) then
     for j:= i+1 to numob-2 do if (j in allmeasures) then
      for k:= j+1 to numob-1 do if (k in allmeasures) then
       for l:= k+1 to numob do if (l in allmeasures) then 
        begin
             found := false;
             classify(i,j,k,l,end1,middle,end2,measures,a,b,c,d,found);
             if found then
                begin
                   p1:= 0; p2:= 0; 
                   p1:= get1prob(a,b,c,d,1);
                   p2:= get1prob(a,b,c,d,2);
                   if sdebug then sepmess2(a,b,c,d,p1,p2);
                   if ((p1 > signif) and (not trivial(a,b,c,d,1))) then 
                      score:= score + p1
                   else score:= (score - (weight * (1 - p1)));
                   if ((p2 > signif) and (not trivial(a,b,c,d,2))) then 
                      score:= score + p2
                   else score:= (score - (weight * (1 - p2)));
                   if (p1 > (signif/numtets)) and (not trivial(a,b,c,d,1)) then 
                      bonsucc := bonsucc+1;
                   if (p2 > (signif/numtets)) and (not trivial(a,b,c,d,2)) then 
                      bonsucc := bonsucc+1;
                end;
       end;
  end;  (* separated *)


procedure do1x2x1(var treksep: trekseptype; var measures: modasclusters);
  var bonsucc,numtets: integer;
      i,j,sep: vertextype;
      score: real;
      found: boolean;
  begin
   for i:= 1 to numlatents-1 do
     for j:= i+1 to numlatents do
       if (treksep[i,j] = [i,j]) then (* if i and j are still adjacent *)
         begin
           sep:= 1; 
           found:= false;         
           while (not found) and (sep <= numlatents) do
             begin
               if ((sep <> i) and (sep <> j) and 
                   (not (treksep[i,sep] = []) or (treksep[j,sep] = []))) then
                 begin             
                   separated(i,sep,j,measures,bonsucc,numtets,score);
                   if sdebug then sepmess3(bonsucc,numtets,score);
                   if (bonsucc = numtets) then  
                      begin
                        treksep[i,j]:= treksep[i,j] + [sep];
                        treksep[j,i]:= treksep[j,i] + [sep];
                        indepmessage2(i,j,sep,ofilvar,1);
                        found:= true;
                      end;
                 end; (* if sep <> i *)
               sep:= sep+1;
             end; (* while *)
         end; (* if treksep *)
     twriteeoln(ofilvar);
  end; (* do1x2x1 *)


(* End of  1st order trek-separation *)
(*********************************************)


(*********************************)
(* Marking Uncertain Adjacencies *)


 (* this returns true if all the paths between i and j have the same 
    back1,x,forw1 subpath. Inter is the vertexset that occurs on all
    undirected paths between i and j. *)
function commonsubpaths(var ijpaths: udpathsptr; var inter: vertexset): boolean;
 var x, back1, forw1,a: vertextype;
     triple: vertexset;
     curpath: udpathsptr;
     found,stillgood: boolean;
 begin
   found:= false;
   x:= numob;
   while (not found) and (x < (numob + numlatents)) do
    begin
      stillgood:= true;
      x:= x+1;
      if x in inter then
        begin 
          triple:= [];
          curpath:= ijpaths;
          while (stillgood and (not found)) do
             begin
               if (sizeofset(curpath^.vertices) > 2) then
                  begin
                    a:= 1;
                    repeat a:= a+1; until (curpath^.patharray[a] = x);
                    back1:= curpath^.patharray[a-1];
                    forw1:= curpath^.patharray[a+1];
                    if (triple = []) then triple:= [back1,x,forw1]
                    else if (triple <> [back1,x,forw1]) then stillgood:= false;
                  end;
               curpath:= curpath^.next;
               found:= (stillgood and (curpath = nil));
             end; (* while *)
        end; (* x in inter *)
    end; (* while *)
   commonsubpaths:= found;
 end; (* commonsubpaths *)


function iscollider(a,x,b: vertextype; var pattern: adjtype): boolean;
  begin
    iscollider:= (pattern^[a,x] = into2) and (pattern^[b,x] = into2);
  end;


   {  x is a definite noncollider between a and b if either
      a *-* x -> b, or a <- x *-* b, or 
      a and b are not adjacent and not a *-> x <-* b    }
function isdefnoncollider(a,x,b: vertextype; var pattern: adjtype): boolean;
  begin
    isdefnoncollider:= ((pattern^[a,x] = outof2) or 
                        (pattern^[b,x] = outof2) or 
                        ((pattern^[a,b] = notthere) and
                         (not iscollider(a,x,b,pattern))));
  end;

                        
    (* this checks if any vertex x on all the undirected paths between i and j is
       a definite non-collider on each such path. Inter is the vertices that are
       on all the undirected paths between i and j *)
function commondefnoncollider(var ijpaths: udpathsptr; var inter: vertexset): boolean;
 var x, back1, forw1, a: vertextype;
     curpath: udpathsptr;
     found,stillgood: boolean;
 begin
   found:= false;
   x:= numob;
   while (not found) and (x < (numob + numlatents)) do
    begin
      stillgood:= true;
      x:= x+1;
      if x in inter then
        begin 
          curpath:= ijpaths;
          while (stillgood and (not found)) do
             begin
               if (sizeofset(curpath^.vertices) > 2) then
                  begin
                    a:= 1;
                    repeat a:= a+1; until (curpath^.patharray[a] = x);
                    back1:= curpath^.patharray[a-1];
                    forw1:= curpath^.patharray[a+1];
                    stillgood:= isdefnoncollider(back1,x,forw1,pattern);
                  end;
               curpath:= curpath^.next;
               found:= (stillgood and (curpath = nil));
             end; (* while *)
        end; (* x in inter *)
    end; (* while *)
   commondefnoncollider:= found;
 end; (* commondefnoncollider *)


function collideroneverypath(var ijpaths: udpathsptr; var pattern: adjtype): boolean;
 var x, a: vertextype;
     curpath: udpathsptr;
     found,stillgood: boolean;
 begin
   stillgood:= true;
   curpath:= ijpaths;
   while (stillgood and (curpath <> nil)) do
     begin
       if (sizeofset(curpath^.vertices) > 2) then
          begin
            found:= false;
            a:= 2;
            while ((not found) and (a+1 <= curpath^.length)) do
              begin
                 found:= iscollider(curpath^.patharray[a-1],
                                  curpath^.patharray[a],
                                  curpath^.patharray[a+1],
                                  pattern);
                 a:= a+1;
              end;
            stillgood:= found;           
          end; (* if *)
       if stillgood then curpath:= curpath^.next; 
     end; (* while *)
   collideroneverypath:= stillgood;
 end; (* collideroneverypath *)


  (*  checks an adjacency in the pattern output by mimbuild to see if
     ?? should be attached.  Any of 4 conditions suffice to eliminate ??s.
     1) If there are no paths besides the adjacecny connecting x and y in 
     the pattern.  2) If every alternative path has the same subpath a,b,c.
     3) If there is some definite non-collider z on every path, or
     4) If there is a collider on every path. *) 
procedure checkpairuncertain(i,j: vertextype; var pattern:adjtype; 
                             var uncertains: uncertaintype;
                             var udpathmatrix: udpathstype);
   var  inter: vertexset;
        a: integer;
        altpaths: boolean;
        curpos: udpathsptr;
   begin
     curpos:=  udpathmatrix^[i,j];
     (* initialize inter to all the vertices that might occur on
        on paths between i and j *)
     inter:= [];
     for a:= numob+1 to (numob+numlatents) do
       if not(a in [i,j]) then inter:= inter + [a];
     (* check if there is a path between i and j besides the adjacency, 
        and let inter hold vertices that occur on every such path *)
     altpaths:= false;
     while curpos <> nil do
        begin
          if ((curpos^.vertices - [i,j]) <> []) then 
             begin
               altpaths:= true;
               inter := inter * (curpos^.vertices - [i,j]);
             end;
          curpos := curpos^.next;
        end;
     if altpaths then
       begin
         if inter <> [] then
           if commonsubpaths(udpathmatrix^[i,j], inter) then 
              uncertains[i-numob,j-numob]:= false
           else if commondefnoncollider(udpathmatrix^[i,j], inter) then
                     uncertains[i-numob,j-numob]:= false;
         if (uncertains[i-numob,j-numob] and
             collideroneverypath(udpathmatrix^[i,j],pattern)) then 
                  uncertains[i-numob,j-numob]:= false;                        
       end
     else uncertains[i-numob,j-numob]:= false;   (* no paths but the adjacency *)
   end; (* checkpairuncertain *)


   (* checks adjacencies output by mimbuild for ??s. To do so it needs
      to convert the pattern output to an undirected graph, which is
      done by first converting it to a directed graph and then calculating
      all the paths. *)
procedure markuncertain(var pattern: adjtype; var uncertains: uncertaintype);

label 999;
var udpathmatrix: udpathstype;
    i,j: vertextype;
begin
  tnewudpaths(udpathmatrix,fatal); (* check for enough space *)
  if fatal in errors then goto 999;
  for i:= 1 to maxtheos do
    for j:= 1 to maxtheos do uncertains[i,j]:= true;
  udcalcpath(pattern, udpathmatrix, 1);
  for i:= numob+1 to (numob + numlatents - 1) do
   for j:= i+1 to (numob + numlatents) do
     if (pattern^[i,j] <> notthere) then
       checkpairuncertain(i,j,pattern,uncertains,udpathmatrix);
  dispose(udpathmatrix);
999: end; (* markuncertains *)
       

(* Marking Uncertain Adjacencies *)
(*********************************)

(*******************************************************)
(*  Orienting *)

  (* for each triple i,j,k, if i-j and j-k, i.e. treksep[i,j] = [i,j]
     etc, and not i-k, then unless j separates i and k, j must be a
     collider.  otherwise either i and k are independent, in which case
     j is a collider, or some vertex not equal to j separates i and k,
     in which case there can't be a trek between i-k through j, or else
     it would take 2nd order info to separate i and k.  *)
procedure findcolliders(treksep: trekseptype; var pattern: adjtype);    
   var i,j,k: vertextype;
   begin
   for i:= 1 to numlatents-2 do
      for j:= i+1 to numlatents-1 do 
         for k:= j+1 to numlatents do 
            if (pattern^[numob+i,numob+j] <> notthere) and
               (pattern^[numob+j,numob+k] <> notthere) and
               (pattern^[numob+i,numob+k] =  notthere) and
               (not ((pattern^[numob+i,numob+j] = into2) and 
                     (pattern^[numob+k,numob+j] = into2)))
               and (not (j in treksep[i,k]))
            then
               begin
                 pattern^[numob+i,numob+j]:= into2;
                 pattern^[numob+k,numob+j]:= into2;
               end
            else
            if (pattern^[numob+i,numob+j] <> notthere) and
               (pattern^[numob+i,numob+k] <> notthere) and
               (pattern^[numob+j,numob+k] =  notthere) and
               (not ((pattern^[numob+j,numob+i] = into2) and 
                     (pattern^[numob+k,numob+i] = into2))) 
                     and (not (i in treksep[j,k])) then
               begin
                 pattern^[numob+j,numob+i]:= into2;
                 pattern^[numob+k,numob+i]:= into2;
               end
            else 
            if (pattern^[numob+j,numob+k] <> notthere) and
               (pattern^[numob+i,numob+k] <> notthere) and
               (pattern^[numob+i,numob+j] =  notthere) and
               (not ((pattern^[numob+j,numob+k] = into2) and 
                     (pattern^[numob+i,numob+j] = into2))) 
                and (not (k in treksep[i,j])) then
               begin
                 pattern^[numob+j,numob+k]:= into2;
                 pattern^[numob+i,numob+k]:= into2;
               end;
    end; (* findcolliders *)


procedure avoidcycles(v1,v2: vertextype; var pattern: adjtype; var nothingnew: boolean);
   label 999;
   var i,j: vertextype;
   graph: adjtype;
   paths: pathstype;
   begin
    tnewadj(graph,fatal);  (* check for enough space *)
    tnewpaths(paths,fatal);  (* check for enough space *)
    if fatal in errors then goto 999;
    patterntograph(pattern,graph);
    calcpath(graph,paths,1);
    if (paths^[v1,v2] <> nil) then 
       begin
         pattern^[v1,v2]:= into2; 
         pattern^[v2,v1]:= outof2;
         nothingnew:= false; 
       end;
     dispose(paths);
     dispose(graph);
999:   end; (* avoidcycles *)

procedure awayfrom(treksep: trekseptype; var pattern: adjtype);    
  label 999;
  var i,j,k: vertextype;
      nothingnew: boolean;
  begin
    repeat
      nothingnew:= true;
      for i:= 1 to numlatents do 
       for j:= 1 to numlatents do if (i <> j) then
         begin
           for k:= 1 to numlatents do 
             if (j <> k) and (i <> k) and 
                (pattern^[numob+i,numob+j] <> notthere) and 
                (pattern^[numob+j,numob+k] <> notthere) and 
                (pattern^[numob+i,numob+k] = notthere)
             then
                if (pattern^[numob+i,numob+j] = into2) and 
                   (pattern^[numob+k,numob+j] <> into2) and
                   (pattern^[numob+k,numob+j] <> outof2) then
                   begin 
                     pattern^[numob+j,numob+k]:= into2; 
                     pattern^[numob+k,numob+j]:= outof2;
                     nothingnew:= false; 
                   end
                else if (pattern^[numob+i,numob+j] <> into2) and 
                        (pattern^[numob+k,numob+j] = into2) and
                        (pattern^[numob+i,numob+j] <> outof2) then
                       begin
                         pattern^[numob+j,numob+i]:= into2; 
                         pattern^[numob+i,numob+j]:= outof2;
                         nothingnew:= false; 
                       end;
           (* if (pattern^[numob+i,numob+j] = unknown) and 
              (pattern^[numob+j,numob+i] = unknown) then 
                 avoidcycles(numob+i,numob+j,pattern,nothingnew); *)
           if fatal in errors then goto 999;
         end;             
    until nothingnew;
999:  end; (* awayfrom *)

procedure orient(treksep: trekseptype; var pattern: adjtype); 
  label 999;
  var i,j: vertextype;
      uncertain: uncertaintype;
      done,printqmarks: boolean;
  begin
    (* initialize *)
    for i:= 1 to max-1 do
      for j:= i+1 to max do 
        begin
          pattern^[i,j]:= notthere;
          pattern^[j,i]:= notthere;
        end;
     for i:= 1 to numlatents-1 do
      for j:= i+1 to numlatents do 
        if treksep[i,j] = [i,j] then 
          begin
            pattern^[numob+i,numob+j]:= unknown;
            pattern^[numob+j,numob+i]:= unknown;
          end;
     findcolliders(treksep,pattern);
     awayfrom(treksep,pattern);
     if fatal in errors then goto 999;
     twrite(output, 'Mark uncertain adjacencies? (y/n)  [n]  |'); 
     printqmarks:= not compareto('n');
     if printqmarks then markuncertain(pattern, uncertain);
     if fatal in errors then goto 999;
     printmimout(pattern,ofilvar,uncertain,printqmarks);
999:  end; (* orient *)

(* End of  Orienting *)
(*******************************************************)

(** messages for the main control procedure **)
procedure puremessage1(var outfile: text);
  begin
   twriteeoln(outfile);
   twriteln(outfile,'             PURIFY:|');
   twriteln(outfile,'A Program to Find Unidimensional Measurement Models|'); 
   twriteeoln(outfile);
  end;

procedure puremess1;
  begin
     puremessage1(output); 
     puremessage1(ofilvar);
  end;

procedure mainmessage1(var outfile: text);
  begin
   twriteeoln(outfile);
   twriteln(outfile,'              MIMBUILD:|');
   twriteln(outfile,'A Program to Build Linear Latent Variable Models|'); 
   twriteeoln(outfile);
  end;

procedure mainmess1;
  begin
     mainmessage1(output); 
     mainmessage1(ofilvar);
  end;

procedure mainmessage2(var outfile: text);
  var i: vertextype;
  begin
    twriteln(outfile,'Initially Specified Measurement Model|');
    twriteeoln(outfile);
    for i:= 1 to numlatents do printcluster(numob+i,measures[i],outfile);
  end;


procedure mainmessage4x0(var outfile: text);
  begin
    twriteeoln(outfile);
    twriteln(outfile,'INTRA-CONSTRUCT PHASE.|');
    twriteln(outfile,'----------------------|');
  end;

procedure mainmess4x0;
  begin
    if interactive then mainmessage4x0(output);
    mainmessage4x0(ofilvar);
  end;


procedure mainmessage4(var outfile: text);
  begin
    twriteeoln(outfile);
    twriteln(outfile,'CROSS-CONSTRUCT PHASE.|');
    twriteln(outfile,'----------------------|');
  end;

procedure mainmess4;
  begin
    if interactive then mainmessage4(output);
     mainmessage4(ofilvar);
  end;

procedure mainmessage5(var outfile: text; bonfail,numtets: integer);
  var i,j: vertextype;
  begin
     twriteeoln(outfile);
     twriteln(outfile,'--------------------------------------------------------------|');
     twriteln(outfile,'Output Measurement Model|');
     twriteeoln(outfile);
     for i:= 1 to numlatents do printcluster(numob+i,measures[i],outfile);
     twriteeoln(outfile);
     if bonfail = 0 then 
        begin
          twriteln(outfile,'ALL tetrad equations implied by this measurement model|');
          twriteln(outfile,'pass the Bonferroni test.|');
        end
     else if (bonfail > 1) then
        begin
          twriteint(outfile,bonfail);
          twrite(outfile,' out of |');
          twriteint(outfile,numtets);
          twrite(outfile, ' tetrad equations implied by this measurement model|');
          twriteeoln(outfile);
          twriteln(outfile,'fail the Bonferroni test.|');
        end
          else begin
           twriteln(outfile,'1 tetrad equation implied by this measurement model fails|');
           twriteeoln(outfile);
           twriteln(outfile,'the Bonferroni test.|');
              end;
     twriteln(outfile,'--------------------------------------------------|');
     twriteln(outfile,'}|');
     twriteeoln(outfile);
     twriteln(outfile,'/graph|');
     for i:= 1 to numlatents do
       for j:= 1 to numob do
         if j in measures[i] then
           begin
              pshort(numob+i,outfile);
              twrite(outfile,' |');
              pshort(j, outfile);
              twriteeoln(outfile);
           end;
     twriteeoln(outfile);
  end;

(** end of messages for the main control procedure **)

(********************************)
(* Main Control for scales *)


procedure makepure(* var adjacency: adjtype *);
 label 999;
 var i,bonfail,numtets1,numtets2,numtets3: integer;
 begin
   puremess1;
   twrite(output,'Run PURIFY interactively (i), or in batch (b) ? [b]  |');
   interactive:= not (compareto('b'));
   if scaleinit(adjacency) then
     begin
       if fatal in errors then goto 999;
       mainmessage2(ofilvar);
       mainmessage4x0(ofilvar);
       for i:= 1 to numlatents do 
           do4x0(numob+i,measures[i]);
       mainmess4;
       docross(measures,3);
       (* docross(measures,2); *)
       bonfail:= (* bonferoninum2x2(measures,numtets1) + *)
                 bonferoninum3x1(measures,numtets2) +
                 bonferoninum4x0(measures,numtets3);
       mainmessage5(ofilvar, bonfail,(* numtets1+ *) numtets2+numtets3);
       flush(ofilvar);
     end (* if scaleinit *) 
   else twriteln(output,'Please correct these errors before you run PURIFY.|');
999:  end;


procedure buildlv(* var adjacency: adjtype *);
 label 999;
 var i: integer;
     uncertain: uncertaintype; 
 begin
   mainmess1;
   if scaleinit(adjacency) then
     begin
       if fatal in errors then goto 999;
       indeplats(treksep,measures);
       do1x2x1(treksep,measures);
       orient(treksep,pattern);
       if fatal in errors then goto 999;
       flush(ofilvar);
     end (* if scaleinit *) 
   else twriteln(output,'Please correct these errors before you run MIMbuild again.|');
   dispose(pattern);
999:  end;

(* End of Main Control for scales *)
(**************************************)

(*-----------#pc-version-----------------------------------------------------*)
(*end.*)
(*-----------#pc-version-----------------------------------------------------*)