(*-----------#pc-version-----------------------------------------------------*)
(*
implementation
uses init,mon,readco,makemod,calctet,calctet1,eqs,rsutil,io,up,part,sys,scal;
*)
(*-----------#pc-version-----------------------------------------------------*)


(*-----------#unix-version---------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "util.h"
#include "io.h"
#include "calctet.h"
#include "calctet1.h"
#include "sys.h"
#include "eqs.h"
#include "part.h"
#include "scal.h"
#include "mon.h"
#include "mon1.h"
#include "rsutil.h"
#include "up.h"
#include "init.h"
#include "readco.h"
#include "makemod.h"
#include "cross.h"
#include "newcalc.h"
#include "indep1.h"
#include "matrices.h"
#include "testbuildm.h"
#include "gib.h"
#include "thomas.h"

(*-----------#unix-version---------------------------------------------------*)



function tround(*num:real):integer*);
begin
   if num > tmaxint then
   begin
	myerror('Number too large, expected an Integer.|',fatal,false);
	tround:= 0;
   end
   else if num < tminint then
   begin
	myerror('Number too small, expected an Integer.|',fatal,false);
	tround:= 0;
   end
   else tround:= round(num);
end;

procedure doexit;
var closed: integer;
begin
 wantexit := true;
 if fileexists[outputf] then
       flush(ofilvar);
end;


procedure myerror(* message: inbuf; err: errortype; showpos: boolean *);
(* for error messages which should end with a carriage return *)
begin
   errormess(message,err,showpos);
   twriteeoln(output);   twriteeoln(output);
end;

function listposition (* (vertex: vertexnametype; insert: boolean;
		      var vertexnames: namearray; 
                      var wasinserted: boolean): vertextype; *);
var i,pos: integer;

begin
   wasinserted := false;
   i := 0;
   repeat
        i := i + 1;
   until (vertexnames[i] = vertex) or (i = maxvertex) or 
         (vertexnames[i][1] = ' ');
   if (vertexnames[i] = vertex) then 
       begin
           listposition := i;
           if insert then wasinserted := false;
       end
   else if insert then
     begin
       if (vertexnames[i][1] = ' ') then
	 begin
	    vertexnames[i] := vertex;
	    listposition := i;
            wasinserted := true;
	 end
       else if (vertexnames[maxvertex][1] <> ' ') then
	 begin
           (* so error message wont mention lineno or file name *)
           lineno := 0;
           myerror('Too many variables.|                     ',
                    ignoreall,false);
         end
      end
   else 
     begin
       errormess('Variable not found:  |',ignoreall,false);
       writevername(vertex,output);twriteeoln(output);
       listposition := 0;
     end;
end;
          
procedure disposepathslist(* var paths: pathsptr; depth: integer *);
(* this procedure disposes of a list of paths. *)
label 999;
var temp: pathsptr;

begin (* disposepathslist *)
   if paths <> nil then
	 begin
	    temp := paths^.next;
            while (paths <> nil) do
	     if ((depth <= paths^.level) or (paths^.level = 0)) then
	       begin
		  dispose(paths);
		  paths := temp;
		  if paths <> nil then temp := paths^.next;
	       end
             else goto 999;
	 end; (* if *)
999: end; (* disposepathlist *)

procedure disposetreklist(* var treks: treksptr; depth: integer *);
(* this procedure disposes of a list of trekpairs. *)
label 999;
var temp: treksptr;  
begin (* disposetreklist *)
    if treks <> nil then
 	 begin
 	    temp := treks^.next;
 	    while (treks <> nil) do
 	       begin
                 if (depth <= treks^.level) or (treks^.level = 0) then
    		      begin
 		         dispose(treks);
 		         treks := temp;
 		         if treks <> nil then temp := treks^.next;
		      end
		  else goto 999;
              end; (* if *)
         end;
999: end; (* disposetreklist *)

procedure disposeindeplist(* var indep1: indepptr *);
var temp,curpos: indepptr;
begin
   curpos := indep1;
   repeat
      if curpos <> nil then
        begin
            temp := curpos^.fwd;
            dispose(curpos);
            curpos := temp;
        end;
     until curpos = nil;
end;

procedure initinput(* id: commandset; covchange: boolean *); 
var i,j: vertextype;
    changedcomm,changedfor,changedtemp,changedcov,changedgraph,changedindep,
    changedbayes,changedraw,changedcells,changedev,changedlinmod: boolean;
begin
   isacycle := false;
   changedfor := false;
   changedcomm := false;
   changedtemp := false;
   changedbayes := havereadbayes;
   changedlinmod := havereadlinearmodel;
   changedcov := havereadcov;
   changedgraph := havereadgraph;
   changedraw := havereadraw;
   changedcells := havereadcells;
   changedev := hadevidence;
   changedindep := havereadindep;
   numedges := 0;
   ranpart := false; rantet := false; 
   curpos := 1;
   graphvertices := [];
   numob := 0;
   numtheo := 0;
   max := 0;
   initev;
   tfilename := blankfilename;
   ffilename := blankfilename;
   fcfilename := blankfilename;
   gfilename := blankfilename;

(* if not wipe out data then covariance is in id *)
   if (not (covariance in id)) then dfilename := blankfilename;  
   for i:=1 to maxvertex do
     begin
        if temptiers[i] <> [] then changedtemp := true;
        temptiers[i] := [];
        for j:=1 to maxvertex do 
             begin
               adjacency^[i,j] := notthere;
               if forbiddenadj^[i,j] <> notthere then 
                  changedfor := true;
               forbiddenadj^[i,j] := notthere;
               if commonfor^[i,j] <> notthere then changedcomm := true;
               commonfor^[i,j] := notthere;
               parm^[i,j]:= 0.0;
             end;
     end;
   disposeindeplist(indep1);
   errors := [];
   if covchange then
     for i := 1 to maxvertex do 
       begin
         fillexogd(i,2,0,1);
         if theoretical[i] then
           begin
              vernames[i] := blankname;
              if origmax > 0 then origmax := origmax - 1;
              if max > 0 then max := max - 1;
              theoretical[i] := false;
           end;
         vernames[i] := blankname;
         rho^[i,j] := 0;
         cov^[i,j] := 0;
         covvertices := [];
         sampsize := 0;
       end;
  if not firsttime then 
       begin
         if changedfor then twriteln(output,'Erasing previous Forbidden Edges|');
         if changedcomm then twriteln(output,'Erasing previous Forbidden Common Causes|');
         if changedtemp then twriteln(output,'Erasing previous Temporal Order|');
         if changedev then twriteln(output,'Erasing previous Evidence|');
         if havereadgraph and changedgraph and (not (graph in id)) then
             twriteln(output,'Erasing previous Graph|');
         if havereadcov and changedcov and (not (covariance in id)) then
             twriteln(output,'Erasing previous Covariance Matrix|');
         if havereadbayes and changedbayes and (not (showprogress in id)) then twriteln(output,'Erasing Bayes Network|');
         if havereadlinearmodel and changedlinmod then twriteln(output,'Erasing Linear Model|');
         if havereadcells and changedcells and (not (cellcounts in id)) then
            twriteln(output,'Erasing previous Cell Data|');
         if havereadraw and changedraw and (not (raw in id)) then
            twriteln(output,'Erasing previous Raw Data|'); 
         if havereadindep and changedindep and (not (independence in id)) then
            twriteln(output,'Erasing previous Independence Data|');      
       end;
   if covchange then
      begin
         havereadcov := false;
         havereadraw := false;
         havereadcells := false;
         havereadindep := false;
      end;
   havereadgraph := false;
   havereadbayes := false;
   havereadlinearmodel := false;
   firsttime := false;
end;  { procedure }

procedure insertcommand(cname: commandstring; id: commands; place: integer);
var i: integer;
begin
   for i := 1 to commandlength do
       commandarray[place].cstring[i] := cname[i];
   commandarray[place].cid := id;
end;


function findcommand(id: commands): integer;
var i: 1..numcommands; found: boolean;
begin
   i := 1;
   found := false;
   while (i <= numcommands) and not found do
      if commandarray[i].cid = id then 
         begin
            found := true;
            findcommand := i;
         end
      else i := i + 1;
end;    
     
procedure getnecessary(* id: commands *);
var changed: boolean; i: integer; notallthere: boolean;
begin		
 notallthere := false;
 if (not aaabort) then
  begin
    if (fileforout in commandneed[id]) then
     begin
      if ((ofilename = blankfilename) or (ofilename = nullfilename)) then
          begin
            getfile(ofilvar,'Output file: |                          ',outputf,
                    ofilename,changed);
            if (ofilename = blankfilename) or (ofilename = nullfilename) then
               begin
                   myerror('Output file name not specified.|          ',fatal,false);
                   fileexists[outputf] := false;
               end
             else fileexists[outputf] := true;
          end;
     end;
    if (lineardata in commandneed[id]) and ((not havereadcov) or (discrete) or (indepdata)) then
           begin
              writecname(commandarray[findcommand(id)].cstring,output);
              twriteln(output,' requires a covariance matrix.|');
              notallthere := true;
           end;
    if (agraph in commandneed[id]) and (not havereadgraph) then
           begin
              writecname(commandarray[findcommand(id)].cstring,output);
              twriteln(output,' requires a graph.|');
              notallthere := true;
           end;        
    if (model in commandneed[id]) and (not (havereadlinearmodel or havereadbayes)) then
           begin
              writecname(commandarray[findcommand(id)].cstring,output);
              twriteln(output,' requires either a linear model or a Bayesian network.|');
              notallthere := true;
           end;        
    if (data in commandneed[id]) and (not (lineardata in commandneed[id])) and
         (not (havereadcov or havereadraw or havereadcells or havereadindep)) then
           begin
              writecname(commandarray[findcommand(id)].cstring,output);
              twriteln(output,' requires data.|');
              notallthere := true;
           end; 
    if (bayesnet in commandneed[id]) and (not havereadbayes) then
         begin
            writecname(commandarray[findcommand(id)].cstring,output);
            twriteln(output,' requires a Bayes Network.|');
            notallthere := true;
         end;
    if (discretedata in commandneed[id]) and (not (havereadcells or havereadraw) or
       (not discrete)) then
         begin
            writecname(commandarray[findcommand(id)].cstring,output);
            twriteln(output,' requires raw discrete data.|');
            notallthere := true;
         end;
  end;
 if notallthere then errors := errors + [fatal];
 if (not aaabort) and (not (id in [monte,statwriter,gibbs,lrt,
                                   readcon,makemodel,addval,
                                   removeval,removeallval,split,cross])) then
         begin
            outputheading;
            givestatus(ofilvar,false,id);
         end;
end;  { procedure }

procedure inittreks;
var i,j: integer;
begin
  for i := 1 to maxvertex do
    for j := 1 to maxvertex do
     begin
       disposepathslist(paths1^[i,j],0);
       if (i <= j) then disposetreklist(treks1^[i,j],0);
       paths1^[i,j] := nil;
       treks1^[i,j] := nil;
     end;
end;

(* check to see if a correrr latent (a capital Z variable)
   is an effect of another correrr latent *)
function impropercorrerr(adj:adjtype):boolean;
label 999;
var lat1,lat2:vertextype;
begin
    for lat1:= numob+1 to max-1 do
    if correrrsrc(lat1) then
    for lat2:= lat1+1 to max do
    if correrrsrc(lat2) and (adj^[lat1,lat2] <> notthere)
    then
      begin
      impropercorrerr:=true;
      goto 999;
      end;
    impropercorrerr:=false;
999:
end;

procedure dostatwriter;
label 999;
  var maxnum: integer;
      statfiletypes: numberset;
  begin
    if impropercorrerr(adjacency) then 
    begin 
       myerror('Improper graph file: can not correlate errors of Z variables.|', fatal, false);
       goto 999;
    end;
    maxnum:= 3;
    statfiletypes:= [];
    repeat 
      twriteeoln(output);
      twriteln(output,'1 = EQS Input File|');
      twriteln(output,'2 = LISREL Input File|');
      twriteln(output,'3 = SAS\CALIS Input File|');
      twriteeoln(output);
      twriteln(output,'List the types of input files you want below.|');
      twriteln(output,'Separate each number by a space, and|');
      twrite(output,'use no delimiters. List =  |');
      getnumbers(statfiletypes, maxnum);
      if statfiletypes = [] then 
         twriteln(output,'Please enter a number, or Cntrl-g to exit. |');
    until (aaabort or (statfiletypes <> []));
    if 1 in statfiletypes then makeoneeqsfile;
    if 2 in statfiletypes then makeonelisrelfile;
    if 3 in statfiletypes then makeonecalisfile;
999:
  end; (* dostatwriter *)

(************************)

procedure dochangegraph;
begin
   if not havereadgraph then
       myerror('No graph.|             ', ignoreall,false)
   else
     begin
        twriteln(output,'Graph: |');
	writegraph(adjacency,output,false,true);
	twriteln(output,'Edges to be deleted. Hit <CR> when done.|'); twriteeoln(output);
        if (currentin = inp) and (max > 1) then
          begin
             twriteln(output,'Example of input: |');
             writename(1,output);
             twrite(output,' |');
             writename(2,output);
             twriteeoln(output);
          end;
	readgraph(true,false,input,adjacency,vernames,false);
	if not myeoln then twriteln(output,buf); twriteeoln(output);
	twriteln(output,'Edges to be added. Hit <CR> when done.|'); twriteeoln(output);
	readgraph(true,true,input,adjacency,vernames,false); twriteeoln(output);
        twriteln(output,'Graph: |');
	writegraph(adjacency,output,false,true);
	rantet := false; ranpart := false;
     end;
end;

procedure getsignificance;
begin
   signif := getnumber(0,1,false,
                      'Significance level |                    ',signif);
   rantet := false;
   ranpart := false;
end;

procedure readytorun(id: commands);
begin
     getnecessary(id);
     if (not rantet) then
       if (id in [prttr,regress,tetrad,monte]) then
         if (not aaabort) then
    	   begin
	      inittreks;
	      if (not aaabort) then calcpath(adjacency,paths1,1);
(*              if (not aaabort) then 
                  ucalcpath(adjacency,upaths1,1);  *)
	      if (not aaabort) then calctrek(treks1,paths1,1);
           end;
     if (not rantet) and (not aaabort) then
         if  ((id = compareedge) or (id = prttr)) then
	          begin
                     calccorr;
                     calctetrad(false);
                     if (not aaabort) then rantet := true;
                  end;
     if aaabort then
	    begin
	       twriteln(output,'Fatal error: cannot proceed|');
               ifilename := nullfilename;
	       ofilename := nullfilename;
	       ranpart := false; rantet := false;
	    end;
end;

procedure docompareedge(switchset: commandset); 
var finished,wanttet,wantpart,wantcommon,wasinserted: boolean;
    head,tail: vertexnametype; headpos,tailpos: integer;
begin
    twrite(output,'Edge:  |');
    finished := false;
    readedge(input,head,tail,finished,true);
    headpos := listposition(head,false,vernames,wasinserted);
    tailpos := listposition(tail,false,vernames,wasinserted);
    if adjacency^[headpos,tailpos] <> notthere then
         myerror('Edge is already in graph.|                ',fatal,false)
    else
      begin
         if common in switchset then
	   begin
             twrite(ofilvar,'Edge added: |');
             writevername(head,ofilvar);
             twrite(ofilvar,' C  |');
             writevername(tail,ofilvar);
             twriteln(ofilvar,'  |');
	   end
         else 
	   begin
             twrite(ofilvar,'Edge added: |');
             writevername(head,ofilvar);
             twrite(ofilvar,' -->  |');
             writevername(tail,ofilvar);
             twriteln(ofilvar,'  |');
	   end;
         twriteeoln(ofilvar);
         if (head = tail) and (head[1] <> ' ') then
            myerror('Head not allowed to equal tail.|          ',ignoreall,true);
         wanttet := false; wantpart := false;
         if (not (ignoreall in errors)) then
	      begin
	         wantcommon := common in switchset;
		 if (tetrad in switchset) or (switchset = [null]) or
		    (switchset = [common]) then
                      wanttet := true;
                 docompare(wantcommon,wanttet,headpos,tailpos,
                           impbegin,impend);
	  end;  { if }
        end;
end;	

procedure parseknowledge(var knowfile: text);
label 999;
var id: commands; firstsection,printederr: boolean;
begin
   printederr := false;
   justreaddata := false;
   firstsection := true;
   lineno := 0;
   id := null;
   while not eof(knowfile) do
      begin
         if (id <> parameters) and (id <> values) then
             readfileline(knowfile,buf);
         readwhitespace;
         if buf[curpos] = '{' then readcomment(knowfile) 
         else if buf[curpos] = '/' then
            begin
                printederr := false;
                id := parsecommand(false);
                if (id in datacommands) then 
                  if firstsection then 
                      begin
                        execcommand(id);
                        justreaddata := true;
                      end
                  else myerror('Data section must be first section in input file.|',
                              fatal,false)
                else execcommand(id);
                firstsection := false;
                if fatal in errors then 
                   begin
                      initinput([null],id in datacommands);
                      goto 999;
                   end;
            end
         else if (not myeoln) and (not printederr) then
             begin
                 myerror('Section Header required here: input being ignored.|',fatal,false);
                 printederr := true;
             end
       end;
999: end;

procedure parseparameters(var knowfile: text);
label 999;
var stop: boolean;
begin
   stop := false;
   while not eof(knowfile) and not stop do
      begin
         readfileline(knowfile,buf);
         if buf[curpos] = '/' then stop := true
         else execcommand(parsecommand(false));
         if fatal in errors then goto 999;
      end;
999: end; 

procedure getknowledge;
var changed: boolean; i,j: integer; count: integer;
begin
   getfile(knowfile,'Input File: |',graphf,knowfilename,changed);
   if not aaabort then
     begin
      curfilename := knowfilename;
      currentin := know;
      parseknowledge(knowfile);
      currentin := inp;
      count := 1;
      for i := 1 to maxcalc do
         for j := 1 to maxcalc do 
            if i < j then
	       begin
    		  pairtoint^[i,j] := count;
	          pairtoint^[j,i] := count;
   	          inttopair^[count].head := i;
	          inttopair^[count].tail := j;
	          count := count + 1;
	       end;
     end;
end;

procedure initforbid;
var i,j: vertextype;
begin
 for i := 1 to max do 
   begin
      temporder[i] := [];
      for j := 1 to max do 
       (* delete forbidden that came from temporal tiers *)
          if forbiddenadj^[i,j] = 0 then forbiddenadj^[i,j] := notthere;
   end;
end;

procedure calcorder;
label 999;
var i,j,k,l,m: vertextype;
begin
    initforbid;
    for i := 1 to max-1 do
      for j := i + 1 to max do
        for k := 1 to max do
           for l := 1 to max do
              if (k in temptiers[i]) and (l in temptiers[j]) then
                  if adjacency^[l,k] <> notthere then
                     begin
                        errormess('Ignoring temporal tiers because inconsistent with graph: |',fatal,false);
                        writename(l,output);
                        twrite(output,' -> |');
                        writename(k,output);
                        twriteeoln(output);
                        initforbid;
                        for m := 1 to max do temptiers[m] := [];
			goto 999;
                     end
                  else 
                     begin
                       temporder[k] := temporder[k] + [l];
                       if forbiddenadj^[l,k] <> 1 then 
                           forbiddenadj^[l,k] := 0;
                     end;
999: end;

procedure formtemporder(add: boolean);
begin
 if max >0 then
   begin
    if currentin = inp then writetemptiers(output);
    readtemptiers(add);
    calcorder;
    if currentin = inp then writetemptiers(output);
   end
 else
   begin
   if add then
    twriteln(output,'Addtemporal requires either data, a graph, a linear model or a Bayes net.|')
    else twriteln(output, 'No temporal tiers to remove|');
   end;
end;


procedure readnet;
var i,j: integer; varorderb: eqorder;
begin                  
   for j:= 1 to maxvertex do
       begin
           cats[j].num:= 2;
           for i:= 1 to maxcategories do cats[j].value[i] := i-1;
       end;
    varorderb := varorder;
    while (varorder <> nil) do
        begin
           varorder := varorder^.next;
           dispose(varorderb);
           varorderb:= varorder;
        end;
     new(varorder);
     varorder^.next := nil;
     varorder^.varbl := 0;
(*     formeqsorder(varorder,adjacency); *)
     readbayesnet(cats,varorder,knowfile,adjacency);
     havereadlinearmodel := false;
     havereadbayes := true;
     havereadgraph := true;
end;

procedure writehelp;
begin
   twriteln(output,'The main TETRAD II commands are: |');
   twriteln(output,'BUILD       Constructs a set of causal models from covariance |');
   twriteln(output,'            matrix or raw continuous data.|');
   twriteln(output,'ESTIMATE    Estimates the parameters of a discrete Bayesian Network.|');
   twriteln(output,'EXIT        Exits the program.|');
   twriteln(output,'MAKEMODEL   Creates a statistical model for the Monte Carlo module.|');
   twriteln(output,'MIMBUILD    Creates a set of linear multiple indicator models.|');
   twriteln(output,'MONTE       Creates pseudo-random samples from a given statistical model.|');
   twriteln(output,'PURIFY      Purifies indicators from a multiple indicator measurement model.|');
   twriteln(output,'SEARCH      Searches for respecifications of a given linear structural |');
   twriteln(output,'            equation model.|');
   twriteln(output,'STATUS      Prints the values of all parameters controlling output.|');
   twriteln(output,'STATWRITER  Prepares input files for EQS, LISREL, or CALIS.|');
   twriteln(output,'TETRADS     Computes the tetrad constraints implied by a model |');
   twriteln(output,'            and those that are judged to hold in the population.|');
   twriteln(output,'UPDATE      Computes conditional probabilities in a discrete Bayes Network.|');
end;

procedure execcommand(* id: commands *);
var switchset: commandset; changed: boolean; 
    temp: real; i,j: integer; c: char; forbid: edgerec; 
    temppart: real; ok: boolean; 
begin
 if (currentin = know) and (not (id in knowcommands)) then
    myerror('Command cannot be executed from Input File.|',ignoreall,true);
 if errors - [ignorepart,warning] = [] then
   begin
      if (not (id in [addedges,removeedges,help,partial,tetrad,
		      compareedge])) and (currentin <> know) then checkend;
      case id of
        tcom1,tcom2,tcom3,tcom4,tcom5,tcom6,tcom7,tcom8,tcom9,tcom10: ;
        rocfile:;
        testbuildm: comparepags1;
        randgraph: randomgraph1;  
        regress: ;
        means: getmu0(false,knowfile,vernames,changed);
        listvars: for i := 1 to max do
                   begin
                     pshort(i,output);
                     twriteeoln(output);
                   end;
	initial:;
        cequiv: ; 
	makemodel:;
        updatethenet: ;
        values: parseparameters(knowfile); 
        parameters: parseparameters(knowfile);
        estimate: ;
        readlinear:readlinearmodel(knowfile,adjacency);
        graph: getgraph;
        addtemp: formtemporder(true);
        monte: ;
        gibbs: ;
        lrt: ;
        removetemp: formtemporder(false);
        scales: ;
        purify: ;
	null: ;
        knowledge: getknowledge;
        statwriter: ;
        correlations: ;
        changegraph: dochangegraph;
	compareedge: getswitches([tetrad,partial,common],switchset);
	exit: doexit;
        addval: begin getnecessary(addval); if not aaabort then addev; end;
        removeval: begin 
                     getnecessary(removeval);
                     if not aaabort then removeev; 
                   end;
        removeallval: begin
                        getnecessary(removeallval); 
                        if not aaabort then removeallevidence;
                       end;
        help: ;
        showprogress: begin
                         initinput([showprogress],true);  
                         readnet;
                      end;
        independence: readpag(false,true,knowfile,temppag.augraph);
(*        independence:  begin
                        indepdata := true;
                        discrete := false;
                        celldata := false;
                        initinput([independence],true);
                        getcovariance(vernames,rho,cov);
                      end; *)
        covariance: begin
                  indepdata := false;
                  discrete := false;
                  celldata := false;
                  initinput([covariance],true);
                  getcovariance(vernames,rho,cov);
                end;  
        raw:  begin
                indepdata := false;
                discrete := true;
                celldata := false;
                initinput([raw],true);
                getcovariance(vernames,rho,cov);
              end;
        cellcounts: begin
                      indepdata := false;
                      discrete := true;
                      celldata := true;
                      initinput([cellcounts],true);
                      getcovariance(vernames,rho,cov);
                    end;
        readcon: begin
                    indepdata := false;
                    discrete := false;
                    celldata := false;
                    initinput([covariance],true);
                    readrawcontinuous(cov,rho);
                 end;
        outputfile: begin
                            getfile(ofilvar,
			    'Output file: |                          ',
			    outputf,ofilename,changed);
                            if (ofilename = blankfilename) or (ofilename = nullfilename) then
                                fileexists[outputf] := false
                            else fileexists[outputf] := true;
                     end;
	partial: ;
	tetrad: ;
(*        prttr: printpart := getnumber(0,100000000,true,'Printpartial |                    ',printpart); *)
        prttr: ;
        finalsets: ;
        lambdac: lambda := getnumber(0,100,false,'Lambda: |',1);
        significance: getsignificance;
        status: givestatus(output,true,status);
	setparameters:  begin
                         if maxtime = -1 then
                            maxtime := 60 * getnumber(-1,60000,false,
	                    'Settime (minutes)  |                    ',maxtime)
                         else maxtime := 60 * getnumber(-1,60000,false,
                            'Settime (minutes)  |',maxtime/60);
                         if maxtime <= 0 then maxtime:= -1.0;
                        end;
        decimals: begin
                    temp := getnumber(0,8,false,'Decimal points |',decpoints);
                    if not (fatal in errors) then decpoints := tround(temp);
                  end;
        apercent: begin
                    percent := getnumber(-1,1,false,'Width |',percent);
                  end;
        adepth: begin
		    temp := getnumber(-1,100,true,
	              'Depth |',depth); 
                    if not (fatal in errors) then depth := tround(temp);
                    if depth <= 0 then depth:= -1;
                end;
        addedges: edgeread(false,false);
        removeedges: edgeread(false,true);
        common: edgeread(true,true);
        forbidden: edgeread(true,false);
        acyclica: acyclic := getreply('ACYCLIC |','y',acyclic);
        lly: ll := getreply('LL |','y',ll);
        lmy: lm := getreply('LM |','y',lm);
        mmy: mm := getreply('MM |','y',mm);
        mly: ml := getreply('ML |','y',ml);
        nocommon: comm := getreply('Common |','y',comm);
        singleconnection: singleconn := getreply('Singleconnection |','y',singleconn);
        ndepth:  begin
                    temp := getnumber(-1,100,true,'NDepth |',nochangedepth); 
                    if not (fatal in errors) then nochangedepth := tround(temp);
                 end;
	aweight: begin
 		   weight := getnumber(0,1000,false,'Weight |',weight);
                   rantet := false;
   	           ranpart := false;
		end;
	split:;
	cross:;
      end;
    if id in runcommands then readytorun(id);
    if (not aaabort) then
      if id in runcommands then
        case id of
             tcom1: calcbicscore;
	     tcom2: asearch(temppag);
	     tcom3: updatedata;
             tcom4,tcom5,tcom6,tcom7,tcom8,tcom9,tcom10: ;
	     initial:;
(*             regress: linestimate(adjacency,rho,parm,ok); *)
             regress: doregress; 
	     makemodel:createmodel;
             cequiv: checkequiv; 
             statwriter: dostatwriter;
             estimate: doestimate(adjacency,ofilvar,false,MLexternal,0);
             updatethenet: updatebayesnetwork;  
(*             updatethenet: mse;   *)
	     correlations: showcorrelations;
	     compareedge: docompareedge(switchset);
             monte:  if not aaabort then domonte; 
             gibbs:  if not aaabort then dogibbs;
             lrt:  if not aaabort then dolrt;              
             finalsets: newsearch(adjacency);
	     prttr: autosearch;
	     partial: begin
                        getpartials;
                        ranpart := true;
                      end;
             tetrad: begin
                        calccorr;
                        calctetrad(true);
                        rantet := true;
                     end;
             scales: if not aaabort then buildlv(adjacency);
             purify: if not aaabort then makepure(adjacency);
             split: if not(aaabort) and not(fatal in errors) then dosplit;
             rocfile: if not(aaabort) and not(fatal in errors) then makerocfile;
             cross: if not(aaabort) and not(fatal in errors) then docrossvalidate;
           end;
(*        if (not aaabort) and (id in runcommands) then outputresults(id); *)
	{function scrapped        updatefile;}
      end;
   errors := errors - [fatal];
   if fileexists[outputf] then flush(ofilvar);
end;

procedure loadcommand;
(* these must be loaded in alphabetical order. *)
begin
    insertcommand('                    ',null,1);
    insertcommand('/BAYESNETWORK       ',showprogress,2);
    insertcommand('/COVARIANCE         ',covariance,3);
    insertcommand('/CELLCOUNTS         ',cellcounts,4);
    insertcommand('/GRAPH              ',graph,5);
    insertcommand('/INDEPENDENCE       ',independence,6);
    insertcommand('/KNOWLEDGE          ',parameters,7);
    insertcommand('/RAW                ',raw,8);
    insertcommand('/VALUE              ',values,9);
    insertcommand('ACYCLIC             ',acyclica,10);
    insertcommand('ADDEVIDENCE         ',addval,11);    
    insertcommand('ADDTEMPORAL         ',addtemp,12);   
    insertcommand('ALLOWCOMMON         ',removeedges,13);
    insertcommand('ALLOWDIRECT         ',addedges,14);
    insertcommand('BUILD               ',partial,15);
    insertcommand('DEPTH               ',adepth,16);
    insertcommand('STATWRITER          ',statwriter,17);
    insertcommand('ESTIMATE            ',estimate,18);
    insertcommand('EXIT                ',exit,19);  
    insertcommand('FORBIDCOMMON        ',common,20);  
    insertcommand('FORBIDDIRECT        ',forbidden,21);
    insertcommand('HELP                ',help,22);      
    insertcommand('INPUT               ',knowledge,23);
    insertcommand('MIMBUILD            ',scales,24);
    insertcommand('MONTE               ',monte,25);
    insertcommand('COMMON              ',nocommon,26);
    insertcommand('LL                  ',lly,27);
    insertcommand('LM                  ',lmy,28);
    insertcommand('ML                  ',mly,29);
    insertcommand('MM                  ',mmy,30);
    insertcommand('OUTPUT              ',outputfile,31);
    insertcommand('REMOVEALLEVIDENCE   ',removeallval,32);
    insertcommand('REMOVEEVIDENCE      ',removeval,33);
    insertcommand('REMOVETEMPORAL      ',removetemp,34);
    insertcommand('NEWSEARCH           ',finalsets,35);
    insertcommand('SETTIME             ',setparameters,36);
    insertcommand('SIGNIFICANCE        ',significance,37);
    insertcommand('SINGLECONNECTION    ',singleconnection,38);
    insertcommand('STATUS              ',status,39);
    insertcommand('TETRADS             ',tetrad,40);
    insertcommand('UPDATE              ',updatethenet,41);
    insertcommand('WEIGHT              ',aweight,42);
    insertcommand('WIDTH               ',apercent,43);
    insertcommand('/CONTINUOUSRAW      ',readcon,44);
    insertcommand('MAKEMODEL           ',makemodel,45);
    insertcommand('/LINEARMODEL        ',readlinear,46);
    insertcommand('@INIT               ',initial,47);
    insertcommand('TESTMONTE           ',testmonte,48);
    insertcommand('TESTSEARCH          ',testsearch,49);
    insertcommand('TESTBUILDL          ',testbuildl,50);
    insertcommand('TESTBUILDM          ',testbuildm,51);
    insertcommand('RANDGRAPH           ',randgraph,52); 
    insertcommand('ROC                 ',rocfile,53);
    insertcommand('PURIFY              ',purify,54);
    insertcommand('LISTVARS            ',listvars,55);
    insertcommand('DECIMALS            ',decimals,56);
    insertcommand('SPLIT               ',split,57);
    insertcommand('CROSS               ',cross,58);
    insertcommand('/MEANS              ',means,59);
    insertcommand('SEARCH              ',prttr,60);
    insertcommand('COMPARE             ',regress,61); 
    insertcommand('LRT                 ',lrt,62);
    insertcommand('GIBBS               ',gibbs,63);
    insertcommand('CEQUIV              ',cequiv,64);
    insertcommand('LAMBDA              ',lambdac,65);
    insertcommand('BICSCORE            ',tcom1,66);
    insertcommand('MAGSEARCH           ',tcom2,67); 
    insertcommand('UPDATA              ',tcom3,68); 
end;

procedure tnewindep(* var ptr: indepptr; err: errortype *);
var size: word;
begin
   size := sizeof(indepptr);
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create adj|',err,false);
end;

procedure tnewinsigcorr(* var ptr: insigcorrptr; err: errortype *);
var size: word;
begin
   size := sizeof(insigcorrarr);
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create insigcorrptr|',err,false);
end;

procedure tnewedgenames(* var ptr: edgenametypes; err: errortype *);
var size: word;
begin
   size := sizeof(edgenamearr);
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create edgenametype|',err,false);
end;

procedure tnewpairtoint(* var ptr: pairtointtype; err: errortype *);
var size: word;
begin
   size := sizeof(pairtointarr);
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create pairtoint|',err,false);
end;

procedure tnewinttopair(* var ptr: inttopairtype; err: errortype *);
var size: word;
begin
   size := sizeof(inttopairarr);
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create inttopair|',err,false);
end;
procedure tnewedgetovert(* var ptr: edgetoverttype; err: errortype *);
var size: word;
begin
   size := sizeof(edgetoverttype);
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create edgetovertype|',err,false);
end;

procedure tnewadj(* var ptr: adjtype; err: errortype *);
var size: word;
begin
   size := sizeof(ptr) * maxvertex * maxvertex;
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create adjtype|',err,false);
end;

procedure tnewcov(* var ptr: covarray; err: errortype *);
var size: word;
begin
   size := sizeof(ptr) * maxvertex * maxvertex;
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create covarray|',err,false);
end;

procedure tnewpriorcov(* var ptr: priorarray; err: errortype *);
var size: word;
begin
   size := sizeof(ptr) * maxparms * maxparms;
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create priorarray|',err,false);
end;

procedure tnewpaths(* var ptr: pathstype; err: errortype *);
var size: word;
begin
   size := sizeof(ptr) * maxvertex * maxvertex;
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create paths|',err,false);
end;

procedure tnewtreks(* var ptr: trekstype; err: errortype *);
var size: word;
begin
   size := sizeof(ptr) * maxvertex * maxvertex;
   if size < heapavail then new(ptr)
   else errormess('Out of memory: Cannot create treks|',err,false);
end;

procedure tnewudpaths(* var ptr: udpathstype; err: errortype *);
var size: word;
begin
   size := sizeof(ptr) * maxvertex * maxvertex;
   if size < heapavail then new(ptr)
   else errormess('Out of memory|',err,false);
end;

procedure writepaths1(i,j: integer);
begin
   if paths1^[i,j] = nil then writeln('nil') else writeln('not nil');
end;

(*-----------#pc-version-----------------------------------------------------*)
(*end.*)
(*-----------#pc-version-----------------------------------------------------*)
