import java.awt.*;

class Histogram extends Canvas {

        int numberBins;
        float dataPoints[] = new float[10000];
        int numPoints;

        //Constructor expects the number of data points to be
        //graphed "nPoints", the number of classes or "bins"
        //in the histogram "nBins", and an array of the data
        //to be graphed.

        public Histogram(int nPoints, int nBins, float[] data){
                int i;
                this.numberBins = nBins;
                this.numPoints = nPoints;
                for(i = 0; i < nPoints; ++i){
                        this.dataPoints[i] = data[i];
                 }
        }

        //drawNew is the non-constructor method that redraws
        //a graph in the same instance - it expects the same
        //arguments as the constructor

        public void drawNew(int nPoints, int nBins, float[] data){

                int i;
                numberBins = nBins;
                numPoints = nPoints;
                for(i = 0; i < nPoints; ++i){
                        dataPoints[i] = data[i];
                }
                this.repaint();
        }

        //paint overrides paint in Frame - it is the
        //place where all the calculations and drawing
        //for Histogram are done

        public void paint(Graphics gCanvas){

        //local varibles for paint

                float widthBin;
                int widthBinPlot;
                int i,j,k;
                float minValue;
                float maxValue;
                Rectangle plotRect = new Rectangle();
                int maxBinCount;
                int correction;
                Rectangle boundsRect;

                int binCount[] =  new int[1000];
                int binCountPlot[] = new int[1000];
                float boundary[] = new float[1001];
                int boundaryPlot[] = new int[1001];
                int numYTicks;
                float yTicksInterval;
                Font graphFont;

                //determine the size of the drawing area
                //and define a sub-area for the graph -
                //it is smaller by 60 pixels

                boundsRect = this.bounds();
                plotRect.x = boundsRect.x + 30;
                plotRect.y = boundsRect.y + 30;
                plotRect.width = boundsRect.width - 60;
                plotRect.height = boundsRect.height - 100;

                //find min and max datapoints

                minValue = dataPoints[0];
                maxValue = dataPoints[0];

                for(i = 1; i < numPoints; ++i){
                        if(dataPoints[i] < minValue)
                                minValue = dataPoints[i];
                        if(dataPoints[i] > maxValue)
                                maxValue = dataPoints[i];
                }
 
		for(i = 0; i < 20; i++) {
                  float min = (float) (-1.0 + i*0.1);
                  if(minValue < min) {
                    minValue = min - (float) 0.1;
                    break;
                    }
                  }

		for(i = 0; i < 20; i++) {
                  float max = (float) (-1.0 + i*0.1);
                  if(maxValue < max) {
                    maxValue = max;
                    break;
                    }
                  }

                //determine the width of the bins - for drawing purposes
                //variables to be drawn have two names - one for the
                //real values and one for the transformed drawing values
                //The convention is variablename and variablenamePlot

                widthBin = (maxValue-minValue)/(float)numberBins;
                //Test
                int width = Math.round(widthBin*((float) 1000.0));
                widthBin = (float) width*((float) 0.001);
                //Test
                widthBinPlot = (int)((float)plotRect.width/(float)numberBins);

                //boundary[] is an array of the x-values of teh
                //boundaries of the bins

                boundary[0] = minValue;

                for(i = 1; i < numberBins + 1; ++i){
                        //boundary[i] = boundary[i - 1] + widthBin;
                        boundary[i] = boundary[0] + i*widthBin;
                }

                //initialize bin values to 0

                for(i = 0; i < numberBins; ++i){
                        binCount[i] = 0;
                }

                //calculate height of each bin

                for(j = 0; j < numberBins; ++j){
                        for(i = 0; i < numPoints; ++i){
                                if(dataPoints[i] >= boundary[j] &&
                                        dataPoints[i] < boundary[j + 1])
                                        binCount[j] = binCount[j] + 1;
                        }
                }

                //determine the heigth of the bin with the
                //higest values

                maxBinCount = binCount[0];
                for(i = 1; i < numberBins; ++i){
                        if(binCount[i] > maxBinCount)
                                maxBinCount = binCount[i];
                }

                //determine ploting values for the boundaries
                //of each bin

                boundaryPlot[0] = plotRect.x + 1;
                for(i = 1; i < numberBins + 1; ++i){
                        boundaryPlot[i] = boundaryPlot[i-1] +
                                widthBinPlot;
                }

                //find out how the resulting graph doesn't match up
                //in pixels with the originally defined area and
                //correct the drawing area definition

                correction = (plotRect.x + plotRect.width - 1) -
                                (boundaryPlot[numberBins - 1] +
widthBinPlot);

                plotRect.width = plotRect.width - correction;
                plotRect.x = plotRect.x + correction/2;

                //draw the bounding rectangle of the graph

                gCanvas.setColor(Color.black);
                gCanvas.drawRect(plotRect.x, plotRect.y,
                                        plotRect.width,
plotRect.height);

                //recalculate the plot values of the bin values
                //in the new drawing rectangle

                boundaryPlot[0] = plotRect.x + 1;
                for(i = 1; i < numberBins + 1; ++i){
                        boundaryPlot[i] = boundaryPlot[i-1] +
                                widthBinPlot;
                }

               
                //determine the plot values for the bin heights

                for(i = 0; i < numberBins + 1; ++i){

                        binCountPlot[i] = plotRect.y + plotRect.height -
                                (int)((((float)binCount[i]/
                               
(float)maxBinCount))*(float)plotRect.height);
                }

                gCanvas.setColor(Color.red);

                //draw the bins

                for(i = 0; i < numberBins; ++i){
                        gCanvas.drawRect(boundaryPlot[i],
                                binCountPlot[i],
                                widthBinPlot,
                                (int)((((float)binCount[i]/
                               
(float)maxBinCount))*(float)plotRect.height) - 1);
                }

                //label the axes - thes routines prove 10 ticks for
                //the x-axis and five for the y-axis

                graphFont = new Font("TimesRoman", Font.PLAIN, 10);
                gCanvas.setFont(graphFont);

                int xAxisTickXPlot[] = new int[1000];
                int xAxisTickYPlot[] = new int[1000];
                int yAxisTickXPlot[] = new int[1000];
                int yAxisTickYPlot[] = new int[1000];
                float yAxisTickY[] = new float[1000];

                int inc;
                inc = (numberBins + 1)/10;
                if(inc < 1)
                        inc = 1;

                for(i = 0; i < numberBins + 1; i = i + inc){
                        xAxisTickXPlot[i] = boundaryPlot[i] - 5;
                        xAxisTickYPlot[i] = plotRect.y + plotRect.height
+ 10;
                       
gCanvas.drawString(Float.toString(boundary[i]),
xAxisTickXPlot[i],
                                xAxisTickYPlot[i]);
                }
                
                numYTicks = 5;
             
                yTicksInterval = ((float)maxBinCount/(float)(numYTicks -
1));

                yAxisTickY[0] = 0.0F;

                for(i = 1; i < numYTicks; ++i){
                        yAxisTickY[i] = yAxisTickY[i - 1] +
(float)yTicksInterval;
                }

                for(i = 0; i < numYTicks; ++i){
                        yAxisTickYPlot[i] = plotRect.y + plotRect.height
-
                                (int)((((yAxisTickY[i]/
                               
(float)maxBinCount))*(float)plotRect.height));
                        yAxisTickXPlot[i] = plotRect.x - 20;
                       
gCanvas.drawString(Float.toString(Math.round(yAxisTickY[i])),
yAxisTickXPlot[i],
                                yAxisTickYPlot[i]);
                }
                        
        }

public boolean handleEvent(Event evt) {
  if(evt.id == Event.WINDOW_DESTROY) {
    System.exit(0);
    return false;
    }
  else return super.handleEvent(evt);
  }
}


