//Copyright (C) 1997 Frank C. Wimberly.  All rights reserved.


import java.util.*;  //Needed for Stack class and its methods

class diGraph extends Graph {

  protected int ntiers;
  protected int tier[];
  protected int tiers[][];
  protected int nvars[];
  protected int numParents[];
  protected int parent[][];

  diGraph(int nvertices, int edgeMatrix[][], String name, String vertexNames[])
  {
    super(nvertices, edgeMatrix, name, vertexNames);
    ntiers = 0;
    tier = new int[nvertices];
    tiers = new int[nvertices][nvertices];
    nvars = new int[nvertices];
    numParents = new int[nvertices];
    parent = new int[nvertices][nvertices];
  }

  diGraph(causalGraph G, String name, String vertexNames[])
  {
    super(G, name, vertexNames);
  }

  public boolean adjacent(int tail, int head) 
  { 
    if(edgeMatrix[tail][head] == OUTOF && edgeMatrix[head][tail] == INTO) return true;
    else return false;
  }

  public boolean isCyclic( )
  { 
    boolean ans = false;
    this.isPath(0, 0);
    for (int i = 0; i < nvertices; i++)
       if( paths[i][i] == 1) {
          ans = true;
          return ans;
       }
     return ans;
  }

  public void findParents( ) {
  for (int vertex = 0; vertex < nvertices; vertex++) {
    int parents = -1;
    for(int v = 0; v < nvertices; v++)
      if(adjacent(v, vertex)) {
        parents++;
        parent[vertex][parents] = v;
        }
    numParents[vertex] = parents + 1;
    }
  }

public int getNumParents(int vertex) 
  {
  return numParents[vertex];
  }

public int getParent(int vertex, int i)
  {
  return parent[vertex][i];
  }

  public void setupTiers( )
  {
    int i, j, t;

    for(i = 0; i < nvertices; i++) {
      tier[i] = -1;
      if(numParents[i] == 0) tier[i] = 0;
      }

    for(t = 1; t < nvertices; t++)
      for(i = 0; i < nvertices; i++)
        for(j = 0; j < nvertices; j++)
          if(tier[j] == t-1 && adjacent(j, i)) {
            tier[i] = t;
            if(t > ntiers) ntiers = t;
            }
    ntiers++;
    //for(i = 0; i < nvertices; i++) {
    //  System.out.println("Node " + i + " is at " + tier[i]);
    //  }
    for(j = 0; j < ntiers; j++) {
      nvars[j] = 0;  
      for(i = 0; i < nvertices; i++) if(tier[i] == j) {
        tiers[j][nvars[j]] = i;  //add variable i to list of variables in tier j
        nvars[j]++;              //increment number of variables in tier j 
        }
      }
  }

  public boolean dSeparated(int x, int y, nodeList z)
  {
     boolean ans = false; boolean above[] = new boolean[nvertices];
     boolean below[] = new boolean[nvertices];
     orderedPair current;
     int i, j, v, subscript;
   
     for ( i = 0; i < nvertices; i++) {
       above[i] = false;
       below[i] = false;
       }

     Stack stack = new Stack( );
     orderedPair Xabove = new orderedPair(x, 1);
     orderedPair Xbelow = new orderedPair(x, -1);
     stack.push(Xabove);
     stack.push(Xbelow);
     below[x] = true;
     above[x] = true;

     while( !stack.empty( )) {
       current = (orderedPair) stack.pop( );
       v = current.valueofFirst( );
       subscript = current.valueofSecond( );
       if(subscript < 0) {
         for ( i = 0; i < nvertices; i++ ) //parents
           //if ( adj[i][v] == 1 )
           if(adjacent(i, v))
             if(!below[i] && (z.isElement(i) < 0)) {
                below[i] = true;
                orderedPair Vbelow = new orderedPair(i, -1);
                stack.push(Vbelow);
                }
         for ( j = 0; j < nvertices; j++) //children
           //if ( adj[v][j] == 1)
           if(adjacent(v, j))
             if(!above[j]) {
               above[j] = true;
               orderedPair Tabove = new orderedPair(j, 1);
               stack.push(Tabove);
               }
           above[v] = true;
         } // v < 0
         else {
           if ( z.isElement( v ) >= 0 ) { // v known
             for ( i = 0; i < nvertices; i++ ) //parentsv
               //if ( adj[i][v] == 1 )
               if(adjacent(i, v))
                 if ( (z.isElement( i ) < 0) && !below[i] ) {
                   below[i] = true;
                   orderedPair Tbelow = new orderedPair(i, -1);
                   stack.push(Tbelow);
                   }
           }
           else // v not known
             for ( j = 0; j < nvertices; j++ ) //children
               //if ( adj[v][j] == 1 )
               if(adjacent(v, j))
                 if( !above[j] ) {
                   above[j] = true;
                   orderedPair Sabove = new orderedPair(j, 1);
                   stack.push(Sabove);
                   }
       }
     } // while
     if (below[y] || above[y]) ans = false; else ans = true;
     return ans;
  }

}

