(*-----------#pc-version-----------------------------------------------------*)
(*
implementation
uses sys;
*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "linbayes.h"
#include "util.h"
#include "sys.h" 
#include "io.h" 
#include "mon.h" 
#include "part.h" 
#include "rsutil.h" 
#include "newcalc.h"
#include "recalc.h"
#include "numeric.h"
#include "calctet.h"


const
pi = 3.1415926535;

function expon(* a,b: real): real; *);
var temp: real;
begin
   temp := b * ln(a);
   expon := exp(temp);
end;

function cfunc(n: integer; a: real): real;
var temp: real; i: vertextype;
begin
  temp := 0;
  for i := 1 to n do 
    temp := temp + gammln((a + 1 - i)/2);
  cfunc := temp;
end; 

{procedure getmum(nu: real; m,n: integer; var smeans,mu0: vertarray);
var i: vertextype;
begin
   for i := 1 to n do 
       mum[i] := (nu * mu0[i] + m * smeans[i])/(nu + m);
end;}

procedure gettm(var subtm,truecov,subt0: covarray; var mu0,smeans: vertarray; n: integer;
                nu: real; size: vertextype);
var i,j: vertextype;
begin
   for i := 1 to size do
     for j := 1 to size do 
        subtm^[i,j] := subt0^[i,j] + cov^[map[i],map[j]] * n + 
                      ((nu * n)/(nu + n)) * (mu0[map[i]] - smeans[map[i]]) * 
                       (mu0[map[j]] - smeans[map[j]]);
end;

procedure gett0(var t0,cov: covarray; a,nu: real; n: vertextype);
var i,j: vertextype; gamma: real;
begin
  gamma := a - n + 1;
  for i := 1 to n do
     for j := 1 to n do
        t0^[i,j] := cov^[i,j] * 
              (nu * gamma * (gamma  - 2)) / ((nu + 1) * gamma);
end;

procedure newgetmu0;
var i: vertextype; varorder: eqorder;
begin
   for i := 1 to numob do
     mu0[i] := exogd[i].parm1;
end;
 
procedure formmap(v: vertextype; var map: parray; includev: boolean; var numvar: vertextype);
var i,j: vertextype;
begin
   j := 1;
   numvar := 0;
   for i := 1 to max do 
     if (adjacency^[i,v] <> notthere) or ((i = v) and includev) then
        begin
           numvar := numvar + 1;
           map[j] := i;
           j := j + 1;
        end;
end;

procedure formsubt0(var t0inv,subt0: covarray; map: parray; size: vertextype);
var ok: boolean;
begin
   copymat(t0inv,tempt0inv);
   for i := 1 to size do
      for j := 1 to size do
        tempt0^[i,j] := tempt0inv^[map[i],map[j]];
   inverse(tempt0,subt0,size,ok);
end;

function determinant(var mat: covarray; n: integer): real;
var valdeterm: real; j: integer; done: array[1..maxvertex] of boolean;
  
  procedure doit(term: real; m,k: integer);
  var j,r,sign: integer;
  begin
     if k > n then
       begin
         sign := 1;
         if odd(m) then sign := -1;
         valdeterm := valdeterm + sign * term;
       end
     else
        if term <> 0.0 then
           begin
             r := 0;
             for j := n downto 1 do if done[j] then r := r + 1
             else 
                begin
                  done[j] := true;
                  doit(term * mat^[k,j],m+r,k+1);
                  done[j] := false;
                end
           end
  end;

begin
   for j := 1 to n do done[j] := false;
   valdeterm := 0.0;
   doit(1.0,0,1);
   determinant := valdeterm;
end;

procedure getprobmod(nu,a: real; m: integer; var subt0,subtm: covarray;
                      var theprob: real; n: vertextype);
var g0,g1,g2,g3,g4,g5,g6,g7,g8,g9,g10,awy: real;
begin
if n = 0 then theprob := 0.0
else 
 begin
   awy := a - numob + n;
   g0 := ln(2) * n * m / 2 ;
   g1 := ln(pi) * (-1 * n * m)/2;
   g3 := ln(nu/(nu + m)) * n/2;
   g4 := cfunc(n,awy+m);
   g5 := cfunc(n,awy);
   g6 := determinant(subt0,n);
   g7 := ln(g6) * awy/2;
   g8 := determinant(subtm,n);
   g9 := ln(g8) * (awy+m)/(-1 * 2);
   theprob := (*g0 + *) g1 + g3 + g4  - g5 + g7 + g9;
 end;
end;

(*
procedure getprobmod(nu,a: real; m: integer; var t0,tm: covarray; 
                      var theprob: real; thevset: vertexset; n: vertextype);
var g1,g2,g3,g4,g5,g6,g7,g8,g9,g10,awy: real;
begin
if n = 0 then theprob := 0.0
else 
 begin
   awy := a - numob + n;
   readtomat(t0inv,temp1t0,temp2t0,thevset);
   readtomat(tminv,temp1tm,temp2tm,thevset);
   g1 := expon(pi,(-1 * n * m)/2);
   g3 := expon(nu/(nu + m),n/2);
   g4 := cfunc(n,awy+m);
   g5 := cfunc(n,awy);
   g6 := determinant(temp2t0,n);
   g7 := expon(g6,awy/2);
   g8 := determinant(temp2tm,n);
   g9 := expon(g8,(awy+m)/-2);
   g10 := g1 * g3 * (g5/g4) * g7 * g9;
   theprob := ln(g10); 
   theprob := 
   ln(
    expon(pi,(-1 * n * m)/2) * expon(nu/(nu + m),n/2) *
    (cfunc(n,awy+m) / cfunc(n,awy)) * expon(determinant(temp2t0,n),awy/2) * 
    expon(determinant(temp2tm,n), (awy+m)/-2));
 end;
end;
*)
procedure disposelinear;
begin
  dispose(t0);
  dispose(tm);
  dispose(t0inv);
  dispose(tempt0);
  dispose(tempt0inv);
  dispose(subt0);
  dispose(subtm);
end;

procedure initlinear;
var i,j: vertextype; ok: boolean;
begin
  new(t0);
  new(tm);
  new(t0inv);
  new(tempt0);
  new(tempt0inv);
  new(subt0);
  new(subtm);
  nu := getnumber(0,100000000,false,'nu: |',nu);
  a := getnumber(0,100000000,false,'Alpha: |',a); 
  newgetmu0;
  gett0(t0,cov,a,nu,numob);
  copymat(t0,tempt0);
  inverse(tempt0,t0inv,numob,ok);
end;

procedure getlinscore(* var lambda,thescore: real *); 
var i,j,k,numvar: vertextype; theprob1,theprob2: real;
begin
   thescore := 0.0;
   for i := 1 to max do
      begin
         formmap(i,map,true,numvar);
         formsubt0(t0inv,subt0,map,numvar);
         for j := 1 to max do
           for k := 1 to max do
             begin
               if j <> k then truecov^[j,k] := 0
               else truecov^[j,k] := 1;
             end;
         gettm(subtm,truecov,subt0,mu0,smeans,round(sampsize),nu,numvar);
         getprobmod(nu,a,round(sampsize),subt0,subtm,theprob1,numvar);
         if numvar = 1 then theprob2 := 0 else
             begin
               formmap(i,map,false,numvar);
               formsubt0(t0inv,subt0,map,numvar);
               gettm(subtm,cov,subt0,mu0,smeans,round(sampsize),nu,numvar);
               getprobmod(nu,a,round(sampsize),subt0,subtm,theprob2,numvar);
             end;
         lscorearr[i] := theprob1 - theprob2;
         thescore := thescore + theprob1 - theprob2;
      end;
end;

procedure getmodlinscore(* lambda,locscore: real; v: vertextype *);
var theprob1,theprob2: real; numvar: vertextype;
begin
   score := score - lscorearr[v];
   formmap(v,map,true,numvar);
   formsubt0(t0inv,subt0,map,numvar);
   gettm(subtm,cov,subt0,mu0,smeans,round(sampsize),nu,numvar);
   getprobmod(nu,a,round(sampsize),subt0,subtm,theprob1,numvar);
   if numvar = 1 then theprob2 := 0
      else
         begin
           formmap(v,map,false,numvar);
           formsubt0(t0inv,subt0,map,numvar);
           gettm(subtm,cov,subt0,mu0,smeans,round(sampsize),nu,numvar);
           getprobmod(nu,a,round(sampsize),subt0,subtm,theprob2,numvar);
        end;
   locscore := theprob1 - theprob2;
end;


procedure assigncov(var cov: covarray; i,j: vertextype; val: real);
begin
   cov^[i,j] := val;
end;

procedure assignadj(var adjacency: adjtype; i,j: vertextype; val: integer);
begin
   adjacency^[i,j] := val;
end;

procedure assignparents(var parents: tempset; i: vertextype);
var num: vertextype;
begin
   num := 0;
   parents[i] := [];
   repeat
     write('num: ');
     readln(num);
     if num <> 0 then parents[i] := parents[i] + [num];
   until num = 0;
end;



