(*-----------#pc-version-----------------------------------------------------*)
(*
implementation
uses io,util,rsutil,calctet,calctet1,sys;
*)
(*-----------#pc-version-----------------------------------------------------*)

(*-----------#unix-version-----------------------------------------------------*)
#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "mon.h"
#include "calctet.h"
#include "calctet1.h"
#include "io.h"
#include "sys.h"
#include "util.h"
#include "eqs.h"
#include "scal.h"
#include "rsutil.h"
#include "cspecial.h"
#include "makemod.h"
#include "cross.h"
#include "linbayes.h"
#include "gib.h"
#include "up.h"
#include "part.h"
#include "indep1.h"

(*-----------#unix-version-----------------------------------------------------*)

(****************************)
(*   MON.P   *)

(* functions devoted to generated sample data from a known model by using
    pseudo random number generators.  *)


(* Worries: 1) how to exit if aaabort true, e.g. in readcutoff, or in getdatatypes
            2) concat different in turbo
            3) getnumber: if the number is in integer, the prompt should be displayed
               as an integer, but isn't. *) 

var 
   havevalue, novalueyet : vertexset;
   newmax,montesamp, numdatasets, datasetnum : integer;
   curout, coefin: text;
   datatype: 1..6;  (*1=raw, 2=sampcov, 3=sampcor,
                      4=popcov, 5=popcor *)
   tiers : tiersptr;
   minprob: real;
   rsvernames:namearray;
   rearr: array[1..maxvertex] of vertextype; 
   

(**************************************)
(* Utilities *)

  (* counts the number of elements in a numberset *)
function card(datatypes: numberset): integer;
var count,i: integer;
begin
   count := 0;
   for i := 1 to 14 do
     if i in datatypes then count := count + 1;
   card := count;
end;

(* procedure scramble;
var arrset: vertexset; done: boolean; randnum: vertextype; i,newseed: integer;
begin
   initrand(newseed);
   arrset := [];
   for i := 1 to numob
     begin
      done := false;
      repeat
         randnum := round(thisrand * (numob - 1));
         if (not (randnum in arrset)) then 
            begin
               rearr[i] := randnum + 1;
               arrset := arrset + [randnum];
               done := true;
            end;
      until done;
     end;
end; *)

procedure scramble;
var arrset: vertexset; done: boolean; randnum: vertextype; i,newseed: integer;
begin
   initrand(newseed);
   arrset := [];
   for i := 1 to numob do rearr[i] := i;
end; 


procedure makecorrmat(numvars: integer; var cov, cor: covarray); 
  label 999;
  var i,j: integer;
      sd: vertarray;
  begin
     for i:= 1 to numvars do
      if i in graphvertices then   (* c *)
        begin
          sd[i]:= sqrt(cov^[i,i]);
          if sd[i] = 0 then
             begin
                 errormess('Variance equal to 0. Cannot form correlation matrix.|',fatal,false);
                 twrite(output,': |');
	         writename(i,output); twriteeoln(output);
                 goto 999;
             end;
        end;
     for i:= 1 to numvars do
       if i in graphvertices then (* c *)
         for j:= i to numvars do
            if j in graphvertices then (* c *)
	     begin
	        cor^[i,j]:= cov^[i,j]/(sd[i] * sd[j]);
	        cor^[j,i]:= cor^[i,j];
             end; 
999:  end;  (* makecorrat *)

(*********************************)
(* Disposing of all the pointers *)

procedure disposepointrs(*var tiers: tiersptr; var varorder: eqorder*);
  var i: integer;
      tiersa: tiersptr;
      varorderb: eqorder;
    begin
       varorderb:= varorder;
       while (varorder <> nil) do
          begin
              varorder := varorderb^.next;
              dispose(varorderb);
              varorderb:= varorder;
          end;
       tiersa:= tiers;
       while (tiers <> nil) do
          begin
              tiers := tiersa^.next;
              dispose(tiersa);
              tiersa:= tiers;
          end;
    end;

(* End of Utilities *)
(********************)

(* ************ Prob stufff  ***********************)

function randunif(*lower, upper: real): real *);
(* Draws a value at random from a distribution 
    uniform over the interval lower to upper. *)

begin 
    randunif:= lower + thisrand*(upper - lower);
end;	


 function randstnorm: real;
 (* Returns a value selected from a standard normal distribution *)
 (* Uses Box-Muller method. Ann. Math. Statistics 29, 1958 pp.610-11
    or #267 of CACM collected algorithms *)

 const twopi=6.2831853072;	
 var  y,x,t  :real;

  begin
      y := thisrand;
      x:= sqrt(-2*ln(y));
      t:= thisrand*twopi;
      randstnorm := x*cos(t)
  end;	(* function RandStNorm *)
    

function randnorm(*mean, variance: real): real*);
 (* Draws a value at random from normal distribution with 
    mean and variance passed in as parameters *)

  begin
    randnorm:= mean + randstnorm * sqrt(variance);
 end;	

function randlognorm(* median, variance: real): real; *);
(* Draws a value at random from a lognormal distribution with 
    median and variance factor passed as parameters *)

begin
      randlognorm:= exp(ln(median) + randstnorm*ln(sqrt(variance)));
end;	

(*******************************************)
(* Printing Routines *)    

procedure writemat(*mat : covarray; var outfile : text; datatype: integer*);
  var i,j: integer;
  begin
    if card(datatypes) > 1 then
      begin
       twriteeoln(outfile);
       case datatype of
        2: twriteln(outfile, 'Sample Covariance Matrix|');
        3: twriteln(outfile, 'Sample Correlation Matrix|');
        4: twriteln(outfile, 'Population Covariance Matrix|');
        5: twriteln(outfile, 'Population Correlation Matrix|');
        (* 6: twriteln(outfile, 'Implied Covariances Symbolically|');*)
       end;
       twriteeoln(outfile);
     end; (* if card *)
    for i := 1 to numob do 
     if i in graphvertices then
       begin pshort(i,outfile); twrite(outfile,'  |'); end;
    twriteeoln(outfile);
    for i:= 1 to numob do
     if i in graphvertices then
      begin
        for j:= 1 to i do 
         if j in graphvertices then
           begin
              tpwritereal(outfile, mat^[i,j], 10,decpoints,false);
	      twrite(outfile,' |');
              if (j mod 7  = 0) or (j = i) then twriteeoln(outfile);
           end;
      end;
    twriteeoln(outfile);
  end;

procedure writecormatrices(var outfile: text);
   begin
     if (sampcov in datatypes) then writemat(scov, outfile, 2);
     if (not (fatal in errors)) and (sampcor in datatypes) then writemat(scor, outfile, 3);
     if (popcov in datatypes) and (not minternal) then writemat(truecov, outfile, 4);
     if (not (fatal in errors)) and (popcor in datatypes) then writemat(truecor, outfile, 5);
     flush(outfile);
   end;

procedure printdist(var cur: probptr; effect:vertextype; 
                       var outfile:text; defaults:boolean;
                       prevpar: parrecord);
  var i,j,k: integer;
      thisprob: real;
  begin
    if cur^.varnum = effect then 
     begin
      if prevpar.numprev > 0 then
         begin
           twrite(outfile,' when |');
           for i:= 1 to prevpar.numprev do
             begin
               writename(prevpar.parn[i], outfile);
               twrite(outfile,'=|');
               twriteint(outfile,tround(prevpar.parv[i]));
               twrite(outfile,' |');
             end;
           twriteeoln(outfile);
         end; (* if there are parents *)
      for j:= 1 to cats[effect].num do
       begin
         if defaults then thisprob:= 1/cats[effect].num
         else thisprob:= cur^.cutoff[j]-cur^.cutoff[j-1];
         if (j>1) and (((j-1) mod 4) = 0) then twriteeoln(outfile);
         twrite(outfile,'  p(|');
         writename(effect,outfile);
         twrite(outfile,'=|');
         twriteint(outfile,tround(cats[effect].value[j]));
         twrite(outfile,')=|');
         if thisprob <> -1  then 
            tpwritereal(outfile,thisprob,8,decpoints,false)
         else 
             begin
                twrite(outfile,' U|');
                for k := 1 to 8 do twrite(outfile,' |');
             end;
       end;
      twriteeoln(outfile);
     end (* if varnum=effect *)
    else
     begin
       prevpar.numprev:= prevpar.numprev+1;
       prevpar.parn[prevpar.numprev]:= cur^.varnum;
       for j:= 1 to cats[cur^.varnum].num do
         begin
           prevpar.parv[prevpar.numprev]:= cats[cur^.varnum].value[j];
           printdist(cur^.nextvalue[j],effect,outfile,defaults,prevpar);
         end;
     end;
  end;


(* this prints out the bayesian network, which is stored in cats. 
   eqs holds the order of the vars, default is true
   if we are creating a default file that someone should modify. *)
procedure writediscretemodel(* var cats: categoryarray; var eqs: eqorder; 
                             var graph: adjtype; var outfile: text; default: boolean *);
  var i,j: integer;
      front: eqorder;
      prevpar: parrecord;
  begin
    (* Write the vertices and the number and values of their categories *)
    front:= eqs;
    twriteeoln(outfile);
    twriteln(outfile,'/BAYESNETWORK|');
    twriteln(outfile,'          Number of   Values of|');
    twriteln(outfile,'Variable  Categories  Categories|');
    while (eqs^.next <> nil) do
     begin
       i:= eqs^.varbl;
       pwritename(i,outfile,9); twrite(outfile,'  |');
       tpwriteint(outfile,cats[i].num,4,true);
	twrite(outfile,'         |');
       for j:= 1 to cats[i].num do
         tpwriteint(outfile,tround(cats[i].value[j]),4,false);
       twriteeoln(outfile);
       eqs := eqs^.next;
     end;
    eqs := front;
    twriteln(outfile,'The Probability Distribution|');
    while (eqs^.next <> nil) do
      begin
        twriteln(outfile,'----------------------------|');
        i:= eqs^.varbl;
        writename(i,outfile);
        twrite(outfile,'  Parents: |');
        if (causes[i] <> []) then 
         begin
           for j:= 1 to origmax do
            if j in graphvertices then   (* c *)
             if (graph^[j,i] <> notthere) then 
               begin writename(j,outfile); twrite(outfile,' |'); end;
             end;
           twriteeoln(outfile);
        prevpar.numprev:= 0;
        for j:= 1 to maxvertex do 
           begin
             prevpar.parn[j]:= 0;
             prevpar.parv[j]:= 0;
           end;
        printdist(cats[i].condprob,i,outfile,default,prevpar);
        eqs:= eqs^.next;
      end;
    twriteeoln(outfile);
    eqs:= front;
  end; (* writediscretemodel *)



procedure writerawheading(var outfile: text);
  var i: integer;
  begin
   twrite(outfile,' |');
   for i:= 1 to numob do
     if i in graphvertices then
         if undercontinuous then pwritename(rearr[i],outfile,12)
         else pwritename(rearr[i],outfile,6);
    twriteeoln(outfile);
  end;

procedure writerawcas(var varval: vertarray; var outfile: text);
   var i: integer;
   begin
     for i:= 1 to numob do
       if i in graphvertices then
        if undercontinuous then tpwritereal(outfile,varval[rearr[i]],12,decpoints,false)
                           else tpwriteint(outfile,tround(varval[rearr[i]]),6,false);
     twriteeoln(outfile);
   end;


procedure printeqorder(list:eqorder);
   begin
     while (list^.next <> nil) do
       begin
	 pwritename(list^.varbl,output,6);
         list:= list^.next;
       end;
     writeln;
   end; 

procedure writeequation(*var outfile: text; var cureqs: eqorder;
                        var graph: adjtype; var curparm: covarray*);
   var cause, effect: integer;
   begin
     effect := cureqs^.varbl;
     writename(effect,outfile);
     twrite(outfile,' = |');
     for cause := 1 to origmax do
      if cause in graphvertices then          (* c *)
       if graph^[cause, effect] <> notthere then 
	begin
            tpwritereal(outfile, curparm^[cause,effect],6,3,true);
	    writename(cause,outfile);
	    twrite(outfile,' + |');
	end;
     twrite(outfile, 'e|');
     twriteint(outfile,effect); 
     cureqs := cureqs^.next;
     twriteeoln(outfile); 
   end;


procedure writeexogdist(var outfile: text; var exogd: exogarray);
  var i: vertextype;
  begin
    twriteln(outfile,'Distribution over exogenous variables|'); 
    twriteeoln(outfile);
    twriteln(outfile,' Error    Distributional|');
    twriteln(outfile,'term for     Family                  Parameters|');
    twriteln(outfile,'--------  --------------             ----------|');
    for i:= 1 to maxvertex do
      if (i in graphvertices) then
        begin
          twrite(outfile,'  |');
          pwritename(i,outfile,5);
          twrite(outfile,'    |');
          case exogd[i].probtype of
            1: begin
                twrite(outfile,' Uniform |');
                twrite(outfile,'         Lower: |');
                tpwritereal(outfile,exogd[i].parm1,8,decpoints,false);       
                twrite(outfile,'  Upper:    |');
                tpwritereal(outfile,exogd[i].parm2,8,decpoints,false);       
                twriteeoln(outfile);
               end;
            2: begin
                twrite(outfile,' Normal|');
                twrite(outfile,'           Mean: |');
                tpwritereal(outfile,exogd[i].parm1,8,decpoints,false);       
                twrite(outfile,'   Variance: |');
                tpwritereal(outfile,exogd[i].parm2,8,decpoints,false);       
                twriteeoln(outfile);
               end;
            3: begin
                twrite(outfile,' Lognormal|');
                twrite(outfile,'        Median: |');
                tpwritereal(outfile,exogd[i].parm1,8,decpoints,false);       
                twrite(outfile,' Variance: |');
                tpwritereal(outfile,exogd[i].parm2,8,decpoints,false);       
                twriteeoln(outfile);
               end;
          end; (* case *)
       end; (* if i in graphvertices *)
      twriteeoln(outfile);
    end; (* writeexog *)


procedure writelinearmodel(cureqs: eqorder; var outfile : text; 
                           var graph: adjtype; var curparm: covarray);
   begin
     twriteln(outfile,'The Generating Model|'); 
     twriteln(outfile,'Linear Structural Equation Model|'); 
     twriteeoln(outfile);
     writeexogdist(outfile,exogd);
     twriteln(outfile,'Structural Equations|');
     twriteln(outfile,'--------------------|');
     while (cureqs^.next <> nil) do 
        writeequation(outfile,cureqs,graph,curparm);
     twriteeoln(output);
   end;

procedure writecomment(var outfile: text);
  begin
     twriteln(outfile,'{|');
     if undercontinuous then 
         writelinearmodel(varorder,outfile,adjacency,parm)
     else 
       begin
         twriteln(outfile,'The Generating Model|'); twriteeoln(outfile);
         writediscretemodel(cats,varorder,adjacency,outfile,false);
       end;
     twriteln(outfile,'  }|');
  end;

(* End of Printing Routines *)
(****************************)

(**************************)
(*  Initializing Routines *)

procedure monteinit;
     label 999;
     var i,j: integer;
     begin
       tnewcov(scov,fatal);       (* check if enough space *)
(*       if (not minternal) then tnewcov(truecov,fatal); *)
       tnewcov(scor,fatal);
       if (not minternal) then tnewcov(truecor,fatal);
       if fatal in errors then goto 999;
       prnton:= true;
       minprob:= 0.001;
       parmsfixed:= false;
       undercontinuous := true; 
       montesamp := 2000; 
       numdatasets := 1;
       datasetnum := 0;
       lowparm := 0.5; highparm := 1.5; percentneg:= 0;
       standardassumpts:= true;
       answer:= 'n'; 
       yes:= 'y'; no:= 'n';
       havevalue := [];       
       novalueyet := graphvertices;
       if not minternal then datatypes:= [];
       datatype:= sampcor;
       tiers := nil;
999:  end;  (* monteinit *)

procedure datainit;
    var i,j : integer;
    begin
      for j:= 1 to maxvertex do
       begin
        smeans[j]:= 0.0;
        for i:=1 to maxvertex do scov^[i,j]:= 0.0;
       end;
     end;  (* datainit  *)

(* End of Initializing Routines *)
(********************************)

(***************************************)
(* Forming the tiers of dependencies *)

  (* this finds all the direct causes of a vertex i
     and stores it in the global array causes[i],
     each cell of which is a vertexset *)
 procedure getcauses(* vert : vertextype; var graph: adjtype;
                       var causes: relationsets *);
    var q : vertextype;
    begin
      causes[vert] := [];
      for q := 1 to origmax do
       if q in graphvertices then  (* c *)
         if (graph^[q,vert] <> notthere) then 
               causes[vert] := causes[vert] + [q];
    end; (* getcauses *)


  (* this finds all the direct effects of a vertex i
     and stores it in the global variable effects[i],
     which is a vertexset *)
 procedure geteffects(vert : integer; var graph: adjtype);
    var q : integer;
    begin
      effects[vert] := [];
      for q := 1 to origmax do
       if q in graphvertices then  (* c *)
         if (graph^[vert,q] <> notthere) then 
               effects[vert] := effects[vert] + [q];
    end; (* geteffects *)


  (* this forms an ordered list of variables such that no member
     of the list depends directly on any variable later in the list.
     this ensures that when we go to fill in values for an effect, all
     of its causes are already calculated. *)
 procedure formeqsorder(* var cureqs: eqorder; var graph: adjtype *);
     var   i : vertextype;
           frontptr: eqorder;
     begin (* formeqsorder *)
       frontptr:= cureqs;
       havevalue := [];             
       novalueyet := graphvertices;
       repeat
          for i:= 1 to origmax do
          if i in graphvertices then  (* c *)
           begin
            getcauses(i,graph,causes);
            if (i in novalueyet) then
              if (causes[i] <= havevalue) then
                 begin
                    havevalue := havevalue + [i];
                    novalueyet := novalueyet - [i];
                    cureqs^.varbl := i;
                    new(cureqs^.next);
                    cureqs := cureqs^.next;
                    cureqs^.next := nil;
                    cureqs^.varbl := 0;
                 end;
           end;
        until (novalueyet = []);
        cureqs := frontptr;
     end; (* formeqsorder *) 


(* End of Forming the tiers of dependencies *)           
(********************************************)

(***********************************)
(*  Getting the User's Preferences *)

procedure getoutfilename(var outfilename: inbuf);
  var changed : boolean;
      curfilename, numstr: inbuf;
  begin
    if (numdatasets > 1) then
      begin
         twriteln(output,'Enter a filename with no extension for the|');
         twriteln(output,'files that will contain the data.|');
       end
    else
        begin
            twriteln(output,'Enter a name for the file that will contain|');
            twriteln(output,'the data TETRAD is about to generate.|');
        end;
    twriteeoln(output);
    outfilename := 'montedat';
    outfilename[10] := chr(0);
    getfilename('Filename: |',outfilename, changed);
    twriteeoln(output);    
    if not aaabort then
    begin
    if (numdatasets > 1) then
      begin
         twrite(output,'TETRAD will create |');
	 twriteint(output, numdatasets);
	 twriteln(output, ' data files called|');
         putstr(outfilename, true); write('.1 through ');
         tconcat(outfilename,'.',curfilename);         
         tstr(numdatasets,numstr);
         tconcat(curfilename,numstr,curfilename);         
         putstr(curfilename, true);
         twriteeoln(output);
      end
    else
      begin
         twrite(output,'TETRAD will create a file called |');
         putstr(outfilename, true);
         twriteeoln(output);
      end;
    twriteeoln(output);
    end;
  end;  (* getoutfilename *)

(* This forms the name of the outfile, opens it, and writes a comment which
   specifies the generating model, be it linear or discrete. *)
procedure startoutfile(datasetnum: integer; outfilename: inbuf; var curout: text; var settoblank: boolean);
  label 999;
  var numstr: inbuf;
  begin
    settoblank := false;
    if numdatasets > 1 then 
       begin
         tstr(datasetnum,numstr);
         tconcat(outfilename,'.',outfilename);
         tconcat(outfilename,numstr,outfilename);
        end;
    rewritefile(curout, outfilename);
    if (outfilename <> blankfilename) and (outfilename <> nullfilename) then writecomment(curout)
    else settoblank := true;
    if (fatal in errors) then goto 999;
999:  end;


procedure getdatatypes(var datatypes: numberset);
  var maxnum: integer;
  begin
    if undercontinuous then maxnum:= 6 else maxnum:= 3;
    repeat 
      twriteeoln(output);
      twriteln(output,'1 = Raw sample data|');
      twriteln(output,'2 = Sample covariance matrix|');
      twriteln(output,'3 = Sample correlation matrix|');
      if undercontinuous then
        begin
          twriteln(output,'4 = Population covariance matrix|');
          twriteln(output,'5 = Population correlation matrix|');
          (* twriteln(output,'6 = Implied Covariances Symbolically|'); *)
        end;
      twriteeoln(output);
      twriteln(output,'List the data types you want below.|');
      twriteln(output,'Separate each number by a space, and|');
      twrite(output,'use no delimiters. List =  |');
      getnumbers(datatypes, maxnum);
      if datatypes = [] then twriteln(output,'Please enter a number. |');
    until (aaabort or (datatypes <> []));
  end;
    

procedure getuserprefs;
   begin
     if (datatypes * [1,2,3] <> []) then  (* want sample data *)
       begin
         twriteeoln(output);
         numdatasets:= tround(getnumber(1, 100, true, 
               'Number of data sets: |                    ', numdatasets));
         montesamp:= tround (getnumber(2, maxmontesamp, true, 
                'Sample size: |                    ', montesamp));
       end
     else montesamp:= maxmontesamp; (* population sample size for statistics *)
     getoutfilename(outfilename);
   end;  (* getuserprefs *)

(* End of Getting the User's Preferences *)
(*****************************************)

 (************************************************)
 (* Getting the parameters for discrete data  *)


procedure getcatval(i,j: integer; var cats: categoryarray);
  var k: integer;
      noprob: boolean;
  begin
     repeat
        cats[i].value[j]:= getnumber(-99, 99, true,  
                                     ' Value: |         ', tround(cats[i].value[j]));
        noprob:= true;
        for k:= 1 to j-1 do
	   if tround(cats[i].value[j]) = tround(cats[i].value[k]) then
	      begin
                twrite(output,'Enter a new value. This value |');
                twriteln(output,'has already been used.|');
                noprob:= false;
              end;                         
     until noprob;
  end;
     
procedure getcategories(*var cats: categoryarray*);     
  var i,j,k: integer;
      varset: numberset;
  begin
    for i:= 1 to origmax do
      if i in graphvertices then  (* c *)
	begin
	twriteint(output,i); twrite(output,') |');
	writename(i,output);twriteeoln(output);
	end;
    twriteln(output,'All variables default as binary valued (0,1).|');
    twriteln(output,'Enter the numbers of any you wish to change.|');
    twrite(output,'Numbers = |');
    getnumbers(varset, origmax);
    for i:= 1 to origmax do
      if i in varset then
       begin
         twrite(output,'Variable is |');
	 writename(i,output); twriteeoln(output);
         cats[i].num:= tround(getnumber(2, maxcategories, true,  
              'Number of categories: | ', 2));
	 twriteln(output,'Non-integers will be rounded.|');
         for j:= 1 to cats[i].num do
           begin
             twrite(output,'Category |');
	     tpwriteint(output,j,4,false);
             getcatval(i,j,cats);
           end;
       end;
  end;  (* getcqategories *)

procedure sort(var nums: probarray; numcats: categorynum);
  var temp: probarray;
      i,j: integer;
  begin
   temp:= nums;
   for i:= 1 to numcats do
     begin
        nums[i]:= 1;
        for j:= 1 to numcats do
          if (temp[j] > nums[i-1]) and (temp[j] <= nums[i]) then 
             nums[i]:= temp[j];
     end;
  end;


procedure getcutoffs(num: categorynum; var cutoffs: probarray);
  var i: integer; 
      cand: real;
  begin
    for i:= 0 to maxcategories do cutoffs[i]:= 0;
    for i:= 1 to (num-1) do 
       begin
         repeat 
            cand:= randunif(0, 1);
         until ((cand >= (0.0 + minprob)) and (cand <= (1.0 - minprob)));
         cutoffs[i]:= cand;
       end;
    sort(cutoffs,num);
    cutoffs[num]:=1.00;
  end; (* getcutoffs *)


function getnextparent(parents: vertexset): vertextype;
  var i: vertextype;
      found: boolean;
  begin
    i:= 1;
    found:= false;
    while (not found) and (i <= origmax) do
      begin
       found:= (i in parents);
       i:= i+1;
      end;
    getnextparent:= i-1;
  end;


(************************************************)
(* Reading in the Bayesian Network from a file *)


function getnextvalue:real;
  begin
   readtochar('=');
   getnextvalue:= readnum(false, ignoreall);
  end;

procedure readcutoffs(vertex:vertextype; numcats: categorynum; 
                      var cutoffs: probarray; var infile: text);
  var i: integer; undef: boolean;
	value:real;
	numcat:catrange;
  begin
    readfileline(infile,buf);
    for i:= 0 to maxcategories do cutoffs[i]:= 0;
    replacecharwithspace(')');
    for i:= 1 to numcats do
	begin
	  value := getnextvalue;
	  numcat := getnumcat(vertex,value,cats);
          cutoffs[numcat]:= getnextnum(infile,numcats);
          readwhitespace;
          if myeoln and (i < numcats ) 
	  then begin readfileline(infile,buf); replacecharwithspace(')'); end;
	end;
    (* make the cutoffs into cdf *)
    for i:= 1 to numcats do
       cutoffs[i]:= cutoffs[i] + cutoffs[i - 1];

    if (cutoffs[numcats] < 0.99) or
       (cutoffs[numcats] > 1.01) then 
       myerror('Probabilities must sum to 1',fatal,true);
    cutoffs[numcats]:=1.00;
  end;


procedure filloutcondprob(var cur: probptr; effect: vertextype; 
                           parentlist:eqorder; var infile: text; 
                           var cats: categoryarray);
   var i: integer;
       nextparent: vertextype;
       value:real; numcat:catrange;
   begin
     if parentlist^.next = nil then
       begin
         cur^.varnum:= effect;
         getcutoffs(cats[effect].num, cur^.cutoff)
       end
     else
      begin
       nextparent := parentlist^.varbl;
       cur^.varnum:= nextparent;
       parentlist:= parentlist^.next;
       for numcat:= 1 to cats[nextparent].num do
         begin
           new(cur^.nextvalue[numcat]);
           for i:= 1 to maxcategories do 
              cur^.nextvalue[numcat]^.nextvalue[i]:= nil;
           filloutcondprob(cur^.nextvalue[numcat],effect,parentlist,
                           infile,cats);
         end;
      end;
   end;  (* filloutcondprob *)

(* this is only called when there are parents *)
procedure readcondprob(var cur: probptr; effect: vertextype; 
                       var infile: text; var cats: categoryarray);
 var i: integer;
     lcur: probptr;
     parname: vertexnametype;
     nextparent: vertextype;
     value:real; numcat:catrange;
     wasinserted:boolean;
 begin
  lcur:= cur;
  repeat
    readfileline(infile,buf);
    if not (isdashedline or eof(infile)) then
      begin
        readtochar('n'); 
        while not myeoln do
         begin
           replacecharwithspace('=');
           readwhitespace;
           readname(parname,ignoreall);        
           cur^.varnum:= listposition(parname,false,vernames,wasinserted);
           value := readnum(false,fatal);
           numcat := getnumcat(cur^.varnum,value,cats);
           if cur^.nextvalue[numcat] = nil then
             begin
              new(cur^.nextvalue[numcat]);
              for i:= 1 to maxcategories do 
                cur^.nextvalue[numcat]^.nextvalue[i]:= nil;
             end;
           cur:= cur^.nextvalue[numcat];
         end;
        cur^.varnum:= effect;
        readcutoffs(effect,cats[effect].num, cur^.cutoff,infile);
      end;
    cur:= lcur;
   until isdashedline or eof(infile);
 end;  (* readcondprob *)


procedure eraseprevparentlist(var parentlist: eqorder);
  var temp: eqorder;
  begin
    temp:= parentlist;
    while (temp^.next <> nil) do
      begin 
        temp:= temp^.next;
        dispose(parentlist);
        parentlist:= temp;
      end;
  end; (* eraseprevparentlist *)        

procedure push(temp: eqorder; var parentlist:eqorder);
  begin
    temp^.next:= parentlist;
    parentlist:= temp;
  end;

procedure readparents(var parentlist: eqorder);
  var curvarname: vertexnametype;
      temp,curpos: eqorder; wasinserted: boolean;
  begin
     readtochar(':'); 
     readwhitespace;
     new(parentlist);
     parentlist^.next := nil;
     parentlist^.varbl := 0;
     curpos := parentlist;
     while (not myeoln) do
       begin
          readname(curvarname, ignoreall);
          new(temp); 
          temp^.varbl := listposition(curvarname,false,vernames,wasinserted);
          if curpos^.varbl = 0 then
             begin
                temp^.next := parentlist;
                parentlist := temp;
		curpos := parentlist;
             end
          else 
            begin
	      temp^.next := curpos^.next;
	      curpos^.next := temp;
	      curpos := temp;
            end;
          readwhitespace;
       end;
  end; (* readparents *)
     
procedure calcparents(var parentlist: eqorder; effect: vertextype);
  var i: vertextype;
      temp: eqorder;
  begin
    for i:= origmax downto 1 do
     if i in graphvertices then  (* c *)
      if i in causes[effect] then
        begin
          new(temp);
          temp^.varbl := i;
          push(temp,parentlist);
        end;
  end;

   (* reads the part of the network file that specifies 
      the variables and the values they take on *)
procedure readcategories(*var cats: categoryarray; 
                         var eqs: eqorder; var bfin: text*);
  var curfilevar: vertextype;
      curvarname: vertexnametype;
      i: integer;
      keyword: wordtype;
      temp: eqorder; wasinserted: boolean;
  begin
   keyword:= 'Categories|';
   repeat readfileline(bfin,buf); until iswordonline(keyword);
   temp:= eqs;
   while (eqs^.next <> nil) do
      begin
        readfileline(bfin,buf);
        readwhitespace;
        readname(curvarname,ignoreall);        
        curfilevar:= listposition(curvarname,false,vernames,wasinserted);
        cats[curfilevar].num:= tround(readnum1(bfin,false,ignoreall));
        for i:= 1 to cats[curfilevar].num do
           cats[curfilevar].value[i]:= tround(readnum1(bfin,false,ignoreall));
        eqs:= eqs^.next;
      end;
   eqs:= temp;
  end; (* readcategories *)


  (*  The while loop for cureqs is to keep track of how many variables
      we have to read from the bayesnet file or how many we have to
      fix some distribution for when fixed is false. We get the effect
      differently depending on fixed, because if the graph file is switched
      between creation of the bayesnet file and reading it in, then the 
      mapping of vertexname to vertextype in vernames can be different 
      than the one in cureqs. *)
procedure getbayesparms(*var cats: categoryarray; cureqs: eqorder;
                        fixed: boolean; var infile: text*);
  var effect: vertextype;
      word: wordtype; i: integer;
      curvarname: vertexnametype;
      parentlist: eqorder; wasinserted: boolean;
  begin
    new(parentlist); parentlist^.next:= nil;
    while (cureqs^.next <> nil) do
      begin
        if (not fixed) then effect:= cureqs^.varbl
        else
          begin
            while (not isdashedline) do readfileline(infile,buf);
            readfileline(infile,buf); 
            readwhitespace; 
            readname(curvarname,ignoreall); 
            effect:= listposition(curvarname,false,vernames,wasinserted);
          end;
        eraseprevparentlist(parentlist);
        if fixed then readparents(parentlist)
         else calcparents(parentlist,effect);
        new(cats[effect].condprob);
        for i:= 1 to maxcategories do 
          cats[effect].condprob^.nextvalue[i]:= nil;
	if fixed then 
          if parentlist^.next = nil then
            begin
              cats[effect].condprob^.varnum:= effect;
              readcutoffs(effect,cats[effect].num, 
                          cats[effect].condprob^.cutoff,infile);
            end
          else readcondprob(cats[effect].condprob,effect,infile,cats)
        else filloutcondprob(cats[effect].condprob,effect,parentlist,
                             infile,cats);
        cureqs:= cureqs^.next;
      end;
  end;  (* getbayesparms *)


procedure removetree(var point: probptr);
var i: integer;
begin
   i := 1;
   while (point^.nextvalue[i] <> nil) and (i <= maxcategories) do
      begin
        removetree(point^.nextvalue[i]);
        dispose(point^.nextvalue[i]);
        point^.nextvalue[i] := nil;
        i := i + 1;
      end;
end;

procedure zremovetree(var point: probptr; v: vertextype);
var i: integer; stop: boolean;
begin
   i := 1;
   stop := false;
   while (not stop) do 
    if v <> point^.varnum then
      begin
        zremovetree(point^.nextvalue[i],v);
        dispose(point^.nextvalue[i]);
        point^.nextvalue[i] := nil;
        i := i + 1;
        if i > maxcategories then stop := true
        else if point^.nextvalue[i] = nil then stop := true;
      end
    else stop := true;
end;

procedure zdisposebayesnet(* var cats: categoryarray *);
var i: vertextype; stop: boolean;
begin
   i := 1;
   stop := false;
   while (not stop) do 
     begin
        if (i in graphvertices) then 
           begin
              zremovetree(cats[i].condprob,i);
              dispose(cats[i].condprob);
           end;
        i := i + 1;
        if (i > maxcategories) then stop := true 
        else if (cats[i].condprob = nil) then stop := true;
      end;
end;

procedure disposebayesnet(*var cats: categoryarray*);
var i: vertextype; stop: boolean;
begin
   i := 1;
   stop := false;
   while (not stop) do
     begin
        if (i in graphvertices) then 
           if cats[i].condprob <> nil then removetree(cats[i].condprob);
        i := i + 1;
        if (i > maxcategories) then stop := true 
        else if (cats[i].condprob = nil) then stop := true;
      end
end;

procedure rearrange(*var localadj,adjacency: adjtype; 
                    var localnames,vernames: namearray*);
var i,j,nextob,nexttheo: integer; stop: boolean; 
    rearrange: array[vertextype] of vertextype;
begin
     for i := 1 to maxvertex do 
          begin
             rearrange[i] := i;
             vernames[i] := blankname;
             for j := 1 to maxvertex do adjacency^[i,j] := notthere;
           end;
     i := 1;
     nextob := 0;
     stop := false;
     while (localnames[i] <> blankname) and (i <= maxvertex) 
            and (not stop) do
        begin
           if not istheoretical(localnames[i]) then
              begin
                   nextob := nextob + 1;
                   rearrange[i] := nextob;
              end;
           if i < maxvertex then i := i + 1
           else stop := true;
        end;
     i := 1;
     nexttheo := nextob + 1;
     stop := false;
     while (localnames[i] <> blankname) and (i <= maxvertex) 
           and (not stop) do
        begin
           if istheoretical(localnames[i]) then 
              begin
                  rearrange[i] := nexttheo;
                  nexttheo := nexttheo + 1;
              end;
           if i < maxvertex then i := i + 1
           else stop := true;
        end;
     for i := 1 to maxvertex do
         begin
             vernames[rearrange[i]] := localnames[i];
             for j := 1 to maxvertex do
                adjacency^[rearrange[i],rearrange[j]] := localadj^[i,j];
         end;
end;

(* procedure READBAYESNET is now in sys.p *)
(*  End of Reading in the Bayesian Network from a file *)
(*******************************************************)



(* End of Getting the parameters for discrete data  *)
(****************************************************)

(********************************************)
(* Getting raw data *)

procedure geterrors(var error: vertarray);
    var
      i: vertextype;
    begin
       for i:=1 to max do
        if i in graphvertices then
         begin
           error[i] := 0;
           case exogd[i].probtype of
               1: error[i] := randunif(exogd[i].parm1, exogd[i].parm2);
               2: error[i] := randnorm(exogd[i].parm1, exogd[i].parm2);
               3: error[i] := randlognorm(exogd[i].parm1, exogd[i].parm2);
           end; (* case *)
         end; (* i = max *)
    end; (* geterrors *)

  (* this takes an array of values for one unit in the sample and
     updates the covariance matrices accordingly *)
procedure updatecovs(var varval: vertarray);
   var i,j: integer;
   begin
    for i := 1 to numob do 
      if i in graphvertices then
       begin
         smeans[i]:= smeans[i] + varval[i];
         for j:= i to numob do
          if j in graphvertices then
           begin
             scov^[i,j]:= scov^[i,j] + (varval[i] * varval[j]);
             scov^[j,i]:= scov^[i,j];
            end; (* j = i to numob *)
       end; (* i = 1 to numob *)
   end; (* updatecovs *)
   
  (* after all members of the sample have been generated, this
     converts sums to expectations. *)
procedure finishcovs;
   var i,j:integer;
   begin
     for i:= 1 to numob do
      if i in graphvertices then
       smeans[i]:= smeans[i]/montesamp;
     for i:= 1 to numob do
      if i in graphvertices then
       for j:= i to numob do
        if j in graphvertices then
         begin
           scov^[i,j]:= scov^[i,j]/montesamp - smeans[i]*smeans[j];
           scov^[j,i]:= scov^[i,j];
         end;
   end; (* finishcovs *)

(* GETNUMCAT get the category number for vertex VERT with value VALUE *)
function getnumcat(*vert: vertextype; value: real; var cats:categoryarray): integer;  *);
   var i: integer; found: boolean;
   begin
     found:= false;
     i:= 0;
     while (not found) and (i < maxcategories) do
       begin
         i:= i+1;
         if cats[vert].value[i] = value then found:= true;
       end;
     getnumcat:= i;
   end;

  (* I use thispar because the index for varval must be an integer,
     where as cur^.varnum is a vertextype. *)
procedure findcondprob(*var cur: probptr; effect: vertextype; 
                        var cutoffs: probarray; 
                        var cats:categoryarray; var varval: vertarray); *);
  var numcat,thispar: integer;
  begin
    thispar:= cur^.varnum;
    if (thispar = effect) then cutoffs:= cur^.cutoff
    else 
     begin
       numcat:= getnumcat(cur^.varnum,varval[thispar],cats);
       cur:= cur^.nextvalue[numcat];
       findcondprob(cur,effect,cutoffs,cats,varval);
     end;
  end;

procedure getdiscreteval(* effect: vertextype; var varval: vertarray;
                         var cats: categoryarray *);
    var  cur: probptr;
         cutoffs: probarray;
         found: boolean;
         i: integer;
         randnum:real;
    begin
      cur:= cats[effect].condprob;
      findcondprob(cur,effect,cutoffs,cats, varval);
      randnum:= randunif(0,1);
      found:= false;
      i:= 1;
      while (not found) and (i <= cats[effect].num) do
       begin
         if (randnum <= cutoffs[i]) then 
            begin
              varval[effect]:= cats[effect].value[i];
              found:= true;
            end;
         i:= i+1;
       end;
     if i>cats[effect].num then 
     	varval[effect]:=cats[effect].value[cats[effect].num];
    end;

 (*  main control routine for generating data. *)
procedure getdata(fromgibs: boolean);
   var  
      error,varval: vertarray;  
      cas : integer;
   
   procedure fillincase(var curgraph: adjtype; var curparm: covarray; 
                         var cureqs: eqorder; var cats: categoryarray);
      var i, effect, cause: integer;
          frontptr: eqorder; 
      begin   (* fillincase *)    
         for i := 1 to maxvertex do
             varval[i] := 0.0;
         frontptr:= cureqs;
         while (cureqs^.next <> nil) do
           begin
              effect := cureqs^.varbl;
              if undercontinuous then
                begin
                  varval[effect] := error[effect];
                  for cause := 1 to origmax do
                    if curgraph^[cause, effect] <> notthere then 
                        varval[effect] := varval[effect] +
                                    curparm^[cause,effect] * varval[cause];
                end
              else getdiscreteval(effect,varval,cats);
              cureqs := cureqs^.next;
           end;
         cureqs := frontptr;
         if 1 in datatypes then writerawcas(varval,curout);
         updatecovs(varval);
      end;  (* fillincase *)

   begin (* getdata *)
     if 1 in datatypes then (* want raw data *) writerawheading(curout);
     for cas := 1 to montesamp do
       begin
         if undercontinuous then geterrors(error);
         fillincase(adjacency,parm,varorder,cats);
       end;
     finishcovs;
     if not fromgibs then flush(curout);
   end;  (* getdata *)

(* End of getting raw data *)
(********************************************)

(********************************************)
(* forming the population covariance matrix *)

procedure formtiers(var tiers: tiersptr);
  var i: integer;
      tiers1: tiersptr;
      covered, notcovered: vertexset;
      tierempty: boolean;
  begin
    covered:= independent;
    notcovered:= graphvertices - independent;
    new(tiers);
    tiers^.next:= nil;
    tiers^.varset:= independent;
    tiers1:= tiers;
    repeat
      tierempty:= true;
      for i:= 1 to origmax do
        if (i in graphvertices) and not(i in covered) and (causes[i] <= covered) then
         begin
           if tierempty then
              begin
                 new(tiers^.next);
                 tiers := tiers^.next;
                 tiers^.next := nil;
                 tiers^.varset := [];
                 tierempty:= false;
              end;
           tiers^.varset:= tiers^.varset + [i];
         end;
      covered:= covered + tiers^.varset;
      notcovered:= notcovered - tiers^.varset;
    until (notcovered = []);
    tiers:= tiers1;
  end;

function formtruevar(x: vertextype; covered: vertexset;
                     var impcov: covarray): real;
var i,j: vertextype; variance: real;
begin
   if minternal then variance := 1.0 else
   if x in independent then 
     case exogd[x].probtype of
       1: variance := (exogd[x].parm2 - exogd[x].parm1)*
                      (exogd[x].parm2 - exogd[x].parm1) / 12;
       2: variance := exogd[x].parm2;
       3: variance := ln(exogd[x].parm2);
     end
   else
     begin
       variance := 0;
       for i := 1 to origmax do
        if (i in covered*graphvertices) then
          variance := variance + parm^[i,x] * parm^[i,x] * impcov^[i,i];
       for i := 1 to origmax - 1 do 
        if (i in covered*graphvertices) then
          for j := i + 1 to origmax do
           if (j in covered*graphvertices) then 
             variance := variance + 
                            2 * parm^[i,x] * parm^[j,x] * impcov^[i,j];
       case exogd[x].probtype of
        1: variance := variance + (exogd[x].parm2 - exogd[x].parm1) / 12;
        2: variance := variance + exogd[x].parm2;
        3: variance := variance + ln(exogd[x].parm2);
       end (* case *)
     end; (* else *)
   formtruevar := variance;
end;


 (* this calculates the product of all the edge coefficients
    and the source's variance along one trek between two vars. *)
function covfrom1trek(trek: treksptr; var impcov: covarray): real;
  var cov1: real; 
      source: vertextype;
  begin
   source:= trek^.source;   
   cov1:= impcov^[source,source];
   if (trek^.trek1 <> nil) then cov1:= cov1 * trek^.trek1^.strength;
   if (trek^.trek2 <> nil) then cov1:= cov1 * trek^.trek2^.strength;
   covfrom1trek:= cov1;
  end;

function formtruecov(x,y: vertextype; var impcov: covarray): real;
var cov: real;
    trek: treksptr;
begin
   cov := 0;
   trek:= treks1^[x,y];
   while (trek <> nil) do
     begin
      cov:= cov + covfrom1trek(trek,impcov);
      trek:= trek^.next;
     end;
   formtruecov := cov;
  
end;

procedure cover(verts, covered: vertexset; var impcov: covarray);
  var i,j:integer;
  begin
    for i:= 1 to origmax do  (* calc the new variances *)
      if (i in verts*graphvertices) then 
        impcov^[i,i]:= formtruevar(i,covered,impcov);
    for i:= 1 to origmax do  (* calc the new covariances *)
      if (i in verts*graphvertices) then 
        for j:= 1 to origmax do
           if (i <> j) and 
              ((j in verts) or (j in covered)) and 
              (impcov^[i,j] = 0.0) then 
              begin 
                impcov^[i,j]:= formtruecov(i,j,impcov);
                impcov^[j,i]:= truecov^[i,j];
              end;
  end; (* cover *)

procedure formtrue(*var impcov: covarray; needtiers,needtreks: boolean*);
  var i,j: integer;
      tiers1 : tiersptr;
      nextset,covered: vertexset;
      loctime: double;
  begin
   for i := 1 to maxvertex do 
     for j := 1 to maxvertex do impcov^[i,j] := 0;
     if needtreks then
       begin
        loctime:= tgettime;
        inittreks;
        calcpath(adjacency,paths1,1);
	calctrek(treks1,paths1,1);
        trektime:= trektime + (tgettime - loctime);
       end;
   if needtiers then formtiers(tiers);
   tiers1 := tiers;
   covered:= [];
   while (tiers1 <> nil) do
     begin
       nextset := tiers1^.varset;
       cover(nextset,covered,impcov);
       covered:= covered + nextset;
       tiers1 := tiers1^.next;
     end;
  end; 

(* End of forming the population covariance matrix *)
(***************************************************)

(********************************************)
(* Main control for Monte Carlo Generator *)

procedure domonte;
label 998,999;
var tbuf:inbuf; settoblank: boolean;
 begin  (* domonte *)
   scramble; 
   monteinit;
   if isacycle then 
         myerror('Monte cannot generate cyclic models.|',fatal,false);
   if fatal in errors then goto 999;
   if prnton then twriteln(output,'The TETRAD Monte Carlo Generator|');
   twriteeoln(output);;
   initrand(newseed);
   undercontinuous:= havereadlinearmodel;
   if (not minternal) then getdatatypes(datatypes);
   if not wantexit then
     begin
      parmsfixed := true;
      if (not minternal) then getuserprefs;
      if (outfilename = blankfilename) or (outfilename = nullfilename) and 
         (not minternal) then 
           begin
              myerror('No output file.|',fatal,false);
              goto 998;
           end;
      new(varorder);
      formeqsorder(varorder,adjacency);
(*      if undercontinuous and prnton then 
        writelinearmodel(varorder,output,adjacency,parm); *)
      if (datatypes * [1,2,3] <> []) then  (* we want sample data *)
        repeat
          datainit;
          initrand(newseed);
          datasetnum:= datasetnum + 1;
          twriteeoln(output);
          twrite(output,'Starting dataset: |');
	  twriteint(output,datasetnum); twriteeoln(output);
          startoutfile(datasetnum,outfilename,curout,settoblank);
          if settoblank then
              begin
                 myerror('No output file.|',fatal,false);
                 goto 998;
              end;

          if (card(datatypes) = 1) then
               if (1 in datatypes) then 
                    if undercontinuous then 
                          twriteln(curout,'/Continuousraw|') 
                    else twriteln(curout,'/Raw|') 
               else twriteln(curout,'/Covariance|');
          twriteint(curout, montesamp);
	  twriteeoln(curout);
          getdata(false);
          if (sampcor in datatypes) then makecorrmat(numob, scov, scor);
          if (datatypes * [popcov,popcor] <> []) 
              then formtrue(truecov,true,false);
          if (popcor in datatypes) then makecorrmat(numob,truecov,truecor);
          if ([2,3,4,5] * datatypes <> []) then writecormatrices(curout);
	  finishfile(curout,tbuf);
        until (wantexit or (datasetnum = numdatasets))
      else if (datatypes * [4,5] <> []) then (* only want population data *)
            begin
              if (not minternal) then 
                 begin
                   startoutfile(datasetnum,outfilename,curout,settoblank);
                   if settoblank then
                     begin
                        myerror('No output file.|',fatal,false);
                        goto 998;
                     end;
                   if (card(datatypes) = 1) and (1 in datatypes) then
                     twriteln(curout,'/Raw|') 
                   else twriteln(curout,'/Covariance|');
                   twriteint(curout, montesamp); twriteeoln(curout);
                 end;
              formtrue(truecov,true,false);
              if (popcor in datatypes) then makecorrmat(numob,truecov, truecor);
              if (not minternal) then writecormatrices(curout);
	      if (not minternal) then finishfile(curout,tbuf);
            end;
      disposepointrs(tiers,varorder);
      twriteeoln(output); twriteln(output,'Done |');
     end; (* if not wantexit *)
998:   dispose(scov);
(*   if (not minternal) then dispose(truecov); *)
   dispose(scor);
   if (not minternal) then dispose(truecor);
999: end; (* domonte *)

procedure fillcondprob(var cur: probptr; effect: vertextype; 
                       parents: vertexset; var cats: categoryarray);
   var i,j: integer;
       nextparent: vertextype;
   begin
     if parents = [] then cur^.varnum:= effect
     else
      begin
       nextparent:= getnextparent(parents);
       cur^.varnum:= nextparent;
       parents:= parents - [nextparent];
       for i:= 1 to cats[nextparent].num do
         begin
           new(cur^.nextvalue[i]);
	   for j := 1 to maxcategories do 
              begin
                 cur^.nextvalue[i]^.multiplier[j]:= -1;
                 cur^.nextvalue[i]^.count[j] := 0;
                 cur^.nextvalue[i]^.cutoff[j]:= 0;
                 cur^.nextvalue[i]^.nextvalue[j]:= nil;
              end;
           fillcondprob(cur^.nextvalue[i],effect,parents,cats);
         end;
      end;
   end;

procedure formbranch(* var adjacency: adjtype; var category: categoryarray;
                     i: vertextype *);
label 999;
var j: vertextype; parents: vertexset;
begin
     if category[i].condprob <> nil then 
         begin
            zremovetree(category[i].condprob,i);
            dispose(category[i].condprob);
            category[i].condprob := nil;
         end;
     category[i].num := maxval[i];
     for j:= 1 to maxval[i] do
	category[i].value[j] := varcatvalue[i,j];  
     getcauses(i,adjacency,causes);
     parents := causes[i];
     new(category[i].condprob);
     if category[i].condprob = nil then
        begin
           myerror('Out of memory. |',fatal,false);
           goto 999;
        end;
     for j:= 1 to maxcategories do 
         begin
            category[i].condprob^.cutoff[j]:= 0;
            category[i].condprob^.nextvalue[j] := nil;
            category[i].condprob^.count[j] := 0;
         end;
     fillcondprob(category[i].condprob,i,parents,category);
999: end;

procedure disposeeqsorder(* var eqs: eqorder *);
var curpos,nextpos: eqorder;
begin
  curpos := eqs;
  repeat
      nextpos := curpos^.next;
      dispose(curpos);
      curpos := nextpos;
  until curpos = nil;
end;

procedure cleartree(var category: categoryarray; root: probptr);
var curpos: probptr; i: integer;
begin
   curpos := root;
   if curpos <> nil then 
      begin
         if curpos^.nextvalue[1] = nil then
             for i := 1 to maxcategories do curpos^.cutoff[i] := 0;
         i := 1;
         while (i <= maxcategories) and (curpos^.nextvalue[i] <> nil) do
           begin
             cleartree(category,curpos^.nextvalue[i]);
             i := i + 1;
           end;
      end;
end;
         
procedure clearcats;
var i: integer;
begin
   for i := 1 to max do cleartree(cats,cats[i].condprob);
end;

procedure formtree(* var adjacency: adjtype; var category: categoryarray *);
var i: vertextype; eqs,curpos: eqorder;
begin
(*   new(eqs);
   formeqsorder(eqs,adjacency); 
   curpos := eqs; *)
   for i := 1 to numobgraph do 
          formbranch(adjacency,category,graphlist[i]);
(*   disposeeqsorder(eqs); *)
end;
   
procedure insertdata(vertex: vertextype; var category: probptr; 
                     var data: vertexarray; var cats: categoryarray);
var curvar:vertextype;
begin
   curvar := category^.varnum;
   if curvar = vertex then
      begin
         category^.cutoff[data[vertex]] := category^.cutoff[data[vertex]] + 1;
         category^.count[data[vertex]] := category^.count[data[vertex]] + 1;
      end
   else insertdata(vertex,category^.nextvalue[data[curvar]],
	                data,cats);
end;      

procedure converttree(vertex: vertextype; var category: probptr;
                      var cats: categoryarray; var estmethod:esttype; 
                      var precision:integer);
var j,total,numcat: integer; temp: probarray;
begin
  if category^.varnum = vertex then
     begin
         total := 0;
         numcat:=cats[vertex].num;
         for j := 1 to numcat do 
             total := total + tround(category^.cutoff[j]);
         if ((total = 0) and (estmethod=MLexternal)) then
            begin
		for j := 0 to cats[vertex].num do category^.cutoff[j] := -1 * j;
            end
         else
            begin
              if ((estmethod = MLinternal) or (estmethod=MLexternal)) then
                for j := 1 to numcat do
		  temp[j] := category^.cutoff[j] / total;
              if estmethod = uniform then
                for j := 1 to numcat do
                  temp[j]:=(category^.cutoff[j]+(precision/numcat))/(total+precision);
              category^.cutoff[0] := 0.0;
              category^.cutoff[cats[vertex].num] := 1.0;
              for j := 1 to numcat - 1 do
                 category^.cutoff[j] := category^.cutoff[j-1] + temp[j];
            end;
      end
  else for j := 1 to cats[category^.varnum].num do 
      converttree(vertex,category^.nextvalue[j],cats,estmethod,precision);
end;      

(* this implements Laplaces rule *)
procedure newconverttree(vertex: vertextype; var category: probptr;
                      var cats: categoryarray; var estmethod:esttype; 
                      var precision:integer);
var j,total,numcat: integer; temp: probarray;
begin
  if category^.varnum = vertex then
     begin
         total := 0;
         numcat:=cats[vertex].num;
         for j := 1 to numcat do 
             total := total + tround(category^.cutoff[j]);
         if ((estmethod = MLinternal) or (estmethod=MLexternal)) then
         for j := 1 to numcat do
		  temp[j] := (category^.cutoff[j] + 1) / (total + cats[vertex].num);
         if estmethod = uniform then
             for j := 1 to numcat do
                 temp[j]:=(category^.cutoff[j]+(precision/numcat))/(total+precision);
         category^.cutoff[0] := 0.0;
         category^.cutoff[cats[vertex].num] := 1.0;
         for j := 1 to numcat - 1 do
              category^.cutoff[j] := category^.cutoff[j-1] + temp[j];
      end
  else for j := 1 to cats[category^.varnum].num do 
      newconverttree(vertex,category^.nextvalue[j],cats,estmethod,precision);
end;      

procedure fillouttree(* var category: categoryarray;estmethod:esttype;precision:integer *);
var i: integer; j: vertextype; datapoint:vertexarray; temp: catrange; allknown: boolean;
begin
   effsampsize := 0;
   for i := 1 to tround(sampsize) do 
     begin
        allknown := true;
	for j:= 1 to numobgraph do
          if graphlist[j] in graphvertices then
              begin
                 temp := getbigdata(i,graphlist[j]);
                 datapoint[graphlist[j]]:= temp;
                 if (datapoint[graphlist[j]] = 0) then allknown := false;
              end;
        if allknown then
	   begin
	      effsampsize := effsampsize + 1;
              for j := 1 to max do 
                if (j in graphvertices) and (not (j in haslatentpar)) then 
          	      insertdata(j,category[j].condprob,datapoint,category);
           end;
     end;
   for j := 1 to max do 
     if (j in graphvertices) and (not (j in haslatentpar)) then 
        converttree(j,category[j].condprob,category,estmethod,precision);
end;

procedure doestimate(* var adjacency: adjtype; var outfile: text; minternal: boolean; 
                       estmethod:esttype;precision:integer *);
var i,j: vertextype; meqs: eqorder; 
begin
    if not minternal then
      for i := 1 to max do 
        for j := 1 to max do
          if (adjacency^[i,j] <> notthere) and 
             ((istheoretical(vernames[i])) or (istheoretical(vernames[j]))) then
                myerror('Cannot estimate graph with latent variables.|',fatal,false);
 if not aaabort then
    begin
       formtree(adjacency,cats);
       if not aaabort then
         begin
           fillouttree(cats,estmethod,precision);
           new(meqs);
           formeqsorder(meqs,adjacency);
         end;
       if (not minternal) then 
           begin
             if not aaabort then
               writediscretemodel(cats,meqs,adjacency,ofilvar,false);
             disposebayesnet(cats);
           end;
       disposeeqsorder(meqs);
       if not minternal then finishfile(ofilvar,ofilename);
     end;
end;

function lfact(n:integer):real;
var
   i:integer;
   val:real;
begin
   val:=0;
   if n>0 then
      for i:=1 to n do val:=val + ln(i);
   lfact:=val;
end;

(* New 6/95 - RS *)
  
 (* This will prodcue a sample covariance matrix, with n=sampsize
    from the population defined by adjacency, parm, and exogd, and
    store the result in scov *)
procedure getsamplecov(*var scov: covarray*);
  begin
    undercontinuous := true; 
    havevalue := [];       
    novalueyet := graphvertices;
    montesamp:= round(sampsize);
    datatypes:= [2];
    new(varorder);
    formeqsorder(varorder,adjacency);
    datainit;
    initrand(newseed);
    getdata(true);
    disposepointrs(tiers,varorder);
  end;

(*this procedure moves tempprobptr to the counts for the first instantiation of the parents*)
procedure setuppointer(condprob:probptr;i:integer;var tempprobptr:probptr);
begin
   while (condprob^.varnum<>i) do condprob:=condprob^.nextvalue[1];
   tempprobptr:=condprob;
end;

procedure setcatspointer(var condprob:probptr;varnum:integer;var tempprobptr:probptr;var first:boolean);
var j:integer;
begin
   if (condprob^.varnum=varnum) then
     begin
      if first then first:=false
        else tempprobptr^.nextvalue[1] := condprob;
      tempprobptr := condprob;
     end
   else
     for j:= 1 to cats[condprob^.varnum].num do
	setcatspointer(condprob^.nextvalue[j],varnum,tempprobptr,first);
end;


function myexpo(base,exponent:integer):real;
begin
  if base=0 then myexpo:= 0
  else if exponent=0 then myexpo:= 1
  else if exponent < 0 then myexpo := (1 / myexpo(base, (-1 * exponent)))
  else myexpo := base * myexpo(base,exponent - 1);
end;

function choose(n,i:integer):integer;
var
	j,k,answer : integer;
begin
  answer:=1;
  for j:= i+1 to n do
	answer:= answer * j;
  for j:= 2 to n-i do
	answer := round(answer/j);
  choose:=answer;
end;

function getnumberofmodels(n:integer):double;
var
	i,j:integer;
        sum:double;
	farray:array[0..50] of double;
begin
  farray[0]:=1;
  farray[1]:=1;
  for i := 2 to n do
    begin
	sum:=0;
	for j:= 1 to i do
	  if odd(j) then sum:=sum+ (choose(i,j) * myexpo(2,j*(i-j)) * farray[i-j])
	  else sum:=sum - (choose(i,j) * myexpo(2,j*(i-j)) * farray[i-j]);
	farray[i]:=sum;
    end;
   twrite(output,'There are |');
   write(output,farray[n]);
   twriteln(output,' models.|');
   getnumberofmodels:=farray[n];

end;

procedure gregscore(var lscore:real);
var
   i,j,k:integer;
   numcat,numpi,Nij:integer;
   tempprobptr:probptr;

begin
   lscore:=0;
   for i:=1 to max do
      begin
	numcat:=cats[i].num;
	numpi:=cats[i].numpi;
        setuppointer(cats[i].condprob,i,tempprobptr);
	for j := 1 to numpi do
	   begin
	     Nij:=0;
	     for k := 1 to numcat do Nij:=Nij+round(tempprobptr^.cutoff[k]);
	     lscore := lscore + lfact(numcat -1) - lfact(Nij + numcat -1);
	     for k:= 1 to numcat do
	        begin
		  lscore:= lscore + lfact(round(tempprobptr^.cutoff[k]));
	        end;
	     tempprobptr:=tempprobptr^.nextvalue[1];
	   end;
      end;
end;

function gammln (* (xx: real): real; *);
(* This function returns ln(gamma(xx)). *)
        const
                stp = 2.50662827465;
                half = 0.5;
                one = 1.0;
                fpf = 5.5;
        var
                x, tmp, ser: real;
                j: integer;
                cof: array[1..6] of real;
begin
        cof[1] := 76.18009173;
        cof[2] := -86.50532033;
        cof[3] := 24.01409822;
        cof[4] := -1.231739516;
        cof[5] := 0.120858003e-2;
        cof[6] := -0.536382e-5;
        x := xx - one;
        tmp := x + fpf;
        tmp := (x + half) * ln(tmp) - tmp;
        ser := one;
        for j := 1 to 6 do
                begin
                        x := x + one;
                        ser := ser + cof[j] / x;
                end;
        gammln := tmp + ln(stp * ser);
end; (* gammln *)


procedure equivbranch(var lambda,lsco: real; v: vertextype);
var
   j,k:integer;
   numcat,numpi,Nij:integer;
   Npij,Npijk:real;
   tempprobptr:probptr;

begin
   lsco := 0;
   numcat:=cats[v].num;
   numpi:=cats[v].numpi;
   Npij:=lambda/numpi;
   Npijk:=lambda/(numcat * numpi);
   setuppointer(cats[v].condprob,v,tempprobptr);
   for j := 1 to numpi do
     begin
	Nij:=0;
	for k := 1 to numcat do Nij:=Nij+round(tempprobptr^.cutoff[k]);
        lsco := lsco + gammln(Npij) - gammln(Npij + Nij);
        for k:= 1 to numcat do
	      lsco:= lsco + gammln(Npijk + tempprobptr^.cutoff[k]) -
 			  gammln(Npijk);
        tempprobptr:=tempprobptr^.nextvalue[1];
     end;
end;

procedure equivscore(var lambda,lscore:real);
var lsco: real; i: integer;
begin
   lscore := 0;
   for i := 1 to max do 
    if (i in graphvertices) then
     begin
        equivbranch(lambda,lsco,i);
        lscorearr[i] := lsco;
        lscore := lscore + lscorearr[i];
     end;
end;

procedure myfilloutbranch(var category: categoryarray; v: vertextype);
var i: integer; j: vertextype; datapoint:vertexarray; allknown: boolean;
begin
   for i := 1 to round(sampsize) do 
    if (samparray[i]=train) then 
     begin
        allknown := true;
        for j := 1 to numobgraph do datapoint[graphlist[j]]:=getbigdata(i,graphlist[j]);
       	insertdata(v,category[v].condprob,datapoint,category);
     end;
end;

procedure myfillouttree(var category: categoryarray);
var i: integer; j: vertextype; datapoint:vertexarray; allknown: boolean;
begin
   for i := 1 to round(sampsize) do 
(*    if (samparray[i]=trainest) or (samparray[i]=est) then *)
     begin
        allknown := true;
	for j:= 1 to max do
          if (j in graphvertices) then 
           begin
              datapoint[j]:=getbigdata(i,j);
              if datapoint[j] = 0 then allknown := false;
           end;
	if allknown then
             for j := 1 to max do 
              if (j in graphvertices) then
        	insertdata(j,category[j].condprob,datapoint,category);
     end;
end;


procedure setupscore(v: vertextype);
var
   pi,j:integer;
   tempprobptr:probptr;
   first:boolean;
begin
   formbranch(adjacency,cats,v);

   {the following chunk sets the number of parent instantiations}
   pi:=1;
   for j:= 1 to numobgraph do
     if graphlist[j] in causes[v] then pi:=pi*cats[graphlist[j]].num;
   cats[v].numpi:=pi;
   myfilloutbranch(cats,v);

   {the following chunk sets up the pointer structure in the index tree; a given leaf in the cats structure corresponds to one instantiation of the parents and these pointers point from one leaf to the next leaf}
  first:=true;
  setcatspointer(cats[v].condprob,v,tempprobptr,first);
end;

procedure setupscoring;
var
   i,pi,j:integer;
   tempprobptr:probptr;
   first:boolean;
begin
   formtree(adjacency,cats);

   {the following chunk sets the number of parent instantiations}
   for i:= 1 to numobgraph do
   begin
      pi:=1;
      for j:= 1 to numobgraph do
  	  if graphlist[j] in causes[graphlist[i]] then pi:=pi*cats[graphlist[j]].num;
      cats[graphlist[i]].numpi:=pi;
   end;
   myfillouttree(cats);

   {the following chunk sets up the pointer structure in the index tree; a given leaf in the cats structure corresponds to one instantiation of the parents and these pointers point from one leaf to the next leaf}
   for i:= 1 to numobgraph do
	begin
	  first:=true;
	  setcatspointer(cats[graphlist[i]].condprob,graphlist[i],tempprobptr,first);
	end;
end;


(* read in data and graph then estimate and then type init*)

procedure getscore(* var lambda,score: real *);
var
	nummod:double;
begin
	setupscoring;
(*	nummod:=getnumberofmodels(max); *)
	equivscore(lambda,score);
end;

procedure getmodscore(* var lambda,locscore: real; v: vertextype *);
begin
      score := score - lscorearr[v];
      setupscore(v);
      equivbranch(lambda,locscore,v);
end;


procedure entropybranch(var lsco: real; v: vertextype);
var
   j,k:integer;
   numcat,numpi,Nij:integer;
   Npij,Npijk:real;
   tempprobptr:probptr;

begin
   lsco := 0;
   numcat:=cats[v].num;
   numpi:=cats[v].numpi;
   setuppointer(cats[v].condprob,v,tempprobptr);
   for j := 1 to numpi do
     begin
	Nij:=0;
	for k := 1 to numcat do Nij:=Nij+round(tempprobptr^.cutoff[k]);
        for k:= 1 to numcat do
	   begin
	   if ((round(tempprobptr^.cutoff[k])<>0) and (Nij<>0)) then
	      lsco:= lsco + (tempprobptr^.cutoff[k]*ln(tempprobptr^.cutoff[k]/Nij));
{	      if (Nij=0) then twrite(output,'Warning::Nij is zero !!   |');}
	   end;
        tempprobptr:=tempprobptr^.nextvalue[1];
     end;
   lsco:=lsco-(ln(ln(round(sampsize)))*(1/2)*(numcat-1)*numpi);
end;


procedure entropymodscore(* var locscore: real; v: vertextype *);
begin
      score := score - lscorearr[v];
      setupscore(v);
      entropybranch(locscore,v);
end;

procedure entropyscore(*var lscore:real*);
var lsco: real; i: integer;
begin
   setupscoring;   
   lscore := 0;
   for i := 1 to max do 
    if (i in graphvertices) then
     begin
        entropybranch(lsco,i);
        lscorearr[i] := lsco;
        lscore := lscore + lscorearr[i];
     end;
end;

procedure myunif;
var i: vertextype; partial: partialpair; partialedon: partialedonarray; aprob,g2: real; numchecked: integer;
    held: boolean;
begin
for i := 1 to 50 do
  begin
    repeat
      partialedon.length := 0;
      repeat
        repeat
         partial[1] :=  round(randunif(7,107));
        until partial[1] in graphvertices;
        repeat
         partial[2] :=  round(randunif(7,107));
        until partial[2] in graphvertices;
      until partial[1] <> partial[2];
      held := calcindep1(partial,aprob,g2,partialedon,numchecked);
    until aprob <= 0.01;
    writename(partial[1],output);
    write(output,' ');
    writename(partial[2],output);
    writeln(output,' ');
  end;
end; 


   