{$N+,E+,F+,O+}
unit part;
interface
uses tt1,tt2,tt3,tt4,tt5,dos;

{#include "types.i"
#include "part.h"
#include "calctet.h"
#include "io.h"
#include "util.h"
#include "cspecial.h"
#include "indep.h"
#include "sys.h"
#include "eqs.h"
#include "calctet1.h"}

const
    arraylength = 32000;
    numval = 60;
    maxpart = 100;

type 
    cellsarr = array[1..arraylength] of integer;
    cellptr = ^cellsarr;

    partialpair = array[1..2] of integer;
    partialedonarray = record
			  length: integer;
                          arr: array[1..maxpart] of integer;
                       end;

var pathmatrix: pathstype;
    cells: cellptr;
    eqsres: text;
    bloop,eloop,mimb: integer;

procedure checktime; 
procedure getpartials(fromscales: boolean); 

implementation
uses calctet,calctet1,io,util,rsutil,sys,indep,eqs;

const
    edgepath = 1;
    nonedgepath = 3;

type 
    triangleptr = ^trilist;
    trilist = record
                inout,unori,other: vertextype;
                fwd,back: triangleptr;
              end;
    datatype = 0..60;
    edgeptr = ^edgeslist;
    edgeslist = record
                 edge: integer;
                 next: edgeptr;
                 last: edgeptr;
               end;   
    parentstype = array[1..maxvertex] of vertexset;
    nodeptr = ^node;
    node = record
              g2, prob: real;
              edge: integer;
              currentset: vertexset;
              fwd,back: nodeptr;
            end;
    endtype = (in2,out2,knownin2,knownout2,notknown);
    endset = set of endtype;
    endarraytype = array[0..maxvertex,0..maxvertex] of endset;
    endarray = ^endarraytype;
    dseparraytype = array[1..maxvertex,1..maxvertex] of vertexset;
    dseparray = ^dseparraytype;
    smallarray = array[1..25,1..25] of real;

var 
    partscore: real;
    ugraph: endarray;
    zeroes: dseparray;
    parents: parentstype;
    debug: boolean;
    heuristic: boolean;    
    numchecked: integer;
    start: double;
    toomuchtime: boolean;
    dsep: dseparray;
    deftrilist,posstrilist,nocollidelist: triangleptr;    

procedure tnewend(var ptr: endarray; err: errortype);
var size: word;
begin
   size := sizeof(endarraytype);
   if (size < heapavail) then new(ptr)
   else errormess('Out of memory: Cannot run build.|',err,false);
end;

procedure tnewdsep(var ptr: dseparray; err: errortype);
var size: word;
begin
   size := sizeof(dseparraytype);
   if (size < heapavail) then new(ptr)
   else errormess('Out of memory: Cannot run build.|',err,false);
end;

procedure checktime;
var now: double;
begin
         now := tgettime;
         if maxtime <> -1 then 
             if (now - start) > maxtime then toomuchtime := true;
end;


function setsize(aset: vertexset): vertextype;
var i: vertextype; num: integer;
begin
  num := 0;
  for i := 1 to maxvertex do
     if i in aset then num := num + 1;
  setsize := num;
end;

function getpower(base: real; exponent: integer): real;
var i: integer; pow: real;
begin
   pow := 1;
   for i := 1 to exponent do pow := pow * base;
   getpower := pow;
end;

(***************************************************************************)

procedure matinv(matrixa: smallarray; var matrixb: smallarray;
                 var ok: boolean; size: vertextype);
label 999;
const errorbound = 1.0E-10;
var i,j,k,m: vertextype;
    factor,temp: real;
begin
   for i := 1 to size do
     for j := 1 to size do
        if i = j then matrixb[i,j] := 1.0
        else matrixb[i,j] := 0.0;
   for j := 1 to size do
     begin
       i := j;
       while abs(matrixa[i,j]) < errorbound do
          begin
             if i = size then
               begin
                  ok := false;
                  goto 999;
               end;
             i := i + 1;
          end;
       for k := 1 to size do 
           begin
              temp := matrixa[i,k];
              matrixa[i,k] := matrixa[j,k];
              matrixa[j,k] := temp;
              temp := matrixb[i,k];
              matrixb[i,k] := matrixb[j,k];
              matrixb[j,k] := temp;
           end;
        factor := 1.0 / matrixa[j,j];
        for k := 1 to size do 
           begin
              matrixa[j,k] := factor * matrixa[j,k];
              matrixb[j,k] := factor * matrixb[j,k];
           end;
        for m := 1 to size do
           if m <> j then
             begin
                factor := -1 * matrixa[m,j];
                for k := 1 to size do
                   begin
                     matrixa[m,k] := matrixa[m,k] + factor * matrixa[j,k];
                     matrixb[m,k] := matrixb[m,k] + factor * matrixb[j,k];
                   end
             end
    end;
  ok := true;
999: end;

procedure fillinmatrix(tail,head: vertextype; partialedon: partialedonarray;
          var matrixa: smallarray);
var i,j: vertextype;
begin
    for i := 1 to partialedon.length+2 do matrixa[i][i] := 1.0;
    matrixa[1,2] := rho^[tail,head];
    for i := 3 to partialedon.length+2 do
       matrixa[1,i] := rho^[tail,partialedon.arr[i-2]];
    for i := 3 to partialedon.length+2 do
       matrixa[2,i] := rho^[head,partialedon.arr[i-2]];
    for i := 3 to partialedon.length+1 do
      for j := i+1 to partialedon.length+2 do
        matrixa[i,j] := rho^[partialedon.arr[i-2],partialedon.arr[j-2]];
    for i := 1 to partialedon.length+1 do
      for j := i+1 to partialedon.length+2 do
         matrixa[j,i] := matrixa[i,j];
end;

function calcpart(partial: partialpair; partialedon: partialedonarray): real;
var temp: real; matrixa,matrixb: smallarray; ok: boolean;
begin
    if partialedon.length = 0 then 
          calcpart:= rho^[partial[1],partial[2]]
    else 
       begin
         fillinmatrix(partial[1],partial[2],partialedon,matrixa);
         matinv(matrixa,matrixb,ok,partialedon.length+2);
         calcpart := -1 * matrixb[1,2] / (sqrt(matrixb[1,1] * matrixb[2,2]));
       end;
end;
         
(**************************************************************************)

{function calcpart(partial: partialpair; partialedon: partialedonarray): real;
var rhoimpart,rhojmpart: partialpair; ij,ik,jk,ik2,jk2,temp: real; i: integer;
begin
    if partialedon.length = 0 then 
          temp := rho^[partial[1],partial[2]]
    else
      begin
        rhoimpart[1] := partial[1];
        rhoimpart[2] := partialedon.arr[partialedon.length];
        rhojmpart[1] := partial[2];
        rhojmpart[2] := partialedon.arr[partialedon.length];
        partialedon.length := partialedon.length - 1;
        ij := calcpart(partial,partialedon);
        ik := calcpart(rhoimpart,partialedon);
        jk := calcpart(rhojmpart,partialedon);
        ik2 := ik * ik;
        jk2 := jk * jk;
        if (abs(ik2) <= 1) and (abs(jk2) <= 1) then
          temp := (ij - (ik * jk))/ 
                       (sqrt(1 - ik2) *  sqrt(1 - jk2))
        else temp := 2;
        if abs(temp) > 1 then
           begin
		writename(partial[1],ofilvar);
		twrite(ofilvar,' |');
		writename(partial[2],ofilvar);
		twrite(ofilvar,' .|');
                for i := 1 to partialedon.length + 1 do
		   begin
			twrite(ofilvar,' |');
			writename(partialedon.arr[i],ofilvar);
		   end;
                twrite(ofilvar,' does not exist.|');
		twriteeoln(ofilvar);
           end;
   end;
 calcpart := temp;
end; }


procedure calcpartialheld(var partial,prob: real; var held: boolean;
                          length: integer);
var z: real;
begin
  begin
   if ((numchecked mod 10000) = 0) then 
      begin
         checktime;
         numchecked := 1;
      end
   else numchecked := numchecked + 1;
   if abs(partial) <= 1 then
          begin
              z := (1/2) * (sqrt(sampsize - 3 - length)) * ln((1 + abs(partial)) / 
                   (1 - abs(partial)));
              getprob(prob,abs(z));
              held := prob >= signif;
          end
   else held := false;
 end;
end;  { procedure }

procedure outpart(partial,pofstdev: real; part: partialpair; 
                  partialedon: partialedonarray; impld,hold: boolean);
var field,i: integer; impldcheck,heldcheck: char;
begin
         setcheck(impld,impldcheck);
         setcheck(hold,heldcheck);
         if partial >= 0 then field := 4 else field := 3;
	 writename(part[1],ofilvar);
         twrite(ofilvar,' |');
	 writename(part[2],ofilvar);
	 twrite(ofilvar,' . |');
         for i := bloop to eloop-2 do
             if i <= partialedon.length then 
                writename(partialedon.arr[i],ofilvar)
             else twrite(ofilvar,'      |');
         tpwritereal(ofilvar,partial,3,field,false);
	 twrite(ofilvar,'  |');
	 twritechar(ofilvar,impldcheck);
	 twrite(ofilvar,'  |');
	 twritechar(ofilvar,heldcheck);
	 twrite(ofilvar,'  |');
	 tpwritereal(ofilvar,pofstdev,7,4,false)
end;	

function collider(i,j,k: integer): boolean;
begin
   if (knownin2 in ugraph^[i,k]) and (knownin2 in ugraph^[j,k]) then 
      collider := true
   else if ((knownout2 in ugraph^[i,k]) or (knownout2 in ugraph^[j,k])) then
      collider := false
   else if (i < j) then
      collider := (ugraph^[i,k] <> []) and (ugraph^[j,k] <> []) and
               not (k in zeroes^[i,j])
   else collider := (ugraph^[i,k] <> []) and (ugraph^[j,k] <> []) and
               not (k in zeroes^[j,i]);
end;

procedure readarraytoset(var dsepcell: vertexset; partialedon: partialedonarray);
var i: vertextype;
begin
   dsepcell := [];
   for i := 1 to partialedon.length do
      dsepcell := dsepcell + [partialedon.arr[i]];
end;

procedure readsettoarray(parents: vertexset; var vertices: partialedonarray;
                         ver1,ver2: vertextype);
var vertloc,j: integer;
begin
  vertices.length := 0;
  vertloc := 1;
  j := bloop;
  repeat
    if (j in parents) and 
       ((not (j in temporder[ver1])) or (not (j in temporder[ver2]))) then
       begin
         vertices.arr[vertloc] := j;
         vertloc := vertloc + 1;
         vertices.length := vertices.length + 1;
       end;
    j := j + 1;
  until j > eloop;
end;

procedure setdirection(tail,head: vertextype);
begin
end;

function triangle(x,y,z: vertextype): boolean;
begin
   triangle := (ugraph^[x,y] <> []) and (ugraph^[y,z] <> []) and 
               (ugraph^[x,z] <> []);
end;

function innocollidelist(x,y,z: vertextype; var nocollidelist: triangleptr): triangleptr;
var curpos: triangleptr; found: boolean;
begin
   innocollidelist := nil;
   found := false;
   curpos := nocollidelist;
   while (curpos <> nil) and (not found) do
     begin
       if (curpos^.unori = y) and 
          ((curpos^.other = x) and (curpos^.inout = z)) or
          ((curpos^.other = z) and (curpos^.inout = x)) then found := true;
       if found then innocollidelist := curpos 
       else curpos := curpos^.fwd;
     end;
end;

procedure putinlist(x,y,z: vertextype; var trilist: triangleptr);
var newtri: triangleptr;
begin
   new(newtri);
   newtri^.inout := x;
   newtri^.unori := y;
   newtri^.other := z;
   newtri^.back := nil;
   if trilist <> nil then trilist^.back := newtri;
   newtri^.fwd := trilist;
   trilist := newtri;
end;

procedure formtri(x,y,z: vertextype; var deftrilist,posstrilist: triangleptr);
begin
   if (((in2 in ugraph^[y,x]) or (knownin2 in ugraph^[y,x])) and
       ((out2 in ugraph^[z,x]) or (knownout2 in ugraph^[z,x]))) or
      (((in2 in ugraph^[z,x]) or (knownin2 in ugraph^[z,x])) and
       ((out2 in ugraph^[y,x]) or (knownout2 in ugraph^[y,x]))) then
         begin
            if ((notknown in ugraph^[x,y]) or (notknown in ugraph^[z,y])) then
                putinlist(x,y,z,deftrilist);
            if ((notknown in ugraph^[x,z]) or (notknown in ugraph^[y,z])) then
                putinlist(x,z,y,deftrilist);
         end
   else if (not (((in2 in ugraph^[y,x]) or (knownin2 in ugraph^[y,x])) and
           ((in2 in ugraph^[z,x]) or (knownin2 in ugraph^[z,x])))) and
           (not (((out2 in ugraph^[y,x]) or (knownout2 in ugraph^[y,x])) and
           ((out2 in ugraph^[z,x]) or (knownout2 in ugraph^[z,x])))) then
         begin
            if ((notknown in ugraph^[x,y]) or (notknown in ugraph^[z,y])) then
                putinlist(x,y,z,posstrilist);
            if ((notknown in ugraph^[x,z]) or (notknown in ugraph^[y,z])) then
                putinlist(x,z,y,posstrilist);
         end;
end;

procedure inittrilists(var deftrilist,posstrilist,nocollidelis: triangleptr);
var i,j,k: vertextype;
begin
  deftrilist := nil;
  posstrilist := nil;
  nocollidelist := nil;
  for i := bloop to eloop do
    for j := i + 1 to eloop do
       for k := j + 1 to eloop do
          if triangle(i,j,k) then
             begin
               formtri(i,j,k,deftrilist,posstrilist);
               formtri(j,k,i,deftrilist,posstrilist);
               formtri(k,j,i,deftrilist,posstrilist);
             end;
end;

procedure removetri(var curtri,trilist: triangleptr);
begin
    if curtri^.back <> nil then curtri^.back^.fwd := curtri^.fwd;
    if curtri^.fwd <> nil then curtri^.fwd^.back := curtri^.back;
    if curtri = trilist then trilist := trilist^.fwd;
end;

procedure updateposstrilist(var deftrilist,posstrilist,nocollidelist: triangleptr; 
                            var nochange: boolean);
var curpos,next: triangleptr; x,y,z: vertextype;
begin
   curpos := posstrilist;
   while (curpos <> nil) do 
     begin
         x := curpos^.inout;
         y := curpos^.unori;
         z := curpos^.other;
         next := curpos^.fwd;
         if (((in2 in ugraph^[y,x]) or (knownin2 in ugraph^[y,x])) and
             ((out2 in ugraph^[z,x]) or (knownout2 in ugraph^[z,x]))) or
            (((in2 in ugraph^[z,x]) or (knownin2 in ugraph^[z,x])) and
             ((out2 in ugraph^[y,x]) or (knownout2 in ugraph^[y,x]))) and
            ((notknown in ugraph^[x,y]) or (notknown in ugraph^[z,y])) and 
            (innocollidelist(x,y,z,nocollidelist) = nil) then
                begin
                   removetri(curpos,posstrilist);
                   curpos^.fwd := deftrilist;
                   curpos^.back := nil;
                   if deftrilist <> nil then deftrilist^.back := curpos;
                   deftrilist := curpos;
                   nochange := false;
                end
         else if (((in2 in ugraph^[y,x]) or (knownin2 in ugraph^[y,x])) and
             ((in2 in ugraph^[z,x]) or (knownin2 in ugraph^[z,x]))) or
            (((out2 in ugraph^[y,x]) or (knownout2 in ugraph^[y,x])) and
             ((out2 in ugraph^[z,x]) or (knownout2 in ugraph^[z,x]))) or
            ((not (notknown in ugraph^[x,y])) and (not (notknown in ugraph^[z,y]))) or
            (innocollidelist(x,y,z,nocollidelist) <> nil) then
                begin
                  removetri(curpos,posstrilist);
                  dispose(curpos);
                end;
          curpos := next;
     end;
end;

procedure setnocollide(var curtri,deftrilist,nocollidelist: triangleptr);
var nextri: triangleptr;
begin
   nextri := curtri^.fwd;
   removetri(curtri,deftrilist);
   curtri^.fwd := nocollidelist;
   if nocollidelist <> nil then nocollidelist^.back := curtri;
   curtri^.back := nil;
   nocollidelist := curtri;
   curtri := nextri;
end;

procedure adddedge(tail,head: edge; var adj: adjtype);
var i,j: vertextype; newtailpaths: array[1..maxvertex] of vertextype;

begin
   for i := bloop to eloop do
       for j := bloop to eloop do
          if ((i <> tail) or (adj^[i,tail] <> notthere)) and
             ((j <> head) or (adj^[head,j] <> notthere)) then
                adj^[i,j] := nonedgepath;
   adj^[tail,head] := edgepath;
end;

procedure propagate(tail,head: vertextype; var nocollidelist: triangleptr;
                    bothends: boolean; var adj: adjtype);
var i: integer; curpos: triangleptr;
begin
    if bothends then ugraph^[head,tail] := ugraph^[head,tail] + [out2] - [notknown];
    ugraph^[tail,head] := ugraph^[tail,head] + [in2] - [notknown];
    if bothends and 
      (not
        (((in2 in ugraph^[head,tail]) or (knownin2 in ugraph^[head,tail])) or
         ((out2 in ugraph^[tail,head]) or (knownout2 in ugraph^[tail,head]))))
       then adddedge(tail,head,adj);
    for i := bloop to eloop do
      if (i <> tail) and (i <> head) and 
         (not (out2 in ugraph^[tail,head])) and
         (not (knownout2 in ugraph^[tail,head])) and
         (ugraph^[i,head] <> []) and 
         (not (in2 in ugraph^[i,head])) and
         (not (knownin2 in ugraph^[i,head])) and
         (not (out2 in ugraph^[head,i])) and
         (not (knownout2 in ugraph^[head,i])) and
         (not ((in2 in ugraph^[head,i]) and (out2 in ugraph^[i,head]))) then
              if ugraph^[tail,i] = [] then propagate(head,i,nocollidelist,true,adj)
              else 
                 begin
                   curpos := innocollidelist(tail,head,i,nocollidelist);
                   if curpos <> nil then
                       begin
                         removetri(curpos,nocollidelist);
                         propagate(head,i,nocollidelist,true,adj);
                       end;
                 end;
end;     

procedure discrimpath(var curtri,deftrilist,nocollidelist: triangleptr; 
                      var curpath: vertexset; nearend,curend: vertextype; 
                      var founddis,nochange: boolean; var adj: adjtype);
var i,other: vertextype; nextri: triangleptr; 
begin
   other := curtri^.other;
   i := bloop - 1;
   repeat
      i := i + 1;
      if ((in2 in ugraph^[i,curend]) or (knownin2 in ugraph^[i,curend])) and
         (not (i in curpath)) and (i <> curtri^.unori) and (ugraph^[i,other] = []) and
         (((in2 in ugraph^[nearend,curend]) or (knownin2 in ugraph^[nearend,curend])) and
          ((out2 in ugraph^[other,curend]) or (knownout2 in ugraph^[other,curend]))) or
         (((in2 in ugraph^[other,curend]) or (knownin2 in ugraph^[other,curend])) and
          ((out2 in ugraph^[nearend,curend]) or (knownout2 in ugraph^[nearend,curend]))) then
              begin
                 nochange := false;
                 founddis := true;
                 if (curtri^.unori in dsep^[i,curtri^.other]) then
                   begin
                      if (in2 in ugraph^[curtri^.inout,curtri^.unori]) or 
                         (knownin2 in ugraph^[curtri^.inout,curtri^.unori]) then
                            begin
                                nextri := curtri^.fwd;
                                propagate(curtri^.unori,curtri^.other,
                                          nocollidelist,true,adj);
                                removetri(curtri,deftrilist);
                                dispose(curtri);
                                curtri := nextri;
                            end
                      else if (in2 in ugraph^[curtri^.other,curtri^.unori]) or 
                         (knownin2 in ugraph^[curtri^.other,curtri^.unori]) then
                             begin
                                 nextri := curtri^.fwd;
                                 propagate(curtri^.unori,curtri^.inout,
                                           nocollidelist,true,adj);
                                 removetri(curtri,deftrilist);
                                 dispose(curtri);
                                 curtri := nextri;
                             end
                      else setnocollide(curtri,deftrilist,nocollidelist);
                     end
                 else
                    begin
                      nextri := curtri^.fwd;
                      propagate(curtri^.inout,curtri^.unori,nocollidelist,false,adj);
                      propagate(curtri^.other,curtri^.unori,nocollidelist,false,adj);
                      removetri(curtri,deftrilist);
                      dispose(curtri);
                      curtri := nextri;
                    end;
               end;
   until founddis or (i > eloop);                      
   i := bloop - 1;
   while (i <= eloop) and (not founddis) do
    begin
      i := i + 1;
      if ((in2 in ugraph^[i,curend]) or (knownin2 in ugraph^[i,curend])) and
         (not (i in curpath)) and (i <> curtri^.unori) and (ugraph^[i,other] <> []) and
         (((in2 in ugraph^[nearend,curend]) or (knownin2 in ugraph^[nearend,curend])) and
          ((out2 in ugraph^[other,curend]) or (knownout2 in ugraph^[other,curend]))) or
         (((in2 in ugraph^[other,curend]) or (knownin2 in ugraph^[other,curend])) and
          ((out2 in ugraph^[nearend,curend]) or (knownout2 in ugraph^[nearend,curend]))) then
              begin
                curpath := curpath + [i];
                discrimpath(curtri,deftrilist,nocollidelist,curpath,curend,i,founddis,nochange,adj);
                curpath := curpath - [i];
              end;
      end;
end;

procedure suffprop(tail,head: vertextype; var adj: adjtype);
var i: integer;
begin
    ugraph^[head,tail] := ugraph^[head,tail] + [out2] - [notknown];
    ugraph^[tail,head] := ugraph^[tail,head] + [in2] - [notknown];
    adddedge(tail,head,adj);
    for i := bloop to eloop do
      if (i <> tail) and (i <> head) and 
         (not (out2 in ugraph^[tail,head])) and
         (not (knownout2 in ugraph^[tail,head])) and
         (ugraph^[i,head] <> []) and (ugraph^[tail,i] = []) and
         (not (in2 in ugraph^[i,head])) and
         (not (knownin2 in ugraph^[i,head])) and
         (not (out2 in ugraph^[head,i])) and
         (not (knownout2 in ugraph^[head,i])) and
         (not ((in2 in ugraph^[head,i]) and (out2 in ugraph^[i,head]))) then
             suffprop(head,i,adj);
end;     

procedure setcollider(dopropagate: boolean; var nocollidelist: triangleptr;
                      var nochange: boolean; var adj: adjtype);
var i,j,k: integer;
begin
   nochange := true;
   for i := bloop to eloop do
     for j := bloop to eloop do
       for k := bloop to eloop do
         if (i <> j) and (j <> k) and (i <> k) then
            if ugraph^[i,j] = [] then
               if collider(i,j,k) then 
                   begin
                      if (not (in2 in ugraph^[i,k])) or
                         (notknown in ugraph^[i,k]) or
                         (not (in2 in ugraph^[j,k])) or
                         (notknown in ugraph^[j,k]) then nochange := false;
                      ugraph^[i,k] := ugraph^[i,k] + [in2] - [notknown];
                      ugraph^[j,k] := ugraph^[j,k] + [in2] - [notknown];;
(*                      if causalsuff then
                        begin
                          if (not (out2 in ugraph^[k,i])) or
                             (notknown in ugraph^[k,i]) or
                             (not (out2 in ugraph^[k,j])) or
                             (notknown in ugraph^[k,j]) then nochange := false;
                          ugraph^[k,i] := ugraph^[k,i] + [out2] - [notknown];
                          ugraph^[k,j] := ugraph^[k,j] + [out2] - [notknown];
                        end; *)
                    end;
   if dopropagate and (not nochange) then
   for i := bloop to eloop do
     for j := bloop to eloop do
       for k := bloop to eloop do
         if (i <> j) and (j <> k) and (i <> k) then
           if ((in2 in ugraph^[i,j]) or (knownin2 in ugraph^[i,j])) and 
              (not (out2 in ugraph^[i,j])) and 
              (not (knownout2 in ugraph^[i,j])) and
              (ugraph^[j,k] <> []) and (not (out2 in ugraph^[j,k])) and 
              (not (knownout2 in ugraph^[j,k])) and
              (not (in2 in ugraph^[k,j])) and 
              (not (knownin2 in ugraph^[k,j])) and
              (ugraph^[i,k] =[]) then
                 if not causalsuff then
                    propagate(j,k,nocollidelist,true,adj)
                 else suffprop(j,k,adj);
end;

procedure calcdpath(var ugraph: endarray; var adj: adjtype);
label 999;
var j,k,i: integer;

begin (* calcpath *)
    for i := bloop to eloop do 
      for j := bloop to eloop do
        if i <> j then
         if ((in2 in ugraph^[i,j]) or (knownin2 in ugraph^[i,j])) and
            ((out2 in ugraph^[j,i]) or (knownout2 in ugraph^[j,i])) and
            (not
              (((in2 in ugraph^[j,i]) or (knownin2 in ugraph^[j,i])) or
               ((out2 in ugraph^[i,j]) or (knownout2 in ugraph^[i,j])))) then
                adj^[i,j] := edgepath
         else adj^[i,j] := notthere;
    for k := bloop to eloop do 
       for i := bloop to eloop do 
           for j := bloop to eloop do        
               if (adj^[i,k] <> notthere) and
                  (adj^[k,j] <> notthere) then
                      adj^[i,j] := nonedgepath;
999: end; (* Calcdpath *)

procedure orientawaycycle(var nochange: boolean; var adj: adjtype;
                          var nocollidelist: triangleptr);
var i,j: vertextype;
begin
  for i := bloop to eloop do
      for j := bloop to eloop do
         if (adj^[i,j] = nonedgepath) and (ugraph^[i,j] <> []) and 
            (not (in2 in ugraph^[i,j])) and (not (knownin2 in ugraph^[i,j])) then
(* do not want to orient away unless path that is not edge;  if there is an
   edge and a directed path, then edge is not first, so test works *)
                begin
                    propagate(i,j,nocollidelist,false,adj);
                    nochange := false;
                end;
end;

procedure orientedges(var deftrilist,posstrilist,nocollidelist: triangleptr;
                      var adj: adjtype);
label 999;
var nochange,founddis: boolean; curpath: vertexset; curpos: triangleptr;
    i,j: vertextype; 
begin
    nochange := true;
    setcollider(true,nocollidelist,nochange,adj);
    calcdpath(ugraph,adj);
    inittrilists(deftrilist,posstrilist,nocollidelist);
    repeat
      if maxtime <> -1 then checktime; 
      if toomuchtime then goto 999;
      nochange := true;
      if nochange then setcollider(true,nocollidelist,nochange,adj);
(*      if nochange then orientawaycycle(nochange,adj,nocollidelist); *)
      if nochange then 
         begin
           curpos := deftrilist;
           while (curpos <> nil) do
              begin
                if maxtime <> -1 then checktime;
                if toomuchtime then goto 999;
                curpath := [];
                founddis := false;
                discrimpath(curpos,deftrilist,nocollidelist,curpath,curpos^.unori,
                            curpos^.inout,founddis,nochange,adj);
                if not founddis then curpos := curpos^.fwd;
              end
          end;
      if nochange then 
          updateposstrilist(deftrilist,posstrilist,nocollidelist,nochange);
    until nochange;
999: end;

procedure printcausalsuff;
var i,j: vertextype; head,tail: vertextype; written: vertexset;
begin
  written := [];
  twriteeoln(ofilvar);
  for i := bloop to eloop-1 do
   for j := i+1 to eloop do
    if (i in graphvertices) and (j in graphvertices) then
     if (ugraph^[i,j] <> []) or (ugraph^[j,i] <> []) then
      begin
          written := written + [i,j];
          if (knownin2 in ugraph^[i,j]) then tail := i
          else if (knownin2 in ugraph^[j,i]) then tail := j
          else if (in2 in ugraph^[i,j]) then tail := i
          else if (in2 in ugraph^[j,i]) then tail := j
          else tail := i;
          if tail = i then head := j 
          else head := i;
	  pwritename(tail,ofilvar,7);
	  twrite(ofilvar,' |');
          if (in2 in ugraph^[head,tail]) and
             (not (knownout2 in ugraph^[head,tail])) then twrite(ofilvar,'<|')
          else twrite(ofilvar,'-|');
          twrite(ofilvar,'-|');
          if (notknown in ugraph^[tail,head]) then twrite(ofilvar,'- |')
          else twrite(ofilvar,'> |');
	  writename(head,ofilvar);
          if ([in2,out2] <= ugraph^[tail,head]) or 
             ([in2,knownout2] <= ugraph^[tail,head]) or             
             ([out2,knownin2] <= ugraph^[tail,head]) or
             ([out2,in2] <= ugraph^[tail,head]) or 
             ([in2,out2] <= ugraph^[head,tail]) or
             ([in2,knownout2] <= ugraph^[head,tail]) or             
             ([out2,knownin2] <= ugraph^[head,tail]) or
             ([out2,in2] <= ugraph^[head,tail]) then 
                  twrite(ofilvar,' * |');
	twriteeoln(ofilvar);
   end;
(*  for i := bloop to eloop do if (i in graphvertices) and (not (i in written)) then
     begin
	  pwritename(i,ofilvar,7);
          twriteeoln(ofilvar);
     end; *)
end;

procedure disposetri(var atrilist: triangleptr);
var curpos: triangleptr;
begin
  curpos := atrilist;
  if curpos <> nil then
     if curpos^.fwd <> nil then
        begin
          curpos := curpos^.fwd;
          dispose(curpos^.back);
        end
     else
        begin
           dispose(curpos);
           curpos := nil;
        end;
end;

procedure checkpath(var adj: adjtype; var ugraph: endarray);
var i,j,k: vertextype;
begin
    twriteln(ofilvar,'Directed Paths |');
    for i := bloop to eloop do 
       for j := bloop to eloop do
         if i <> j then
            if adj^[i,j] <> notthere then
               begin
                 writename(i,ofilvar);
                 twrite(ofilvar,' to |');
                 writename(j,ofilvar);
                 twriteeoln(ofilvar);
               end;
     twriteeoln(ofilvar);
     twriteeoln(ofilvar);
end;

procedure checknopath(var adj: adjtype; var ugraph: endarray);
var i,j,k: vertextype;
begin
 for i := bloop to eloop do 
   for j := bloop to eloop do
     if i <> j then
       if (not (in2 in ugraph^[j,i])) and (not (knownin2 in ugraph^[j,i])) and
          (ugraph^[i,j] <> []) then adj^[i,j] := 1
       else adj^[i,j] := notthere;
 for k := bloop to eloop do 
   for i := bloop to eloop do 
     for j := bloop to eloop do        
       if (adj^[i,k] <> notthere) and (adj^[k,j] <> notthere) then
          adj^[i,j] := 1;
    twriteln(ofilvar,'Not Connected by Directed Paths |');
    for i := bloop to eloop do 
       for j := bloop to eloop do
         if i <> j then
            if adj^[i,j] = notthere then
                begin
                  writename(i,ofilvar);
                  twrite(ofilvar,' to |');
                  writename(j,ofilvar);
                  twriteeoln(ofilvar);
                end;
     twriteeoln(ofilvar);
     twriteeoln(ofilvar);

end;
          
procedure printthepattern(var adj: adjtype);
var i,j: vertextype; head,tail: vertextype; written: vertexset; 
    curpos: triangleptr; 
begin
  written := [];
  twriteeoln(ofilvar);
  for i := bloop to eloop-1 do
    for j := i+1 to eloop do
     if (i in graphvertices) and (j in graphvertices) then
      if (ugraph^[i,j] <> []) or (ugraph^[j,i] <> []) then
        begin
          written := written + [i,j];
          if (knownin2 in ugraph^[i,j]) then tail := i
          else if (knownin2 in ugraph^[j,i]) then tail := j
          else if (in2 in ugraph^[i,j]) then tail := i
          else if (in2 in ugraph^[j,i]) then tail := j
          else tail := i;
          if tail = i then head := j 
          else head := i;
	  writename(tail,ofilvar);
          if ((in2 in ugraph^[head,tail]) or (knownin2 in ugraph^[head,tail])) then
              twrite(ofilvar,' <|')
          else if ((out2 in ugraph^[head,tail]) or (knownout2 in ugraph^[head,tail])) then twrite(ofilvar,' -|')
          else twrite(ofilvar,' o|');
          twrite(ofilvar,'-|');
          if ((in2 in ugraph^[tail,head]) or (knownin2 in ugraph^[tail,head])) then
              twrite(ofilvar,'> |')
          else if ((out2 in ugraph^[tail,head]) or (knownout2 in ugraph^[tail,head])) then twrite(ofilvar,'- |')
          else twrite(ofilvar,'o |');
	  writename(head,ofilvar);
          if ([in2,out2] <= ugraph^[tail,head]) or 
             ([in2,knownout2] <= ugraph^[tail,head]) or             
             ([out2,knownin2] <= ugraph^[tail,head]) or
             ([out2,in2] <= ugraph^[tail,head]) or 
             ([in2,out2] <= ugraph^[head,tail]) or
             ([in2,knownout2] <= ugraph^[head,tail]) or             
             ([out2,knownin2] <= ugraph^[head,tail]) or
             ([out2,in2] <= ugraph^[head,tail]) then 
                  twrite(ofilvar,' * |');
	twriteeoln(ofilvar);
     end;
  for i := bloop to eloop do if (i in graphvertices) and (not (i in written)) then
     begin
	  pwritename(i,ofilvar,7);
          twriteeoln(ofilvar);
     end;
  twriteeoln(ofilvar);
  curpos := nocollidelist;
  if curpos <> nil then
      twriteln(ofilvar,'The following pairs of edges do not collide: |');
  while curpos <> nil do
     begin
        writename(curpos^.inout,ofilvar);
        twrite(ofilvar,' --- |');
        writename(curpos^.unori,ofilvar);
        twrite(ofilvar,'    |');
        writename(curpos^.unori,ofilvar);
        twrite(ofilvar,' --- |');
        writename(curpos^.other,ofilvar);
	twriteeoln(ofilvar);        
        curpos := curpos^.fwd;
     end;
   checkpath(adj,ugraph);
   checknopath(adj,ugraph);
   disposetri(nocollidelist);
   disposetri(deftrilist);
   disposetri(posstrilist);
end;

procedure printlist;
var i,j: integer;
begin
for i := bloop to eloop do
  for j := bloop to eloop do
     if ((in2 in ugraph^[i,j]) or (knownin2 in ugraph^[i,j])) and
        (not (out2 in ugraph^[i,j])) and (not (knownout2 in ugraph^[i,j])) and
        ((out2 in ugraph^[j,i]) or (knownout2 in ugraph^[j,i])) and
        (not (in2 in ugraph^[j,i])) and (not (knownin2 in ugraph^[j,i])) then
	begin
	    writename(i,ofilvar);
	    twrite(ofilvar,' -> |');
	    writename(j,ofilvar);
	end;
end;

procedure getzeroes(var zero1,zero2,zero3,zero4: vertexset;
                    i,j,k,l,pattern: integer);
begin
   if pattern = 1 then
      begin
        zero1 := zeroes^[i,j];
        zero2 := zeroes^[k,l];
        zero3 := zeroes^[i,k];
        zero4 := zeroes^[j,l];
      end
   else if pattern = 2 then
      begin
        zero1 := zeroes^[i,l];
        zero2 := zeroes^[j,k];
        zero3 := zeroes^[i,j];
        zero4 := zeroes^[k,l];
      end
   else
      begin
        zero1 := zeroes^[i,k];
        zero2 := zeroes^[j,l];
        zero3 := zeroes^[i,l];
        zero4 := zeroes^[j,k];
      end  
end;

function parthold(partial: partialpair; var partialedon: partialedonarray): integer;
var val,prob: real; held: boolean;
begin
      val := calcpart(partial,partialedon);
      calcpartialheld(val,prob,held,partialedon.length);
      if held then parthold := 0 else parthold := 1;
end;

procedure removefile;
begin
   swapvectors;
   exec('\COMMAND.COM','/C DEL \MIMTEST\BLUB.IN'); 
   swapvectors;
   if doserror <> 0 then
     writeln('dos error #',doserror)
   else writeln('exec successful. Child process exit code = ',dosexitcode);
end;

procedure writelis1(partial: partialpair; var partialedon: partialedonarray);
var localadj: adjtype; localparm: covarray; i,j: vertextype; 
    comment,fname: inbuf; oldgraphvert: vertexset; 
begin
   oldgraphvert := graphvertices;
   graphvertices := [partial[1],partial[2]];
   new(localadj);
   new(localparm);
   fname := 'blub.in';
   fname[8] := chr(0);
   comment[1] := chr(0);
   removefile; 
   rewritefile(eqsinp,fname);
   for i := 1 to maxvertex do
     for j := 1 to maxvertex do
       begin
         localadj^[i,j] := notthere;
         localparm^[i,j] := 0.0;
       end;
   for i := 1 to 2 do
      for j := 1 to numob do
         begin
           localadj^[partial[i],j] := adjacency^[partial[i],j];
           if adjacency^[partial[i],j] <> notthere then 
               graphvertices := graphvertices + [j];
         end;
   for i := 1 to partialedon.length do
      for j := 1 to numob do
        begin
          localadj^[partialedon.arr[i],j] := adjacency^[partialedon.arr[i],j];
          if adjacency^[partialedon.arr[i],j] <> notthere then 
               graphvertices := graphvertices + [partialedon.arr[i],j];
        end;
   for i := 1 to partialedon.length do
    begin
       localadj^[partialedon.arr[i],partial[1]] := 0;
       localadj^[partialedon.arr[i],partial[2]] := 0;
    end;
   makelisrel1(localadj,localparm,comment,fname,partial[1],partial[2]);
   flush(eqsinp);
   close(eqsinp);
   graphvertices := oldgraphvert;
   dispose(localadj);
   dispose(localparm);
end;

function stringin(first: inbuf; var second: inbuf; length: integer): boolean;
var i: integer; match: boolean;
begin
    i := 1; match := true; 
    while (i <= length) and match do
	begin
	   if first[i] = second[i] then i := i + 1
	   else match := false;
	end;  
      stringin := match;
end;

procedure readtonum(var buf: inbuf);
var i: integer; c: char;
begin
   repeat
     readit(c);
   until (c in ['0'..'9','.']) or (curpos = buflength);
   curpos := curpos - 1;   
end;

procedure readtonum1(var buf: inbuf);
var i: integer; c: char;
begin
   curpos := 15;
   repeat
     readit(c);
   until (c in ['0'..'9','.','-']) or (curpos = buflength);
   curpos := curpos - 1;   
end;

(**********************************************************************)

procedure writelispart;
var localadj: adjtype; localparm: covarray; i,j,newlat: vertextype; 
    comment,fname: inbuf; oldgraphvert: vertexset; oldmax, num: integer;
begin
   oldgraphvert := graphvertices;
   graphvertices := [];
   oldmax := max;
   new(localadj);
   new(localparm);
   fname := 'blub.in';
   fname[8] := chr(0);
   comment[1] := chr(0);
   removefile; 
   rewritefile(eqsinp,fname);
   for i := 1 to maxvertex do
     for j := 1 to maxvertex do
       begin
         localadj^[i,j] := notthere;
         localparm^[i,j] := 0.0;
       end;
   for i := numob+1 to max do
    begin
      num := 0;
      for j := 1 to numob do
        begin
                localadj^[i,j] := adjacency^[i,j];
                if adjacency^[i,j] <> notthere then 
                   begin
                      num := num + 1;
                      graphvertices := graphvertices + [i,j];
                   end;
        end;
    end;
   newlat := oldmax + 1;
   for i := numob+1 to oldmax-1 do
      for j := i+1 to oldmax do 
        begin
          localadj^[newlat,i] := 1;
          localadj^[newlat,j] := 1;
          graphvertices := graphvertices + [i,j,newlat];
          vernames[newlat,1] := 'Z';
          vernames[newlat,2] := chr(ord('0') + (newlat - oldmax) div 100);
          vernames[newlat,3] := chr(ord('0') + ((newlat - oldmax) mod 100) div 10);
          vernames[newlat,4] := chr(ord('0') + ((newlat - oldmax) mod 100) mod 10);
          vernames[newlat,5] := chr(0);
          newlat := newlat + 1;
          max := max + 1;
        end;
   makelisrelfile(localadj,localparm,comment,fname);
   flush(eqsinp);
   close(eqsinp);
   graphvertices := oldgraphvert;
   max := oldmax;
   dispose(localadj);
   dispose(localparm);
end;
    
procedure readparts;
label 999;
var fname: inbuf; found: boolean; i,j: vertextype; k: integer;
begin
   lineno := 1;
   fname := 'C:\MIMTEST\BLUB.OUT';
   for k := 20 to buflength do fname[i] := chr(0);
   fname[0] := chr(19);
   resetfile(eqsres,fname,know);
   found := false;
   repeat 
       readfileline(eqsres,buf);
       found := stringin('         PHI                      ',buf,34);
   until found or eof(eqsres);
   if eof(eqsres) then goto 999;
   readfileline(eqsres,buf);
   readfileline(eqsres,buf);
   readfileline(eqsres,buf);
   for i := numob+1 to max do 
      begin
         readfileline(eqsres,buf);
         readtonum1(buf);
         for j := numob+1 to i do 
           begin
              rho^[i,j] := readnum(false,warning);
              rho^[j,i] := rho^[i,j];
           end;
        end;
999: end;

procedure calllisrel;
begin
   swapvectors;
   exec('C:\LIS8\LISREL8.EXE', 'C:\MIMTEST\BLUB.IN -o C:\MIMTEST\BLUB.OUT');
   swapvectors;
      if doserror <> 0 then
      writeln('dos error #',doserror)
   else writeln('exec successful. Child process exit code = ',dosexitcode);
end;

procedure lisparts;
begin
     writelispart;
     calllisrel;
     readparts; 
end;

(**********************************************************************)

procedure readeqsfileout(var val,prob: real; var held: boolean;
                     partial: partialpair; var partialedon: partialedonarray);
label 999;
var fname: inbuf; found: boolean; i: integer; head,tail: vertexnametype;
           headpos,tailpos: vertextype; wasinserted,done: boolean;
begin
   lineno := 1;
   writename(partial[1],ofilvar);
   write(ofilvar,' ');
   writename(partial[2],ofilvar);
   write(ofilvar,' . ');
   for i := 1 to partialedon.length do
     writename(partialedon.arr[i],ofilvar);
   flush(ofilvar);
   fname := 'blub.out';
   fname[0] := chr(8);
   resetfile(eqsres,fname,know);
   found := false;
   repeat 
       readfileline(eqsres,buf);
       if partialedon.length <> 0 then
         found := stringin('         MODIFICATION INDICES FOR BETA',buf,38)
       else found := stringin('         MODIFICATION INDICES FOR GAMMA',buf,39)
   until found or eof(eqsres);
   writebuf(buf,ofilvar);
   writeln(ofilvar);
   flush(ofilvar);
   if eof(eqsres) then 
      begin
         val := 100;
         held := true;
         prob := 0.0;
         goto 999;
      end;
   readfileline(eqsres,buf);
   readfileline(eqsres,buf);
   readfileline(eqsres,buf);
   readfileline(eqsres,buf);
   if partialedon.length <> 0 then readfileline(eqsres,buf);
   curpos := 12;
   readwhitespace;
   val := readnum(false,warning);
   if warning in errors then 
      begin
         val := 100;
         prob := 0;
         held := false;
         writeln(ofilvar,'Could not find number.');
         flush(ofilvar);
       end
    else
       begin
          chisquare(val,1,signif,prob);
          held := prob > signif;
        end;
999: end;

procedure partholdla(partial: partialpair; var partialedon: partialedonarray;
                     var val,prob: real; var held: boolean);
var i: integer;
begin
   writelis1(partial,partialedon);
   calllisrel; 
   readeqsfileout(val,prob,held,partial,partialedon); 
end;

(**********************************************************************)

procedure zerocorr(i1,j1,k1,l1,i2,j2,k2,l2: vertextype; 
                   var foundcommon: boolean);
var loc1,loc2,loc3,loc4: partialpair; nullset: partialedonarray;
begin
   loc1[1] := i1; loc1[2] := j1;
    loc2[1] := k1; loc2[2] := l1;
    loc3[1] := i2; loc3[2] := j2;
    loc4[1] := k2; loc4[2] := l2;
    nullset.length := 0;
    if (parthold(loc1,nullset) = 0) or (parthold(loc2,nullset) = 0) and
        (parthold(loc3,nullset) = 0) or (parthold(loc4,nullset) = 0) then
      foundcommon := true;
end;

procedure checkzero(i1,j1,k1,l1,i2,j2,k2,l2,k,depth: integer; 
                   var intersectarray,subarray: partialedonarray;
                   var foundcommon: boolean; var subset: vertexset);
var next,i1j1,k1l1,i2j2,k2l2: integer; 
    loc1,loc2,loc3,loc4: partialpair; nullset: partialedonarray;
begin
    subset := subset + [intersectarray.arr[k]];
    subarray.arr[depth] := intersectarray.arr[k];
    if (depth = subarray.length) then
       begin
          loc1[1] := i1; loc1[2] := j1;
          loc2[1] := k1; loc2[2] := l1;
          loc3[1] := i2; loc3[2] := j2;
          loc4[1] := k2; loc4[2] := l2;
          if (i1 in subset) or (j1 in subset) then i1j1 := 0
          else i1j1 := parthold(loc1,subarray);
          if (k1 in subset) or (l1 in subset) then k1l1 := 0
          else k1l1 := parthold(loc2,subarray);
          if (i2 in subset) or (j2 in subset) then i2j2 := 0
          else i2j2 := parthold(loc3,subarray);
          if (k2 in subset) or (l2 in subset) then k2l2 := 0
          else k2l2 := parthold(loc4,subarray);
          if (i1j1 = 0) and (k1l1 = 0) and (i2j2 = 0) and (k2l2 = 0) then
             foundcommon := true;
       end
    else if (depth < subarray.length) and (not foundcommon) then
         for next := k+1 to intersectarray.length-subarray.length+depth+1 do
                checkzero(i1,j1,k1,l1,i2,j2,k2,l2,next,depth+1,intersectarray,
                          subarray,foundcommon,subset);
    subset := subset - [intersectarray.arr[k]];
end;

procedure writelatent(i,j,k,l,pattern: integer);
begin
    if pattern = 1 then 
	begin
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' - |');
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' |');
	twrite(ofilvar,' needs a latent variable|');
	twriteeoln(ofilvar);
	end
    else if pattern = 2 then
	begin
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' - |');
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' |');
	twrite(ofilvar,' needs a latent variable|');
	twriteeoln(ofilvar);
	end
    else
	begin
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' - |');
	writename(i,ofilvar);twrite(ofilvar,' |');
	writename(l,ofilvar);twrite(ofilvar,' |');
	writename(j,ofilvar);twrite(ofilvar,' |');
	writename(k,ofilvar);twrite(ofilvar,' |');
	twrite(ofilvar,' needs a latent variable|');
	twriteeoln(ofilvar);
	end;
end;

{
procedure needlatent(num4,i,j,k,l,pattern: integer);
var zero1,zero2,zero3,zero4,intersect,subset: vertexset; n,length: integer;
    foundcommon: boolean; intersectarray,subarray: partialedonarray;
    
begin
    needlat[num4,pattern]:= false;
    if heldar^[num4][pattern] = 0 then
        begin
           foundcommon := false;
           getzeroes(zero1,zero2,zero3,zero4,i,j,k,l,pattern);
           intersect := zero1 * zero2 * zero3 * zero4;
           intersectarray.length := 0;
           for n := 0 to numob do 
               if n in intersect then 
                   begin
                      intersectarray.length := intersectarray.length + 1;
                      intersectarray.arr[intersectarray.length] := n;
                   end;
           length := 1;
           subset := [];
           while (length <= intersectarray.length) and (not foundcommon) do
               begin
                   subarray.length := length;
                   for n := 1 to intersectarray.length-length+1 do
                      if pattern = 1 then
                        checkzero(i,j,k,l,i,k,j,l,n,
                                  1,intersectarray,subarray,foundcommon,subset)
                      else if pattern = 2 then
                        checkzero(i,l,j,k,i,j,k,l,n,
                                  1,intersectarray,subarray,foundcommon,subset)
                      else
                        checkzero(i,k,j,l,i,l,j,k,n,
                                 1,intersectarray,subarray,foundcommon,subset);
                   length := length + 1;
                end;
           if (not foundcommon) then
             begin 
              needlat[num4,pattern]:= true;
              if wantpart then writelatent(i,j,k,l,pattern)
             end;
         end;
end;
}

procedure needlatent(i,j,k,l,pattern: vertextype);
label 999;
var zero1,zero2,zero3,zero4,intersect,subset: vertexset; n,length: integer;
    foundcommon: boolean; intersectarray,subarray: partialedonarray;
begin
     foundcommon := false;
     if pattern = 1 then zerocorr(i,j,k,l,i,k,j,l,foundcommon)
     else if pattern = 2 then zerocorr(i,l,j,k,i,j,k,l,foundcommon)
     else zerocorr(i,k,j,l,i,l,j,k,foundcommon);
     if foundcommon then goto 999;
     getzeroes(zero1,zero2,zero3,zero4,i,j,k,l,pattern);
     intersect := zero1 * zero2 * zero3 * zero4;
     intersectarray.length := 0;
     for n := 0 to numob do 
        if n in intersect then 
            begin
               intersectarray.length := intersectarray.length + 1;
               intersectarray.arr[intersectarray.length] := n;
            end;
        length := 1;
        subset := [];
        while (length <= intersectarray.length) and (not foundcommon) do
           begin
              subarray.length := length;
              for n := 1 to intersectarray.length-length+1 do
                if pattern = 1 then
                   checkzero(i,j,k,l,i,k,j,l,n,
                             1,intersectarray,subarray,foundcommon,subset)
                else if pattern = 2 then
                   checkzero(i,l,j,k,i,j,k,l,n,
                             1,intersectarray,subarray,foundcommon,subset)
                else
                   checkzero(i,k,j,l,i,l,j,k,n,
                             1,intersectarray,subarray,foundcommon,subset);
                length := length + 1;
           end;
       if (not foundcommon) and (i in graphvertices) and 
         (j in graphvertices) and (k in graphvertices) and
         (l in graphvertices) then writelatent(i,j,k,l,pattern);
999: end;

procedure checklatents;
label 999;
var i,j,k,l: vertextype; p,q,r: real; locresid,prob: array[1..3] of real;
begin
   for i := 1 to numob-3 do
      for j := i+1 to numob-2 do
          for k := j+1 to numob-1 do
             for l := k+1 to numob do
                if (i in graphvertices) and (j in graphvertices) and 
                   (k in graphvertices) and (l in graphvertices) then
                  begin
                     if aaabort then goto 999;
                     p := ((rho^[i,j] * rho^[k,l]));
                     q := ((rho^[i,k] * rho^[j,l]));
                     r := ((rho^[i,l] * rho^[j,k]));
                     prob[1]:= 0; prob[2]:= 0; prob[3]:= 0;
                     locresid[1] := (p - q);
                     calcstat(i,l,j,k,abs(locresid[1]),prob[1],1);
                     locresid[3] := (q - r);
                     calcstat(i,j,k,l,abs(locresid[3]),prob[3],3);
                     locresid[2] := (r - p);
                     calcstat(i,k,j,l,abs(locresid[2]),prob[2],2);
                     if prob[1] > signif then needlatent(i,j,k,l,1);
                     if prob[2] > signif then needlatent(i,j,k,l,2);
                     if prob[3] > signif then needlatent(i,j,k,l,3);
                  end;
999: end;

procedure deleteedge(var edge,listedge: nodeptr);
begin
 if edge <> nil then
   begin
      if edge^.back <> nil then edge^.back^.fwd := edge^.fwd
      else listedge := edge^.fwd;
      if edge^.fwd <> nil then edge^.fwd^.back := edge^.back;
      dispose(edge);
   end;
end;

procedure leafinsert(var curnode,leaflist: nodeptr);
var curpos: nodeptr; leaveloop,done: boolean;
begin
   curpos := leaflist;
   leaveloop := false;
   done := false;
   if leaflist = nil then 
       begin
         leaflist := curnode;
         leaflist^.fwd := nil;
         leaflist^.back := nil;
       end
   else
    begin
     while (curpos <> nil) and (not leaveloop) and (not done) do
        begin
           if curpos^.prob < curnode^.prob then 
              if curpos^.fwd = nil then
                  begin
                     curpos^.fwd := curnode;
                     curnode^.back := curpos;
                     curnode^.fwd := nil;
                     done := true;
                  end
              else curpos := curpos^.fwd
          else leaveloop := true;
        end;
     if not done then
       begin
         if curpos^.back <> nil then
            begin
              curpos^.back^.fwd := curnode;
              curnode^.back := curpos^.back;
              curpos^.back := curnode;
              curnode^.fwd := curpos;
            end
         else 
            begin
              curnode^.fwd := leaflist;
              curnode^.back := nil;
              leaflist^.back := curnode;
              leaflist := curnode;
            end;
       end;
   end;
end;

procedure moveedge(var edge,listedge1,listedge2: nodeptr);
begin
 if edge <> nil then
   begin
      if edge^.back <> nil then edge^.back^.fwd := edge^.fwd
      else listedge1 := edge^.fwd;
      if edge^.fwd <> nil then edge^.fwd^.back := edge^.back;
      edge^.fwd := nil;
      edge^.back := nil;
      leafinsert(edge,listedge2);
   end;
end;

procedure printpartial(part: partialpair; var partialedon: partialedonarray;
                    var ofilvar: text; prob,val: real);
var i,totalchars: integer; 
begin
   totalchars:=0;
   rswritename(part[1],ofilvar,totalchars);
   twrite(ofilvar,' -- |'); totalchars:= totalchars+4;
   rswritename(part[2],ofilvar,totalchars);
   for i:= 1 to (17 - totalchars) do twrite(ofilvar,' |');   
   totalchars:= 17;
   if (not indepdata) and (not discrete) then
      begin
        twrite(ofilvar,'rho(|'); totalchars:= totalchars+4;
      end;
   rswritename(part[1],ofilvar,totalchars);
   twrite(ofilvar,' |'); totalchars:= totalchars+1;
   rswritename(part[2],ofilvar,totalchars); 
   if (partialedon.length > 0) then 
     begin
       twrite(ofilvar,' . |'); 
       totalchars:= totalchars+3;
       for i := 1 to partialedon.length do
         begin
	   rswritename(partialedon.arr[i],ofilvar,totalchars);
           if i < partialedon.length then 
             begin
               twrite(ofilvar,' |');
               totalchars:= totalchars+1;
             end;
         end;
     end;
   if (not indepdata) and (not discrete) then
        begin
          twrite(ofilvar,') |'); totalchars:= totalchars+2;
        end;
   if totalchars > 48 then begin twriteeoln(ofilvar); totalchars:= 0; end;
   for i:= 1 to (48 - totalchars) do twrite(ofilvar,' |');
   if (not indepdata) then
      begin
        twritereal(ofilvar,val); 
        twrite(ofilvar,'  |');
        twritereal(ofilvar,prob);
        twriteeoln(ofilvar);
      end;
end;

function checkindeplist(partial: partialpair; 
                         var partialedon: partialedonarray): boolean;
var curpos: indepptr; i,j: vertextype;
begin
   checkindeplist := false;
   curpos := indep1;
   while curpos <> nil do
     begin
       if ((curpos^.tail = partial[1]) and (curpos^.head = partial[2])) or
          ((curpos^.head = partial[1]) and (curpos^.tail = partial[2])) then
             begin
                checkindeplist := true;
                i := 1;
                for j := 1 to numob do 
                   if j in curpos^.cond then
                       begin
			   zeroes^[partial[1],partial[2]] := 
     			      zeroes^[partial[1],partial[2]] + [j];
                           partialedon.arr[i] := j;
                           i := i + 1;
                       end;
                 partialedon.length := i - 1;
                 curpos := nil;
             end
        else curpos := curpos^.fwd;
     end;
end;

procedure checkpartials(edge: nodeptr; order,depth,location: integer;
                        var partialedon,vertices: partialedonarray; 
                        var ugraph: endarray; var edgelist: nodeptr;
                        var held: boolean);
label 999;
var i,j: integer; val,prob,g2: real; partial: partialpair;
    theseparents: vertexset;
begin
if (not held) then
 begin  
  partialedon.length := order;
  partial[1] := edgetovert^[edge^.edge].head; 
  partial[2] := edgetovert^[edge^.edge].tail;
  for i := location to vertices.length - (order - depth) + 1 do
   if (not held) then
    with partialedon do
      begin
        arr[depth+1] := vertices.arr[i];
        if depth <> order - 1 then
           checkpartials(edge,order,depth+1,i+1,partialedon,vertices,ugraph,edgelist,held)
        else
          begin
           if mimb = 1 then partholdla(partial,partialedon,g2,prob,held)
           else if indepdata then held := checkindeplist(partial,partialedon)
           else if (not discrete) then
             begin
               val := calcpart(partial,partialedon);
               calcpartialheld(val,prob,held,partialedon.length);
               if toomuchtime then goto 999;
             end
           else held := calcindependence(partial,prob,g2,partialedon,numchecked);
            if held then
              if ugraph^[partial[1],partial[2]] <> [] then

                 begin
                   numedges := numedges - 1;
                   if discrete or (mimb = 1) then
                      printpartial(partial,partialedon,ofilvar,prob,g2)
                   else printpartial(partial,partialedon,ofilvar,prob,val);
                   ugraph^[partial[1],partial[2]] := [];
                   parents[partial[1]] := parents[partial[1]] - [partial[2]];
                   parents[partial[2]] := parents[partial[2]] - [partial[1]];
                   if not causalsuff then
                    begin
                      readarraytoset(dsep^[partial[1],partial[2]],partialedon);
                      dsep^[partial[2],partial[1]] := 
                         dsep^[partial[1],partial[2]];
                    end;
                   for j := 1 to length do 
                     zeroes^[partial[1],partial[2]] := 
                       zeroes^[partial[1],partial[2]] + [arr[j]];
                 end;
           end;
        end; 
 end;
999: end;


procedure initugraph(var adj: adjtype; var ugraph: endarray); var i,j: integer; 
begin
   for i := bloop to eloop do
     for j := bloop to eloop do
        begin
          if adj^[i,j] <> notthere then
             ugraph^[i,j] := [knownin2]
          else if adj^[j,i] <> notthere then
             ugraph^[i,j] := [knownout2]
          else if i < j then ugraph^[i,j] := [notknown]
          else ugraph^[i,j] := [];
          dsep^[i,j] := [0];
        end;
end;

procedure initparents;
var i,j: integer;
begin
   for i := bloop to eloop do 
     if i in graphvertices then
       begin
          for j := bloop to eloop do 
            if j in graphvertices then
             begin
                if (i <> j) and (adjacency^[i,j] = notthere) then
                   parents[i] := parents[i] + [j];
                zeroes^[i,j] := [];
                if (forbiddenadj^[j,i] <> notthere) or
                   (j in temporder[i]) then 
                        parents[i] := parents[i] - [j];
                if causalsuff or (commonfor^[i,j] <> notthere) then
                    begin
                       if ((forbiddenadj^[i,j] <> notthere) and
                           (forbiddenadj^[j,i] <> notthere)) or 
                          ((j in temporder[i]) and 
                           (forbiddenadj^[i,j] <> notthere)) then
                               begin
                                 ugraph^[i,j] := [];
                                 ugraph^[j,i] := [];
                                 parents[i] := parents[i] - [j];
                                 parents[j] := parents[j] - [i];
                               end;
                    end
               end;
        end;
end;

procedure orientfromknow;
var i,j: integer;
begin
  for i := bloop to eloop-1 do
     for j := i+1 to eloop do
        if (ugraph^[i,j] <> []) and (adjacency^[i,j] = notthere) and
           (adjacency^[j,i] = notthere) then 
                begin
                    ugraph^[i,j] := [notknown];
                    ugraph^[j,i] := [notknown];
                 end;
  for i := bloop to eloop do
     for j := bloop to eloop do
       if (ugraph^[i,j] <> []) then
        if (j in temporder[i]) or (forbiddenadj^[j,i] <> notthere) then
              begin
                  ugraph^[i,j] := [knownin2];
                  if (commonfor^[i,j] <> notthere) or causalsuff then
                     ugraph^[j,i] := [knownout2]; 
               end;
end;

procedure disposeedgelist(var edgelist: nodeptr);
var curpos: nodeptr;
begin
  curpos := edgelist;
  while edgelist <> nil do
    begin
       edgelist := edgelist^.fwd;
       dispose(curpos);
       curpos := edgelist;
    end;
end;

procedure checkedges(var edgelist,newlist: nodeptr);
label 999;
var partial: partialpair; temp,curnode: nodeptr; theseparents: vertexset;
    vertices1,vertices2,partialedon: partialedonarray; i: integer; 
    held: boolean;
begin;
   i := 0;
   repeat
      curnode := edgelist;
      i := i + 1;
      while (curnode <> nil) do 
        begin
          if toomuchtime then goto 999;
          partial[1] := edgetovert^[curnode^.edge].head;
          partial[2] := edgetovert^[curnode^.edge].tail;
(* if 2 is later than 1, then 2 is not ancestor of 1, so check only 2 *)
          if (not (partial[2] in temporder[partial[1]])) then 
             begin
               held := false;
               theseparents := parents[partial[1]] - [partial[1],partial[2]];
               readsettoarray(theseparents,vertices1,partial[1],partial[2]);
               if (vertices1.length >= i) then
                 checkpartials(curnode,i,0,1,partialedon,vertices1,ugraph,edgelist,held);
             end;
(* if 1 is later than 2, then 1 is not ancestor of 2, so check only 1 *)
          if (not (partial[1] in temporder[partial[2]])) and (not held) then
             begin
               held := false;
               theseparents := parents[partial[2]] - [partial[1],partial[2]];
               readsettoarray(theseparents,vertices2,partial[1],partial[2]);
               if (vertices2.length >= i) then
                 checkpartials(curnode,i,0,1,partialedon,vertices2,ugraph,edgelist,held);
             end;
          if (ugraph^[partial[1],partial[2]] = []) then
             begin
                temp := curnode;
                curnode := curnode^.fwd;
                deleteedge(temp,edgelist)
             end
          else if ((vertices1.length <= i) and (vertices2.length <= i)) then
             begin
                temp := curnode;
                curnode := curnode^.fwd;
                moveedge(temp,edgelist,newlist);
             end
          else curnode := curnode^.fwd;
       end;
    until (edgelist = nil);
999: end;


function collide(tail,head,i: vertextype): boolean;
begin
   collide := 
      ((in2 in ugraph^[tail,head]) or (knownin2 in ugraph^[tail,head])) and
      ((in2 in ugraph^[i,head]) or (knownin2 in ugraph^[i,head]));
end;

procedure findcond(tail,head: vertextype; var condset,pathset: vertexset);
var i: vertextype;
begin
   for i := bloop to eloop do 
     begin
       if ((ugraph^[head,i] <> []) and (not (i in pathset)) and 
           (not (i in condset))) and 
          ((collide(tail,head,i) or triangle(i,head,tail)) or (tail = 0)) then        
         begin
              condset := condset + [i];
              pathset := pathset + [i];
              findcond(head,i,condset,pathset);
              pathset := pathset - [i];
         end;
      end;
end;

procedure unorient(var ugraph: endarray);
var i,j: vertextype;
begin
  for i := bloop to eloop do
    for j := bloop to eloop do
      if (i >= j) then ugraph^[i,j] := [] 
      else if ugraph^[i,j] <> [] then ugraph^[i,j] := [notknown];     
end;

procedure writprobmessage;
  begin
    twriteln(ofilvar,'------------------------------------------------------|');
    if (not discrete) and (not indepdata) then 
        twriteln(ofilvar,'List of vanishing (partial) correlations that made|')
    else twriteln(ofilvar,'List of (conditional) independence relations that made|');
    twriteln(ofilvar,'TETRAD remove edges from the graph. |');
    twriteeoln(ofilvar);
    if (not discrete) and (not indepdata) then 
      begin
       twriteln(ofilvar,'  Corr. :  Sample (Partial) Correlation|');
       twriteln(ofilvar,'  Prob. :  Probability that the absolute value of the sample |');
       twriteln(ofilvar,'           (partial) correlation exceeds the observed value, |');
       twriteln(ofilvar,'           on the assumption of zero (partial) correlation in|');
       twriteln(ofilvar,'           the population, assuming a multinormal distribution.|');
      end
    else if (not indepdata) then
      begin
       twriteln(ofilvar,'  Prob. :  Probability that the absolute value of the sample |');
       twriteln(ofilvar,'           G-square statistic exceeds the observed value, |');
       twriteln(ofilvar,'           on the assumption of (conditional) independence in|');
       twriteln(ofilvar,'           the population, assuming a multinomial distribution.|');
      end;
    twriteeoln(ofilvar);
    if (not discrete) and (not indepdata) then
      begin
        twriteln(ofilvar,'Edge             (Partial)|');  
        twriteln(ofilvar,'Removed          Correlation                     Corr.     Prob.|');
        twriteln(ofilvar,'-------          -----------                     -----     -----|');
      end
    else if (not indepdata) then
      begin
        twriteln(ofilvar,'Edge             (Conditional)|');  
        twriteln(ofilvar,'Removed          Independence                    G-square  Prob.|');
        twriteln(ofilvar,'-------          ------------                    --------  -----|');
      end
    else
      begin
        twriteln(ofilvar,'Edge             (Conditional)|');  
        twriteln(ofilvar,'Removed          Independence|');
        twriteln(ofilvar,'-------          ------------    |');
      end;
  end;

procedure getpartialnonsuff; 
var i,j: vertextype; curpos: edgeptr; partial: partialpair; num: integer;
    prob,g2: real; partialedon,vertices: partialedonarray; c: char;
    held: boolean; condset,pathset,theseparents: vertexset; nochange: boolean;
    temp,edgelist,newlist,curnode: nodeptr; var adj: adjtype;
label 999;
begin
   tnewadj(adj,fatal);
   if fatal in errors then goto 999;
   writprobmessage;
   numchecked := 1;
   edgelist := nil;
   newlist := nil;
   num := 1;
   toomuchtime := false;
   start := tgettime;
   numedges := round(numob * (numob - 1) / 2);
   if discrete then 
      if sizeof(cellsarr) < heapavail then new(cells)
      else 
         begin
            errormess('Out of memory.|',fatal,false);
            goto 999;
         end;
   initugraph(adjacency,ugraph);
   partialedon.length := 0;
   initparents;
   for i := bloop to eloop - 1 do
      for j := i + 1 to eloop do
        if (i in graphvertices) and (j in graphvertices) then
         begin
            partial[1] := i; partial[2] := j; 
            if mimb=1 then partholdla(partial,partialedon,g2,prob,held)
            else if (not discrete) and (not indepdata) then 
                 calcpartialheld(rho^[i,j],prob,held,0)
            else if (not indepdata) then
                 held := calcindependence(partial,prob,g2,partialedon,numchecked)
            else held := checkindeplist(partial,partialedon);
            if held then 
               begin
                  ugraph^[i,j] := [];
                  ugraph^[j,i] := [];
                  dsep^[i,j] := [];
                  dsep^[j,i] := [];
                  numedges := numedges - 1;
                  parents[i] := parents[i] - [j];
                  parents[j] := parents[j] - [i];
                  if discrete or (mimb=1) then
                      printpartial(partial,partialedon,ofilvar,prob,g2)
                  else printpartial(partial,partialedon,ofilvar,prob,rho^[i,j]);
               end
            else
               begin
                  new(curnode);
                  if discrete then curnode^.prob := g2
                  else curnode^.prob := rho^[i,j];
                  curnode^.edge := edgenames^[i,j];
                  leafinsert(curnode,edgelist);
               end;
          end;
    checkedges(edgelist,newlist);
    disposeedgelist(edgelist);
    if toomuchtime then 
	twriteln(ofilvar,'Search aborted because time limits exceeded.|');
    if not heuristic then
       begin
          orientfromknow;
          setcollider(false,nocollidelist,nochange,adj);    
          for i := bloop to eloop do 
              begin
                 condset := [];
                 pathset := [i];
                 findcond(0,i,condset,pathset);
                 parents[i] := condset;
              end;            
          checkedges(newlist,edgelist);
          unorient(ugraph);
       end;
    disposeedgelist(edgelist);
    disposeedgelist(newlist);
    orientfromknow;
    if not heuristic then orientedges(deftrilist,posstrilist,nocollidelist,adj)
    else 
      begin
         setcollider(true,nocollidelist,nochange,adj);
         calcdpath(ugraph,adj);
      end;
    twriteeoln(ofilvar);
    twriteln(ofilvar,'-----------------------------------------------------|');
    twriteln(ofilvar,'List of Causal Relations: |');
    twriteln(ofilvar,'NOT assuming causal sufficiency|');
    printthepattern(adj);
    if discrete then dispose(cells);
    dispose(adj);
999: end;


procedure fillpartialedon(var partialedon: partialedonarray;  
                          partial: partialpair; k,depth,length: integer);
label 999;
var next,i: vertextype; val,prob: real; held: boolean;
begin
   partialedon.arr[depth] := k;
   if (length = 0) or (depth = length) then
      begin
         val := calcpart(partial,partialedon);
         calcpartialheld(val,prob,held,partialedon.length);
         if held then 
           begin
            for i := 1 to partialedon.length do 
                zeroes^[partial[1],partial[2]] := 
                   zeroes^[partial[1],partial[2]] + [partialedon.arr[i]];
            zeroes^[partial[2],partial[1]] :=  zeroes^[partial[1],partial[2]];
           end;
         if toomuchtime then goto 999;
      end
   else 
      for next := k + 1 to eloop do
        if (next <> partial[1]) and (next <> partial[2]) and 
           (next in graphvertices) then
         fillpartialedon(partialedon,partial,next,depth+1,length);
999: end;

procedure findlatents;
var i,j,k,thelength,length: vertextype; partialedon: partialedonarray; 
    partial: partialpair;
begin
  for i := 1 to numob do
    for j := 1 to numob do zeroes^[i,j] := [i,j];
  thelength := setsize(graphvertices);
  for length := 0 to thelength - 2 do
    begin
       partialedon.length := length;
       for i := 1 to numob - 1 do
         for j := i + 1 to numob do
           if (i in graphvertices) and (j in graphvertices) then
             begin
                partial[1] := i;
                partial[2] := j;
                if (length = 0) then                     
                   fillpartialedon(partialedon,partial,k,1,length)
                else
                  for k := 1 to numob do 
                    if (k <> partial[1]) and (k <> partial[2]) and
                       (k in graphvertices) then
                         fillpartialedon(partialedon,partial,k,1,length);
             end;
     end;
  if not aaabort then checklatents;
end;

procedure getpartialssuff;
label 999;
var i,j: vertextype; curpos: edgeptr; partial: partialpair; 
    totalchars, k,num: integer; prob,g2: real; 
    partialedon,vertices1,vertices2: partialedonarray; c: char;
    held,nochange: boolean; theseparents: vertexset; 
    temp,edgelist,curnode: nodeptr; adj: adjtype;
begin
   tnewadj(adj,fatal);
   if fatal in errors then goto 999;
   writprobmessage;
   nocollidelist := nil;
   numchecked := 1;
   edgelist := nil;
   num := 1;
   toomuchtime := false;
   start := tgettime;
   numedges := round((eloop - bloop + 1) * (eloop - bloop) / 2);
   if discrete then
      if sizeof(cellsarr) < heapavail then new(cells)
      else 
         begin
            writeln(heapavail);
            errormess('Out of memory.|',fatal,false);
            goto 999;
         end;
   initugraph(adjacency,ugraph);
   partialedon.length := 0;
   initparents;
   for i := bloop to eloop - 1 do
      for j := i + 1 to eloop do
        if (i in graphvertices) and (j in graphvertices) then
         begin
            partial[1] := i; partial[2] := j; 
            if mimb=1 then partholdla(partial,partialedon,g2,prob,held)  
            else if (not discrete) and (not indepdata) then
                 calcpartialheld(rho^[i,j],prob,held,0)
            else if (not indepdata) then
                held := calcindependence(partial,prob,g2,partialedon,numchecked)
            else held := checkindeplist(partial,partialedon);
            if held then 
               begin
                  ugraph^[i,j] := [];
                  ugraph^[j,i] := [];
                  numedges := numedges - 1;
                  parents[i] := parents[i] - [j];
                  parents[j] := parents[j] - [i];
                  if discrete or (mimb = 1) then
                     printpartial(partial,partialedon, ofilvar,prob,g2)
                  else printpartial(partial,partialedon, ofilvar,prob,rho^[i,j]);
               end
            else
               begin
                  new(curnode);
                  if discrete then curnode^.prob := g2
                  else curnode^.prob := rho^[i,j];
                  curnode^.edge := edgenames^[i,j];
                  leafinsert(curnode,edgelist);
               end;
          end;
   i := 0;
   repeat
      curnode := edgelist;
      i := i + 1;
      while (curnode <> nil) and (not toomuchtime) do
        begin
          partial[1] := edgetovert^[curnode^.edge].head;
          partial[2] := edgetovert^[curnode^.edge].tail;
          theseparents := parents[partial[1]] - [partial[1],partial[2]];
          readsettoarray(theseparents,vertices1,partial[1],partial[2]);
          held := false;
          if (vertices1.length >= i) then
             checkpartials(curnode,i,0,1,partialedon,vertices1,ugraph,edgelist,held);
          theseparents := parents[partial[2]] - [partial[1],partial[2]];
          readsettoarray(theseparents,vertices2,partial[1],partial[2]);
          if (vertices2.length >= i) and (not held) then
            begin
             held := false;
             checkpartials(curnode,i,0,1,partialedon,vertices2,ugraph,edgelist,held);
            end;
          if (ugraph^[partial[1],partial[2]] = []) or
             ((vertices1.length <= i) and (vertices2.length <= i)) then
             begin
                temp := curnode;
                curnode := curnode^.fwd;
                deleteedge(temp,edgelist)
             end
          else curnode := curnode^.fwd;
       end;
    until (edgelist = nil) or toomuchtime;
    disposeedgelist(edgelist);
    if toomuchtime then 
	 twriteln(ofilvar,'Search aborted because time limits exceeded.|');
    orientfromknow;
    setcollider(true,nocollidelist,nochange,adj);
(*    calcdpath(ugraph,adj); *)
    nochange := true;
    setcollider(true,nocollidelist,nochange,adj);
(*    repeat
        setcollider(true,nocollidelist,nochange,adj);
        if nochange then orientawaycycle(nochange,adj,nocollidelist);
    until nochange; *)
    twriteeoln(ofilvar);
    twriteln(ofilvar,'-----------------------------------------------------|');
    twriteln(ofilvar,'List of Causal Relations: |');
    twriteln(ofilvar,'   Assuming causal sufficiency|');
    twriteln(ofilvar,'   *: no orientation consistent with assumptions|');
    printcausalsuff;
    twriteeoln(ofilvar); twriteeoln(ofilvar);
    if discrete then dispose(cells);
    dispose(adj);
(*    if (numob <= 25) and (not discrete) and (not indepdata) then findlatents; *)
999: end;


procedure chi;
type bin = 0..1;
     binarray = array[bin,bin,bin,bin] of real;
     indarray = array[1..4,1..4,bin,bin] of real;
     margtype = array[1..4,bin] of real;
     difftype = array[1..4,1..4] of real;

var i,j,k,l: integer; g2,expected,c2,observed: real; file1name,file2name: inbuf;
   file1,file2: text; changed: boolean; thearr1,thearr2: binarray;
   index1,index2,index3,index4: bin; prob: real; expectedpair,joint: indarray;
   marginal: margtype; diffc2,diffg2: difftype;

   procedure init;
   var i,j: integer; k,l: bin;
   begin
      for i := 1 to 4 do 
         for j := 1 to 4 do
           begin
             for k := 0 to 1 do
               for l := 0 to 1 do
                  begin
                     marginal[i,k] := 0;
                     expectedpair[i,j,k,l] := 0;
                     joint[i,j,k,l] := 0;
                  end;
             diffg2[i,j] := 0;
             diffc2[i,j] := 0;
          end;
   end;

   procedure fillinjoint(var1,var2: integer);
   var i,j,k,l: bin; index1,index2: integer;
   begin
     for i := 0 to 1 do
       for j := 0 to 1 do
         for k := 0 to 1 do
           for l := 0 to 1 do
             begin
               if var1 = 1 then index1 := i
               else if var1 = 2 then index1 := j
               else if var1 = 3 then index1 := k
               else index1 := l;
               if var2 = 1 then index2 := i
               else if var2 = 2 then index2 := j
               else if var2 = 3 then index2 := k
               else index2 := l;
               joint[var1,var2,index1,index2] := 
                   joint[var1,var2,index1,index2] + thearr1[i,j,k,l];
             end;
    end;

    procedure fillinexpect(var1,var2: integer);
    var i,j: bin;
    begin
      for i := 0 to 1 do
         for j := 0 to 1 do
            expectedpair[var1,var2,i,j] := marginal[var1,i] * marginal[var2,j]/
               sampsize;
    end;
   
   procedure fillinmarginal;
   var i,j,k,l: bin; 
   begin
     for i := 0 to 1 do
       for j := 0 to 1 do
         for k := 0 to 1 do
           for l := 0 to 1 do
             begin
               marginal[1,i] := marginal[1,i] + thearr1[i,j,k,l];
               marginal[2,j] := marginal[2,j] + thearr1[i,j,k,l];
               marginal[3,k] := marginal[3,k] + thearr1[i,j,k,l];
               marginal[4,l] := marginal[4,l] + thearr1[i,j,k,l];
             end;
     end;

       
    procedure getdiff(var1,var2: integer);
    var i,j: bin;
    begin
       for i := 0 to 1 do
         for j := 0 to 1 do
           begin
            diffg2[var1,var2] := joint[var1,var2,i,j] * 
               ln(joint[var1,var2,i,j]/expectedpair[var1,var2,i,j]);
            diffc2[var1,var2] := 
               diffc2[var1,var2] + 
                  sqr(joint[var1,var2,i,j] - expectedpair[var1,var2,i,j])/
                  expectedpair[var1,var2,i,j];
           end;
       end;

     procedure finddiff;
     var i,j: integer; k,l: bin;
     begin
       init;
       fillinmarginal;
       for i := 1 to 3 do
         for j := i + 1 to 4 do 
           begin
              fillinjoint(i,j);
              fillinexpect(i,j);
              getdiff(i,j);
           end;
        end;

                  

begin
  file1name := nullfilename;
  getfile(file1,'file 1: ',know,file1name,changed);
  i := 0;
  repeat
     i := i + 1;
     read(file1,index1);
     read(file1,index2);
     read(file1,index3);
     read(file1,index4);
     readln(file1,prob);
     thearr1[index1,index2,index3,index4] := prob * sampsize;
  until i = 16;
  file2name := nullfilename;
  getfile(file2,'file 2: ',know,file2name,changed); 
  i := 0;
  repeat
     i := i + 1;
     read(file2,index1);
     read(file2,index3);
     read(file2,index2);
     readln(file2,index4);
     readln(file2,prob);
     thearr2[index1,index2,index3,index4] := prob;
  until i = 16;
  i := 0;
  g2 := 0;
  c2 := 0;
  for i := 0 to 1 do
   for j := 0 to 1 do
    for k := 0 to 1 do
      for l := 0 to 1 do
        begin
           expected := thearr1[i,j,k,l];
           observed := thearr2[i,j,k,l];
           c2 := c2 + sqr(observed - expected)/expected;
           g2 := g2 + (observed * ln(observed/expected)); 
        end;
     finddiff;
end;


procedure getpartials;
var madedsep, madeugraph, madezeroes: boolean; 
    partial: partialpair; partialedon: partialedonarray;
    val,prob: real; held: boolean;
begin
   mimb := 0;
   write('Mimbuild type?');
   readln(mimb);
   if mimb <> 0 then
      begin
        bloop := numob + 1;
        eloop := max;
      end
   else 
      begin
        bloop := 1;
        eloop := numob;
      end;
   if mimb = 2 then lisparts;
   madedsep := false;
   madeugraph := false;
   madezeroes := false;
   tnewdsep(dsep,fatal);
   if (not (fatal in errors)) then madedsep := true;
   tnewend(ugraph,fatal);
   if (not (fatal in errors)) then madeugraph := true;
   tnewdsep(zeroes,fatal);
   if (not (fatal in errors)) then madezeroes := true;
   if (not (fatal in errors)) then
      begin
         causalsuff := getreply('Assume no latent common causes? |','t',true);
         if causalsuff then getpartialssuff
         else getpartialnonsuff; 
      end;
     if madedsep then dispose(dsep);
     if madeugraph then dispose(ugraph);
     if madezeroes then dispose(zeroes);
end;

end.
