#include "syst.h" 
#include "tt1.h" 
#include "tt2.h" 
#include "tt3.h"
#include "tt4.h" 
#include "tt5.h" 
#include "io.h"
#include "part.h"
#include "recalc.h"
#include "numeric.h"
#include "calctet.h"
#include "powell.h"

const
   ftol=1.0e-6;
type
   glnarray = array [1..maxvertex] of real;
   glndim = glnarray;
   glnpbynp = array [1..maxvertex,1..maxvertex] of real;
   gl3array = glnarray;
   pindexrec = record
                 cause,effect,atype: vertextype;
               end;
   pindexarr = array[1..maxvertex] of pindexrec;  (* could be more parameters than this *)

   verarray = array[1..maxvertex,1..maxvertex] of vertextype;
      
var
   ncom : integer;
   pcom,xicom : glnarray; 
   fret : real;
   iter,np : integer;
   p : glnarray;
   xi : glnpbynp;
   pindex: pindexarr;

function neglog(var cov,tcov: covarray): real;
label 999;
var i,j: vertextype; tempnum: real; ok: boolean;
begin
  for i:= 1 to numob do
    for j:=1 to numob do
     tempcov^[i,j]:= tcov^[i,j];
  ok := true;
  inverse(tempcov,theinverse,numob,ok);  (* puts inverse of imp.covs in ginverse *)
  sqmatmult(cov,theinverse,tempcov,numob);  (* stores answer in tcov *)
  tempnum := determ(tcov,numob,ok);
  if tempnum < 0 then neglog := -1
  else neglog:= (sampsize/2)*(ln(tempnum) + trace(tempcov,numob));
999: end;

procedure across(x,y: vertextype; var tcov: covarray);
begin
  with theorder do
   begin
        for i := 1 to tierpos[x] - 1 do 
           for j := 1 to tierpos[y] - 1 do 
             if (x <> y) or (i <= j) then
               begin
                 if parm^[tier[x,i],tier[y,j]] =
                    parm^[tier[y,j],tier[x,i]] then 
                     begin
                       tcov^[tier[x,i],tier[y,j]] := parm^[tier[x,i],tier[y,j]];
                       tcov^[tier[y,j],tier[x,i]] := parm^[tier[x,i],tier[y,j]];
                     end;
                 for k := 1 to numparents[tier[y,j]] do 
                     begin
                        tcov^[tier[x,i],tier[y,j]] := 
                           tcov^[tier[x,i],tier[y,j]] +
                              tcov^[tier[x,i],parents[tier[y,j],k]] *
                              parm^[parents[tier[y,j],k],tier[y,j]];
                        tcov^[tier[y,j],tier[x,i]] := tcov^[tier[x,i],tier[y,j]];
                     end;
              end;
    end
end;

procedure formimp(var parm,tcov: covarray);
var i,j: vertextype;
begin
   for i := 1 to max do
     for j := 1 to max do
        tcov^[i,j] := 0;
   for i := 1 to theorder.maxtiers do
     for j := 1 to i do
        across(j,i,tcov);
end;

function doubleheaded(i,j: vertextype): boolean;
begin
   doubleheaded := (parm^[i,j] <> 0) and (parm^[i,j] = parm^[j,i]) and (i <> j);
end;

function findparam(i: integer): integer;
var curpos: integer;
begin
   curpos := 1;
   while (pindex[curpos].atype <> 1) or (pindex[curpos].cause <> i) do 
      curpos := curpos + 1;
   findparam := curpos;
end;

function fnc(x: gl3array): real;
(* make sure parm is initialized to zero *)
var i,j: vertextype; tempnum: real; foundone: boolean; k: integer;
begin
   for i := 1 to ncom do
      if pindex[i].atype = 1 then 
         begin
           if x[i] > 0 then parm^[pindex[i].cause,pindex[i].effect] := x[i]
           else
              begin
                 parm^[pindex[i].cause,pindex[i].effect] := 0.001;
                 x[i] := 0.001;
              end;
         end
      else if pindex[i].atype = 2 then 
          parm^[pindex[i].cause,pindex[i].effect] := x[i]
      else if pindex[i].atype = 3 then
        begin
          parm^[pindex[i].cause,pindex[i].effect] := x[i];
          parm^[pindex[i].effect,pindex[i].cause] := x[i];
        end;
   formimp(parm,tcov);
   tempnum := neglog(cov,tcov);
   while (tempnum < 0) do 
     begin
       i := 1;
       while (i <= max) and (tempnum < 0) do
         begin
            j := 1;
            foundone := false;
            while (j <= max) and (not foundone) do
               begin
                  if doubleheaded(i,j) then 
                    begin
                      k := findparam(i);
                      x[k] := 2 * x[k];
                      parm^[i,i] := 2 * parm^[i,i];
                      foundone := true;
                    end
                  else j := j + 1;
                end;
            if foundone then
               begin
                  formimp(parm,tcov);
                  tempnum := neglog(cov,tcov);
               end;
            i := i + 1;
        end;
    end;
  fnc := tempnum;
end;

function f1dim(x: real): real;
var
   j: integer;
   xt: glndim;
begin
   for j := 1 to ncom do begin
      xt[j] := pcom[j]+x*xicom[j]
   end;
   f1dim := fnc(xt)
end;

function func(x: real): real;
begin
   func := f1dim(x)
end;

procedure mnbrak(var ax,bx,cx,fa,fb,fc: real);
label 1;
const
   gold=1.618034;
   glimit=100.0;
   tiny=1.0e-20;
var
   ulim,u,r,q,fu,dum: real;
function max(a,b: real): real;
   begin
      if (a > b) then max := a else max := b
   end;
function sign(a,b: real): real;
   begin
      if (b > 0.0) then sign := abs(a) else sign := -abs(a)
   end;
begin
   fa := func(ax);
   fb := func(bx);
   if (fb > fa) then begin
      dum := ax;
      ax := bx;
      bx := dum;
      dum := fb;
      fb := fa;
      fa := dum
   end;
   cx := bx+gold*(bx-ax);
   fc := func(cx);
1:   if (fb >= fc) then begin
      r := (bx-ax)*(fb-fc);
      q := (bx-cx)*(fb-fa);
      u := bx-((bx-cx)*q-(bx-ax)*r)/
         (2.0*sign(max(abs(q-r),tiny),q-r));
      ulim := bx+glimit*(cx-bx);
      if ((bx-u)*(u-cx) > 0.0) then begin
         fu := func(u);
         if (fu < fc) then begin
            ax := bx;
            fa := fb;
            bx := u;
            fb := fu;
            goto 1 end
         else if (fu > fb) then begin
            cx := u;
            fc := fu;
            goto 1
         end;
         u := cx+gold*(cx-bx);
         fu := func(u)
      end else if  ((cx-u)*(u-ulim) > 0.0) then begin
         fu := func(u);
         if (fu < fc) then begin
            bx := cx;
            cx := u;
            u := cx+gold*(cx-bx);
            fb := fc;
            fc := fu;
            fu := func(u)
         end
      end else if  ((u-ulim)*(ulim-cx) >= 0.0) then begin
         u := ulim;
         fu := func(u)
      end else begin
         u := cx+gold*(cx-bx);
         fu := func(u)
      end;
      ax := bx;
      bx := cx;
      cx := u;
      fa := fb;
      fb := fc;
      fc := fu;
      goto 1
   end
end;

function mybrent(ax,bx,cx,tol: real; var xmin: real): real;
label 999, 1, 2, 3;
const
   itmax=100;
   cgold=0.3819660;
   zeps=1.0e-10;
var
   a,b,d,e,etemp: real;
   fu,fv,fw,fx: real;
   iter: integer;
   p,q,r,tol1,tol2: real;
   u,v,w,x,xm: real;
function sign(a,b: real): real;
   begin
      if (b > 0.0) then sign := abs(a) else sign := -abs(a)
   end;
begin
   if ax < cx then a := ax else a := cx;
   if ax > cx then b := ax else b := cx;
   v := bx;
   w := v;
   x := v;
   e := 0.0;
   fx := func(x);
   fv := fx;
   fw := fx;
   for iter := 1 to itmax do begin
      xm := 0.5*(a+b);
      tol1 := tol*abs(x)+zeps;
      tol2 := 2.0*tol1;
      if (abs(x-xm) <= (tol2-0.5*(b-a))) then goto 3;
      if (abs(e) > tol1) then begin
         r := (x-w)*(fx-fv);
         q := (x-v)*(fx-fw);
         p := (x-v)*q-(x-w)*r;
         q := 2.0*(q-r);
         if (q > 0.0) then  p := -p;
         q := abs(q);
         etemp := e;
         e := d;
         if((abs(p) >= abs(0.5*q*etemp)) or (p <= q*(a-x))
            or (p >= q*(b-x))) then goto 1;
         d := p/q;
         u := x+d;
         if (((u-a) < tol2) or ((b-u) < tol2)) then d := sign(tol1,xm-x);
         goto 2
      end;
1:      if (x >= xm)  then e := a-x else e := b-x;
      d := cgold*e;
2:      if (abs(d) >= tol1)  then u := x+d else u := x+sign(tol1,d);
      fu := func(u);
      if (fu <= fx)  then begin
         if (u >= x)  then a := x else b := x;
         v := w;
         fv := fw;
         w := x;
         fw := fx;
         x := u;
         fx := fu
      end else begin
         if (u < x)  then a := u else b := u;
         if ((fu <= fw) or (w = x))  then begin
            v := w;
            fv := fw;
            w := u;
            fw := fu
         end else if ((fu <= fv) or (v = x) or (v = 2)) then begin
            v := u;
            fv := fu
         end
      end
   end;
   writeln('pause in routine brent - too many iterations');
   mybrent := 100000000;
   xmin := 10000000;
   goto 999;
3:   xmin := x;
   mybrent := fx
999: ;
end;

procedure linmin(var p,xi: glnarray; n: integer; var fret: real);
const
   tol=1.0e-4;
var
   j: integer;
   xx,xmin,fx,fb,fa,bx,ax: real;
begin
   ncom := n;
   for j := 1 to n do begin
      pcom[j] := p[j];
      xicom[j] := xi[j]
   end;
   ax := 0.0;
   xx := 1.0;
   bx := 2.0;
   mnbrak(ax,xx,bx,fa,fx,fb);
   fret := mybrent(ax,xx,bx,tol,xmin);
   for j := 1 to n do begin
      xi[j] := xmin*xi[j];
      p[j] := p[j]+xi[j]
   end
end;

procedure powell(var p: glnarray; var xi: glnpbynp; n,np: integer;
       ftol: real; var iter: integer; var fret: real);
label 1,99;
const
   itmax=200;
var
   j,ibig,i: integer;
   t,fptt,fp,del: real;
   pt,ptt,xit: glnarray;
begin
   fret := fnc(p);
   for j := 1 to n do begin
      pt[j] := p[j]
   end;
   iter := 0;
1: iter := iter+1;
   fp := fret;
   ibig := 0;
   del := 0.0;
   for i := 1 to n do begin
      for j := 1 to n do begin
         xit[j] := xi[j,i]
      end;
      fptt := fret;
      linmin(p,xit,n,fret);
      if (abs(fptt-fret) > del) then begin
         del := abs(fptt-fret);
         ibig := i
      end
   end;
   if (2.0*abs(fp-fret) <= ftol*(abs(fp)+abs(fret))) then goto 99;
   if (iter = itmax) then begin
      writeln('pause in routine powell');
      writeln('too many interations'); readln
   end;
   for j := 1 to n do begin
      ptt[j] := 2.0*p[j]-pt[j];
      xit[j] := p[j]-pt[j];
      pt[j] := p[j]
   end;
   fptt := fnc(ptt);
   if (fptt >= fp) then goto 1; 
   t := 2.0*(fp-2.0*fret+fptt)*sqr(fp-fret-del)-del*sqr(fp-fptt);
   if (t >= 0.0) then goto 1; 
   linmin(p,xit,n,fret);
   for j := 1 to n do begin
      xi[j,ibig] := xit[j]
   end;
   goto 1;
99:   end;

procedure putintier(x,tiernum: vertextype; var thenext: vertextype);
var i: vertextype;
begin
  with theorder do
   begin
    tier[tiernum,tierpos[tiernum]] := x;
    tierpos[tiernum] := tierpos[tiernum] + 1;
    for i := 1 to numchildren[x] do
       begin
          numparents1[children[x,i]] := numparents1[children[x,i]] - 1;
          if numparents1[children[x,i]] = 0 then
              begin
                 nextinline[thenext].avar := children[x,i];
                 nextinline[thenext].tier := tiernum + 1;
                 if (tiernum + 1) > maxtiers then maxtiers := tiernum + 1;
                 thenext := thenext + 1;
              end;
       end;
     end;
end;

procedure myeqsorder(var mag: graphrec);
var i,j,thenext: vertextype;
begin
 theorder.maxtiers := 1;
 thenext := 1;
 with theorder do
  begin
   for i := 1 to max do
      begin
         tierpos[i] := 1;
         numparents[i] := 0;
         numparents1[i] := 0;
         numchildren[i] := 0;
      end;
   for i := 1 to max do
     for j := 1 to max do
       begin
          if (in2 in mag.augraph^[i,j]) and (out2 in mag.augraph^[j,i]) then
             begin
                numparents[j] := numparents[j] + 1;
                numparents1[j] := numparents1[j] + 1;
                numchildren[i] := numchildren[i] + 1;
                children[i,numchildren[i]] := j;
                parents[j,numparents[j]] := i;
             end;
       end;
    for i := 1 to max do if (numparents[i] = 0) then putintier(i,1,thenext);
    i := 1;
    while (i < thenext) do
       begin
         putintier(nextinline[i].avar,nextinline[i].tier,thenext);
         i := i + 1;
       end;
  end;
end;

procedure initscore(var mag: graphrec);
var i,j: vertextype;
begin
   np := 0;
   myeqsorder(mag);
   with mag do 
     begin
       for i := 1 to max do 
         for j := 1 to max do
            begin
              if i = j then 
                begin
                  np := np + 1;
                  p[np] := 1;
                  pindex[np].cause := i;
                  pindex[np].effect := j;
                  pindex[np].atype := 1;
                  parm^[i,j] := 1;
                end
              else if (in2 in augraph^[i,j]) and (out2 in augraph^[j,i]) then
                begin
                  np := np + 1;
                  p[np] := 0;
                  pindex[np].cause := i;
                  pindex[np].effect := j;
                  pindex[np].atype := 2;
                  parm^[i,j] := 0;
                end
              else if (in2 in augraph^[i,j]) and (in2 in augraph^[j,i]) 
                      and (i < j) then
                begin
                  np := np + 1;
                  p[np] := 0;
                  pindex[np].cause := i;
                  pindex[np].effect := j;
                  pindex[np].atype := 3;
                  parm^[i,j] := 0;
                end
              else 
                begin
                  parm^[i,j] := 0;
                end;
            end;
       for i := 1 to np do 
         for j := 1 to np do
           if i = j then xi[i,j] := 1.0 else xi[i,j] := 0;
      end;
end;       
    
procedure scoremag(* var mag: graphrec; var score: real; var ok: boolean *);
var ndim: integer; df: real;
begin
   initscore(mag);
   ndim := np;
   powell(p,xi,ndim,np,ftol,iter,fret);
   df := ((max * (max + 1))/2) - ndim;
   score := -1 * (fret - df * ln(sampsize * max));
end;

procedure printxi;
var i,j: vertextype;
begin
   for i := 1 to np do
      for j := 1 to np do
        write('i = ',i,' j = ',j,' xi = ',xi[i,j]);
end;
