#include "syst.h"
#include "tt1.h"
#include "tt2.h"
#include "tt3.h"
#include "tt4.h"
#include "tt5.h"
#include "io.h"
#include "util.h"
#include "calctet.h"
#include "sys.h"
#include "eqs.h"
#include "cspecial.h"
#include "part.h"
#include "up.h"
#include "testbuildm.h"
#include "newcalc.h"
#include "matrices.h"
#include "mon.h"
#include "linbayes.h"

var patfile: text; patfilename: inbuf; oldmax: integer;

procedure randomgraph1;
label 999;
var ofile: text; ofilename: packed array[1..20] of char;
    avedegree,aprob,randomnum: real;  newseed,numvar,numpair: integer;
    ingraph: vertexset; i,j: vertextype; numedge: integer; currentmod: modeltype;

begin
   initrand(newseed);
   for i := 1 to max do currentmod.amodel[i] := [];
   numedges := 1;
   ingraph := [];
   write('Output file: ');
   readln(ofilename);
   rewrite(ofile,ofilename);
   write('Number of variables: ');
   readln(numvar);
   write('Average degree: ');
   readln(avedegree);
   numpair := (numvar * (numvar - 1)) div 2;
   aprob := (avedegree * numvar) / (2 * numpair);
   writeln(ofile,'/graph');
   repeat
      for i := 1 to numvar - 1 do
         for j := i + 1 to numvar do
           begin
              randomnum := thisrand;
              if randomnum <= aprob then
                if (not (i in currentmod.amodel[j])) then
                  begin
                      writeln(ofile,'x',i:0,' ','x',j:0);
                      ingraph := ingraph + [i,j];
                      numedges := numedges + 1;
                      currentmod.amodel[j] := currentmod.amodel[j] + [i];
                     if (numedges > round(numvar * avedegree/2)) then goto 999;
                  end;
           end
   until numedges > round(numvar * avedegree/2);
999:   for i := 1 to numvar do 
      if (not (i in ingraph)) then writeln(ofile,'x',i:0);
end;

procedure readpattern(* infromtty: boolean; add: boolean; var fileread: text;
		      var pattern: patternarray; var vertexnames: namearray *);

label 999;
var emptyline,notfound,finished: boolean; c,c1: char; 
    head,tail: vertexnametype; i,j,headpos,tailpos: integer;
    sorter: array[1..maxvertex] of integer;
    obsingraph: vertexset; arrow1,arrow2,wasinserted: boolean;

begin (* readgraph *)
   if infromtty then lineno := 0 else lineno := 1;
   finished := false; twriteeoln(output);
   i := 1;
   buf[1] := 'c';
   while not eof(fileread) and (buf[1] <> '/') do 
         readfileline(fileread,buf);
   while not eof(fileread) and (buf[1] <> '!') do
      begin (*While*)
         emptyline := false;
         checkinterrupt;
         if (lastvarassigned = max2s) then
		begin
		twrite(output,'Maximum number of edges reached. No more edges allowed.|');
		twriteeoln(output);
		end;
         errors := errors - [ignoreall];
         head := '  '; tail := '  ';
	 if not infromtty then
 	    begin
         	readfileline(fileread,buf);
                if buf[1] = '!' then goto 999;
                if myeoln then emptyline := true;
	        if emptyline then goto 999 else
                 begin
                    readit(c);
                    if c = '{' then 
                      begin
                        readcomment(fileread);
                        while myeoln and (not eof(fileread)) do
                            readfileline(fileread,buf);
                      end
                    else curpos := curpos - 1;
                    if (fatal in errors) then goto 999;
                    arrow1 := false;
                    arrow2 := false;
                    readname(head,ignoreall);
                    readwhitespace;
                    readit(c);
                    if c = '<' then arrow1 := true;
                    repeat 
                        readit(c)
                    until c <> '-';
                    if c = '>' then arrow2 := true;
                    readwhitespace;
                    readname(tail,ignoreall);
	 	    lineno := lineno + 1;
               end;
	    end;
         if not emptyline then
           begin
             if (head = tail) and (head[1] <> ' ') then
	        myerror('Head not allowed to equal tail          ',ignoreall,
	     	         true);
                headpos := listposition(head,true,vertexnames,wasinserted);
                tailpos := listposition(tail,true,vertexnames,wasinserted);
                if not (ignoreall in errors) then
		    begin
			if pattern[headpos,tailpos].existence  <> notthere then
                           begin
                 	        twrite(output,'Edge:  |');
				writevername(head,output);
				twrite(output,' --> |');
				writevername(tail,output);
				twriteeoln(output);
 			        myerror('This edge already in graph              ',
			   	    ignoreall,true)
                           end
		        else
			   begin
                             if headpos < tailpos then
                               begin
                                  pattern[headpos,tailpos].existence := 0;
                                  pattern[headpos,tailpos].arrow1 := arrow1;
                                  pattern[headpos,tailpos].arrow2 := arrow2;
                               end
                             else 
                               begin
                                  pattern[tailpos,headpos].existence := 0;
                                  pattern[tailpos,headpos].arrow1 := arrow2;
                                  pattern[tailpos,headpos].arrow2 := arrow1;
                               end;
			   end;
			i := i + 1;
		    end;
                 end;
     end;
999: max := lastvarassigned - 1;
 end;  { readgraph }

function idirectionset(i,j: vertextype; var pattern: patternarray): boolean;
begin
   if i < j then
       idirectionset :=  (pattern[i,j].arrow1 or pattern[i,j].arrow2)
   else idirectionset :=  (pattern[j,i].arrow1 or pattern[j,i].arrow2)
end;

function icollider(i,j,k: integer; var adj: adjtype): boolean;
begin
      icollider := (adj^[i,k] <> notthere) and (adj^[j,k] <> notthere);
end;

procedure isetdirection(tail,head: vertextype; var pattern: patternarray);
begin
   if (tail < head) then pattern[tail,head].arrow1 := true
   else pattern[head,tail].arrow2 := true;
end;

procedure ipropagate(tail,head: vertextype; var pattern: patternarray;
                    var adj: adjtype);
var i: integer;
begin
    isetdirection(tail,head,pattern);
    for i := 1 to numob do
      if (i <> tail) and (i <> head) then
         if (adj^[i,head] <> notthere) or (adj^[head,i] <> notthere) then
            if (adj^[tail,i] = notthere) and (adj^[i,tail] = notthere) then
              if not idirectionset(i,head,pattern) then
                  if icollider(i,tail,head,adj) then 
                     ipropagate(i,head,pattern,adj)
                  else ipropagate(head,i,pattern,adj);
end;     

procedure seticollider(var pattern: patternarray; var adj: adjtype);
var i,j,k: integer;
begin
   for i := 1 to max do
     for j := 1 to max do
       for k := 1 to max do
         if (i <> j) and (j <> k) and (i <> k) then
            if (pattern[i,j].existence = notthere) and (pattern[j,i].existence = notthere)
                then
               if icollider(i,j,k,adj) then
                   begin
                      isetdirection(i,k,pattern);
                      isetdirection(j,k,pattern);
		      ipropagate(i,k,pattern,adj);
                      ipropagate(j,k,pattern,adj);
                   end;
end;

procedure writearrow(var output : text;
		     i,j    :integer);
begin
   writename(i,output);
   twrite(output,' -> |');
   writename(j,output);
   twriteeoln(output);
end; { writearrow }

procedure writeedge(var output : text;
		    i,j	   :integer);
begin
   writename(i,output);
   twrite(output,' -- |');
   writename(j,output);
   twriteeoln(output);
end; { writeedge }


procedure writepattern2;
var
   i,j : integer;
begin
   for i := 1 to (max-1) do
      for j:= (i+1) to max do
	 if (pattern2[i,j].existence<>notthere) then
	 begin
	    if idirectionset(i,j,pattern2) then writearrow(output,i,j)
	    else if idirectionset(j,i,pattern2) then writearrow(output,j,i)
	    else writeedge(output,i,j);
	 end;
	 
end;

procedure writepattern1;
var
   i,j : integer;
begin
   for i := 1 to (max-1) do
      for j:= (i+1) to max do
	 if (pattern1[i,j].existence<>notthere) then
	 begin
	    if idirectionset(i,j,pattern1) then writearrow(output,i,j)
	    else if idirectionset(j,i,pattern1) then writearrow(output,j,i)
	    else writeedge(output,i,j);
	 end;
	 
end;


procedure converttopattern(var adj : adjtype; var pattern: patternarray);
var i,j,k														      : integer;
begin
   for i := 1 to max do
     for j := 1 to max do
        if i <> j then
           if adj^[i,j] <> notthere then
               if (i < j) then pattern[i,j].existence := 0
               else pattern[j,i].existence := 0;
  seticollider(pattern,adj);
(*   for i := 1 to max do
     for j := 1 to max do
        if (i <> j) and (pattern[i,j].existence <> notthere) then
           begin
             if pattern[i,j].arrow1 and pattern[i,j].arrow2 then
                writeln(vernames[i],' <-> ',vernames[j])
             else if pattern[i,j].arrow1 then
                writeln(vernames[i],' -> ',vernames[j])
             else if pattern[i,j].arrow2 then
                writeln(vernames[j],' -> ',vernames[i])              
             else writeln(vernames[i],' - ',vernames[j])
          end; *)
end;

procedure differences(var pattern1,pattern2: patternarray; var ofile: text);
var i,j,exomission, excomission, arrowomission, arrowcomission,
    numedges,nummissing,numarrow,noarrow: integer; found15,found45: boolean;
    missed,found: array[1..10] of real; anindex,numfound: integer; 
    avemissed,avefound,temp: real;
begin
 exomission := 0; excomission := 0; arrowomission := 0; arrowcomission := 0;
 numedges := 0; nummissing := 0; numarrow := 0; noarrow := 0; 
 avemissed := 0.0; avefound := 0.0;  numfound := 0;
 for i := 1 to 10 do 
    begin
      missed[i] := 0.0;
      found[i] := 0.0;
    end;
 for i := 1 to max-1 do 
   for j := i+1 to max do
      begin
        if (pattern1[i,j].existence <> notthere) then
              numedges := numedges + 1
        else nummissing := nummissing + 1;
        if (pattern2[i,j].existence <> notthere) and
           (pattern1[i,j].existence <> notthere) then
               begin
                  numfound := numfound + 1;
                  if trueparm^[i,j] <> 0.0 then temp := trueparm^[i,j] 
                  else temp := trueparm^[j,i];
                  anindex := round(temp * 10 + 0.5);
                  if anindex >= 10.5 then anindex := 10; 
                  writeln(i,' ',j,' ',trueparm^[i,j],' ',anindex);
                  found[anindex] := found[anindex] + 1;
                  avefound := avefound + trueparm^[i,j];
                  if pattern1[i,j].arrow1 then numarrow := numarrow + 1
                  else noarrow := noarrow + 1;
                  if pattern1[i,j].arrow2 then numarrow := numarrow + 1
                  else noarrow := noarrow + 1;       
               end;
        if (pattern1[i,j].existence <> notthere) and 
           (pattern2[i,j].existence = notthere) then 
                begin
                   exomission := exomission + 1;
                   writeln(vernames[i],' ',vernames[j],' missing');
                   anindex := round((trueparm^[i,j] * 10) + 0.5);
                   if anindex >= 10.5 then anindex := 10; 
                   missed[anindex] := missed[anindex] + 1;
                   avemissed := avemissed + trueparm^[i,j];
                end
        else if (pattern2[i,j].existence <> notthere) and 
           (pattern1[i,j].existence = notthere) then 
                begin
                    excomission := excomission + 1;
                    writeln(vernames[i],' ',vernames[j],' added');
                 end;
        if (pattern1[i,j].existence <> notthere) and 
           (pattern2[i,j].existence <> notthere) then
             begin
                if pattern1[i,j].arrow1 and (not pattern2[i,j].arrow1) then
                    begin
                        arrowomission := arrowomission + 1;
                        writeln(vernames[i],' ',vernames[j],' arrow missing');    
                    end
                else if pattern2[i,j].arrow1 and (not pattern1[i,j].arrow1) then
                    begin
                       arrowcomission := arrowcomission + 1;
                       writeln(vernames[i],' ',vernames[j],' arrow added');    
                    end;
                if pattern1[i,j].arrow2 and (not pattern2[i,j].arrow2) then
                    begin
                        arrowomission := arrowomission + 1;
                        writeln(vernames[i],' ',vernames[j],' arrow missing');    
                    end
                else if pattern2[i,j].arrow2 and (not pattern1[i,j].arrow2) then
                    begin
                       arrowcomission := arrowcomission + 1;
                        writeln(vernames[i],' ',vernames[j],' arrow added');    
                    end;
             end;
      end;
   tpwritereal(ofile,(exomission),6,3,false);twrite(ofile,' |');
   tpwritereal(ofile,(numedges),6,3,false);twrite(ofile,' |');
   tpwritereal(ofile,(excomission),6,3,false);twrite(ofile,' |');
   tpwritereal(ofile,(nummissing),6,3,false);twrite(ofile,' |');
   tpwritereal(ofile,(arrowomission),6,3,false);twrite(ofile,' |');
   tpwritereal(ofile,(numarrow),6,3,false);twrite(ofile,' |');
   tpwritereal(ofile,(arrowcomission),6,3,false);twrite(ofile,' |');
   tpwritereal(ofile,(noarrow),6,3,false);
   tpwritereal(ofile,avemissed/exomission,6,3,false);
   tpwritereal(ofile,avefound/numfound,6,3,false);   
   twriteeoln(ofile);
   twriteln(ofile,'missed |');
   for i := 1 to 10 do
     if missed[i] <> 0.0 then 
          begin
            twriteint(ofile,i);
            twrite(ofile,' |'); 
            tpwritereal(ofile,(missed[i]),6,3,false);twrite(ofile,' |');
            twriteeoln(ofile);
          end;
   twriteeoln(ofile);
   twriteln(ofile,'found |');
   for i := 1 to 10 do
       if found[i] <> 0.0 then 
          begin
            twriteint(ofile,i);
            twrite(ofile,' |'); 
            tpwritereal(ofile,(found[i]),6,3,false);twrite(ofile,' |');
            twriteeoln(ofile);
          end;
   twriteeoln(ofile);
end;


procedure convertpattomod(var pattern1: patternarray;
			  var amod: modeltypeptr);
var i,j: vertextype;
begin
  for i := 1 to max do amod^.amodel[i] := [];
  for i := 1 to max do
    for j := 1 to max do
      begin
        if pattern1[i,j].existence = 0 then
           if (not (pattern1[i,j].arrow1)) and (not (pattern1[i,j].arrow2)) then 
              begin
                 amod^.amodel[i] := amod^.amodel[i] + [j];
                 amod^.amodel[j] := amod^.amodel[j] + [i];
              end
           else if pattern1[i,j].arrow1 then amod^.amodel[j] := amod^.amodel[j] + [i]
           else if pattern1[i,j].arrow2 then amod^.amodel[i] := amod^.amodel[i] + [j];
      end;
end;

procedure myconvertpattomod(var pattern1: patternarray;
			  var amod: modeltypeptr);
var i,j: vertextype;
begin
  for i := 1 to max do amod^.amodel[i] := [];
  for i := 1 to max do
    for j := 1 to max do
      begin
        if pattern1[i,j].existence = 0 then
           if (not (pattern1[i,j].arrow1)) and (not (pattern1[i,j].arrow2)) then 
              begin
                 amod^.amodel[i] := amod^.amodel[i] + [j];
                 amod^.amodel[j] := amod^.amodel[j] + [i];
              end
           else if pattern1[i,j].arrow2 then amod^.amodel[j] := amod^.amodel[j] + [i]
           else if pattern1[i,j].arrow1 then amod^.amodel[i] := amod^.amodel[i] + [j];
      end;
end;

procedure comparepatterns1(var ofile: text);
var ifile1,ifile2,ifile3		      : text;
   ifilename1,ifilename2,ifilename3,ofilename : inbuf; 
   changed				      : boolean;
   i,j					      : integer;
   crossent1,crossent2,crossent3,score1,score2,divergence: real;
   boo,ok,mmyfail			      : boolean;
   cp					      : modeltypeptr;
   
begin
   new(cp);
   for i := 1  to maxvertex do
    for j := 1  to maxvertex do
          begin
             pattern2[i,j].existence := notthere;
             pattern2[i,j].arrow1 := false;
             pattern2[i,j].arrow2 := false;
          end;
   readpattern(false,true,patfile,pattern2,vernames);
   max := oldmax;
   myconvertpattomod(pattern2,cp);
   readmodtoadj(adjacency,cp); 
   finddag(adjacency);
   linestimate(adjacency,rho,parm,ok);  (* this estimates the parameters
                                           from the search graph and the sample *)
   minternal := true;                   (* this estimates the correlation matrix
                                           from the search graph and the sample *)
   datatypes := [5];
   domonte;
   minternal := false; 
   getdivergence(populcor,truecor,divergence,ok); (* compare population and estimate
                                               from search *)
   twritereal(ofile,divergence);    
   twrite(ofile,'  |');
   initlinear;
   mygetpatternscore(cp,score2,lambda,mmyfail,1);
   disposelinear;
   twritereal(ofile,score2);
   twrite(ofile,'  |');
   differences(pattern1,pattern2,ofile);
   flush(ofile);
   dispose(cp);
end;

procedure convertmodeltopattern(    cp	    : modeltypeptr;
				var pattern : patternarray);
var i,j,k: integer;
begin
   for i := 1 to max do
      for j := 1 to max do
	 if i <> j then
	 begin
	    if adjm(cp,i,j) then
	       pattern[i,j].existence:=0;
	    if existarrowm(cp,i,j) then
       	       pattern[i,j].arrow2:=true;
	    if existarrowm(cp,j,i) then
       	       pattern[i,j].arrow1:=true;
	 end;
end;

procedure comparepatterns;
var changed,mmyfail  : boolean; i,j,numtoget: integer; 
    ofile    : text; ok: boolean; divergence,score: real;
    cp       : modeltypeptr;
    colset   : vertexset;
begin
   getknowledge; (* this is to get the population correlation matrix *)
   getknowledge; (* this is to get the true graph *)
   linestimate(adjacency,rho,trueparm,ok);  (* this standardizes true parameters *)
   for i := 1 to max do 
      for j := 1 to max do
        populcor^[i,j] := rho^[i,j];
   write('Number to get? ');
   readln(numtoget);
   for i := 1  to maxvertex do
    for j := 1  to maxvertex do
          begin
             pattern1[i,j].existence := notthere;
             pattern1[i,j].arrow1 := false;
             pattern1[i,j].arrow2 := false;
             pattern2[i,j].existence := notthere;
             pattern2[i,j].arrow1 := false;
             pattern2[i,j].arrow2 := false;
          end;
   getfile(ofile,'Output File: |',outputf,ofilename,changed);
   getknowledge;   (* this is to get the sample correlation matrix *)
   getdivergence(populcor,rho,divergence,ok); (* compare population and sample *)
   twritereal(ofile,divergence);
   getknowledge;                        (* this is to get the true graph *);
   new(cp);
   for i := 1 to max do colset:=colset + [i];
   zreadadjtomod(adjacency,cp);   
   findpattern(cp,colset);              (*  this converts true model to pattern *)
   initlinear;
   getthescore(lambda, score);
   disposelinear;
   twrite(ofile,'  |');
   twritereal(ofile,score);
   convertmodeltopattern(cp,pattern1);  (* this converts form pattern stored in *)
   linestimate(adjacency,rho,parm,ok);  (* this estimates the parameters
                                               from the true graph and the sample *)
   minternal := true;                   (* this estimates the correlation matrix
                                           from the true graph and the sample *)
   datatypes := [5];
   domonte;
   minternal := false; 
   getdivergence(populcor,truecor,divergence,ok); (* compare population and estimate
                                               from true *)
   twritereal(ofile,divergence);    
   oldmax := max;
   getfile(patfile,'Input File: |',graphf,patfilename,changed); 
                                        (* get the patterns from the search *)
   max := oldmax;
   for i := 1 to numtoget do comparepatterns1(ofile);
   dispose(cp);
end;


procedure convertmodtopat(* var amod: modeltypeptr *);
var i,j: vertextype;
begin
   for i := 1  to maxvertex do
    for j := 1  to maxvertex do
          begin
             pattern1[i,j].existence := notthere;
             pattern1[i,j].arrow1 := false;
             pattern1[i,j].arrow2 := false;
             pattern2[i,j].existence := notthere;
             pattern2[i,j].arrow1 := false;
             pattern2[i,j].arrow2 := false;
          end;
   readmodtoadj(adjacency,amod); 
   converttopattern(adjacency,pattern1);
   convertpattomod(pattern1,amod);
end;


procedure calcancestors(var adj,an: adjtype);
label 999;
var j,k,i: vertextype;
begin 
    for i := 1 to max do 
      for j := 1 to max do
       if (i in graphvertices) and (j in graphvertices) then     
        if i <> j then
           an^[i,j] := adj^[i,j]
        else an^[i,j] := 0;
    for k := 1 to max do 
       for i := 1 to max do 
           for j := 1 to max do        
             if (i in graphvertices) and (j in graphvertices) and
                (k in graphvertices) then
               if (an^[i,k] <> notthere) and
                  (an^[k,j] <> notthere) then
                      an^[i,j] := 0;
999: end; 

procedure moralize(var adj,moral: adjtype; var ancestors: vertexset);
var i,j,k: vertextype;
begin
   for i := 1 to max do 
      for j := 1 to max do
         begin
            moral^[i,j] := notthere;
            moral^[j,i] := notthere;
         end;
   for i := 1 to max do 
      for j := 1 to max do
        if (i in ancestors) and (j in ancestors) then
         begin
           if adj^[i,j] <> notthere then
              begin
                 moral^[i,j] := 0;
                 moral^[j,i] := 0;
              end
           else 
              begin
                 for k := 1 to max do
                   if (adj^[i,k] <> notthere) and (adj^[j,k] <> notthere) 
                      and (k in ancestors) then
                      begin
                         moral^[i,j] := 0;
                         moral^[j,i] := 0;
                      end;
              end
          end
end;

procedure removeedg(var moral: adjtype; condition: vertexset);
var i,j: vertextype;
begin
   for i := 1 to max do 
       if i in condition then
         for j := 1 to max do 
           begin
             moral^[i,j] := notthere;
             moral^[j,i] := notthere;
           end;
end;
     
function dseparated(var adj,an,an1,moral: adjtype; v1,v2: vertextype; 
                     condition: vertexset): boolean;
var i,j: vertextype; ancestors: vertexset;
begin
  ancestors := [];
  for i := 1 to max do 
     begin
        if (an^[i,v1] <> notthere) or (an^[i,v2] <> notthere) then
           ancestors := ancestors + [i];
        for j := 1 to max do 
            if (j in condition) and (an^[i,j] <> notthere) then
               ancestors := ancestors + [i];
     end;
  moralize(adj,moral,ancestors);
  removeedg(moral,condition);
  calcancestors(moral,an1);
  dseparated := (an1^[v1,v2] = notthere) and (an1^[v2,v1] = notthere);
end;

procedure readvarlist(* var selection: vertexset *);
var leaveloop,wasinserted,firsttime: boolean; name: vertexnametype; i,k: vertextype; 
   begin
     selection := [];
     leaveloop := false;
     firsttime := true;
     i := 0;
     write(output,'Variables: ');
     readttyline(buf);
     while (not leaveloop) do
       begin
         readwhitespace;
         if (not myeoln) then
           begin
             firsttime := false;
             readwhitespace;
             if (not myeoln) then
	       begin
                 readname(name,fatal);
                 if name[1] in ['A'..'Z'] then
   	           myerror('No theoreticals allowed in selection.|  ',
                           fatal,true)
                 else i := i + 1;
                 k := listposition(name,true,vernames,wasinserted);
                 selection := selection + [k];
	       end;
           end
          else leaveloop := true;
        end;
end;
procedure gettheancestors(thevars: vertexset; var ancestors: vertexset; var an: adjtype);
var i,j: vertextype;
begin
   ancestors := [];
   for i := 1 to numob do
     for j := 1 to numob do
        if (j in thevars) and (j <> i) and (an^[i,j] <> notthere) then
          ancestors := ancestors + [i];
end;

procedure dagtopag(var adj: adjtype; var ugraph: endarray);
var an,an1,moral: adjtype; condition,selected: vertexset; i,j,k: vertextype;
begin
   getknowledge; (* get true graph *)
   tnewend(ugraph,fatal);
   for i := 1 to max do
     for j := 1 to max do
       begin
         ugraph^[i,j] := [];
         ugraph^[j,i] := [];
       end;
   tnewadj(an,fatal);
   tnewadj(an1,fatal);
   tnewadj(moral,fatal);
   tnewdsep(dsep,fatal);
   calcancestors(adj,an);
   readvarlist(selected);
   for i := 1 to numob-1 do
     for j := i+1 to numob do
      if (not (i in selected)) and (not (j in selected)) then
       begin
         gettheancestors(selected + [i,j],condition,an);
         condition := condition - [i,j];
         if dseparated(adj,an,an1,moral,i,j,condition) then 
             begin
                dsep^[i,j] := condition;
                dsep^[j,i] := condition;
                ugraph^[i,j] := [];
                ugraph^[j,i] := [];
             end
         else
             begin
                dsep^[i,j] := [0];
                dsep^[j,i] := [0];
                ugraph^[i,j] := [notknown];
                ugraph^[j,i] := [notknown];
             end;
       end
      else
       begin
          dsep^[i,j] := [];
          dsep^[j,i] := [];
          ugraph^[i,j] := [];
          ugraph^[j,i] := [];
       end;
   getpartials;
   dispose(an);
   dispose(an1);
   dispose(moral);
   dispose(dsep);
end;

procedure readpag(* infromtty: boolean; add: boolean; var fileread: text;
		  var augraph: endarray *);

label 999;
var emptyline,notfound,finished: boolean; c,c1: char; 
    head,tail: vertexnametype; i,j,headpos,tailpos: integer;
    arrow1,arrow2,wasinserted,circle1,circle2,outof1,outof2: boolean;

begin 
   if infromtty then lineno := 0 else lineno := 1;
   finished := false; twriteeoln(output);
   i := 1;
   buf[1] := 'c';
(*   while not eof(fileread) and (buf[1] <> '/') do 
         readfileline(fileread,buf); *)
   while not eof(fileread) and (buf[1] <> '!') do
      begin (*While*)
         emptyline := false;
         checkinterrupt;
         if (lastvarassigned = max2s) then
		begin
		twrite(output,'Maximum number of edges reached. No more edges allowed.|');
		twriteeoln(output);
		end;
         errors := errors - [ignoreall];
         head := '  '; tail := '  ';
	 if not infromtty then
 	    begin
         	readfileline(fileread,buf);
                if buf[1] = '!' then goto 999;
                if myeoln then emptyline := true;
	        if emptyline then goto 999 else
                 begin
                    readit(c);
                    if c = '{' then 
                      begin
                        readcomment(fileread);
                        while myeoln and (not eof(fileread)) do
                            readfileline(fileread,buf);
                      end
                    else curpos := curpos - 1;
                    if (fatal in errors) then goto 999;
                    arrow1 := false;
                    arrow2 := false;
                    outof1 := false;
                    outof2 := false;
                    circle1 := false;
                    circle2 := false;
                    readname(head,ignoreall);
                    readwhitespace;
                    readit(c);
                    if c = '<' then arrow1 := true
                    else if c = 'o' then circle1 := true
                    else if c = '-' then outof1 := true;
                    repeat 
                        readit(c)
                    until c <> '-';
                    if c = '>' then arrow2 := true
                    else if c = 'o' then circle2 := true
                    else if c = '-' then outof2 := true;
                    readwhitespace;
                    readname(tail,ignoreall);
	 	    lineno := lineno + 1;
               end;
	    end;
         if not emptyline and (head[1] <> ' ') and (head[1] <> chr(0)) and
            (tail[1] <> ' ') and (tail[1] <> chr(0)) then
           begin
             if (head = tail) and (head[1] <> ' ') then
	        myerror('Head not allowed to equal tail          ',ignoreall,
	     	         true);
                headpos := listposition(head,true,vernames,wasinserted);
                tailpos := listposition(tail,true,vernames,wasinserted);
                if not (ignoreall in errors) then
		    begin
			if augraph^[headpos,tailpos] <> [] then
                           begin
                 	        twrite(output,'Edge:  |');
				writevername(head,output);
				twrite(output,' --> |');
				writevername(tail,output);
				twriteeoln(output);
 			        myerror('This edge already in graph              ',
			   	    ignoreall,true)
                           end
		        else
			   begin
                             if arrow2 then augraph^[headpos,tailpos] := [in2]
                             else if outof2 then augraph^[headpos,tailpos] := [out2]
                             else if circle2 then augraph^[headpos,tailpos] := [notknown];
                             if arrow1 then augraph^[tailpos,headpos] := [in2]
                             else if outof1 then augraph^[tailpos,headpos] := [out2]
                             else if circle1 then augraph^[tailpos,headpos] := [notknown];
			   end;
			i := i + 1;
		    end;
                 end;
     end;
999:end;  


procedure writeedges(i,j: vertextype; var ugraph: endarray);
begin
   writename(i,output);
   write(' ');
   if ugraph^[j,i] <> [] then
      begin
         if in2 in ugraph^[j,i] then write('<')
         else if notknown in ugraph^[j,i] then write('o')
         else if out2 in ugraph^[j,i] then write('-');
         write('-');
         if in2 in ugraph^[i,j] then write('>')
         else if notknown in ugraph^[i,j] then write('o')
         else if out2 in ugraph^[i,j] then write('-');     
      end;
   write(' ');
   writename(j,output);
   write('     ');
end;

procedure comparepags(var ugraph: endarray);
var ugraph2: endarray; var pagfile: text; 
    changed: boolean; pagfilename: inbuf; i,j: vertextype;
    numedges,adjadded,adjmiss,arrowadded,outadded,notadded,numincommon: integer;
begin
   numedges := 0;
   numincommon := 0;
   adjadded := 0;
   adjmiss := 0;
   arrowadded := 0;
   outadded := 0;
   notadded := 0;
   tnewend(ugraph,fatal);
   tnewend(ugraph2,fatal);
   for i := 1 to max do
     for j := 1 to max do
       begin
         ugraph^[i,j] := [];
         ugraph^[j,i] := [];
         ugraph2^[i,j] := [];
         ugraph2^[j,i] := [];
       end;
   getfile(pagfile,'True Graph File: |',graphf,pagfilename,changed); 
   readpag(false,true,pagfile,ugraph);
   getfile(pagfile,'Output of search: |',graphf,pagfilename,changed); 
   readpag(false,true,pagfile,ugraph2);
   for i := 1 to numob do
     for j := 1 to numob do
       begin
         if i < j then
           begin
             if (ugraph^[i,j] <> []) and (ugraph2^[i,j] <> [])
                 then numincommon := numincommon + 1;
             if (ugraph^[i,j] <> []) then
                 numedges := numedges + 1;
             if (ugraph^[i,j] = []) and (ugraph2^[i,j] <> []) then
                 begin
                    adjadded := adjadded + 1;
                    writeedges(i,j,ugraph);
                    writeedges(i,j,ugraph2);
                    writeln;
                 end
             else if (ugraph^[i,j] <> []) and (ugraph2^[i,j] = []) then 
                 begin
                    adjmiss := adjmiss + 1;
                    writeedges(i,j,ugraph);
                    writeedges(i,j,ugraph2);
                    writeln;
                 end;
           end;
         if i <> j then
           begin
             if (ugraph2^[i,j] = [in2]) and (ugraph^[i,j] <> [in2]) and
                (ugraph^[i,j] <> []) then 
                begin
                   arrowadded := arrowadded + 1;
                   writeedges(i,j,ugraph);
                   writeedges(i,j,ugraph2);
                   writeln;
                 end
             else if (ugraph2^[i,j] = [out2]) and (ugraph^[i,j] <> [out2]) and
                (ugraph^[i,j] <> []) then 
                begin
                    writeedges(i,j,ugraph);
                    writeedges(i,j,ugraph2);
                    writeln;
                    outadded := outadded + 1
                 end
             else if (ugraph2^[i,j] = [notknown]) and (ugraph^[i,j] <> [notknown]) and
                (ugraph^[i,j] <> []) then 
                   begin
                    writeedges(i,j,ugraph);
                    writeedges(i,j,ugraph2);
                    writeln;
                    notadded := notadded + 1;
                 end;
           end;
       end;
    writeln(ofilvar,'Edges added ', adjadded/(((numob * (numob - 1))/2) - numedges):6:3,
                    ' Edges deleted ', (adjmiss/numedges):6:3,
                    ' Arrows added ', (arrowadded / (2 * numincommon)):6:3,
                    ' Out of added ', (outadded / (2 * numincommon)):6:3,
                    ' Circle added ', (notadded / (2 * numincommon)):6:3);
    dispose(ugraph);
    dispose(ugraph2);
end;
             
procedure stripelements;
var ofilename: inbuf; ofile: text; colno,colval,sampno,i,j: integer;
begin
   sampno := 0;
   getknowledge; (* get the data file *)
   write('Output file: ');
   readln(ofilename);
   rewrite(ofile,ofilename);
   colno := round(getnumber(0.0,1000.0,true,'Column to remove|',1.0));
   colval := round(getnumber(0.0,1000.0,true,'Value to remove|',1.0));
   for i := 1 to round(sampsize) do
     begin
      if datamat^[i]^[colno] = colval then
         begin
           sampno := sampno + 1;
           for j := 1 to max do
             if j <> colno then write(ofile,varcatvalue[j,datamat^[i]^[j]]:2:0);
           if j = max then writeln(ofile);
         end;
       end;
end;

(*************************************************************************)
procedure readpag1(infromtty: boolean; add,firsttime: boolean; var fileread: text;
		  var newan: adjtype);

label 999;
var emptyline,notfound,finished: boolean; c,c1: char; 
    head,tail: vertexnametype; i,j,headpos,tailpos: integer;
    arrow1,arrow2,wasinserted,circle1,circle2,outof1,outof2: boolean;

begin 
   if infromtty then lineno := 0 else lineno := 1;
   finished := false; twriteeoln(output);
   i := 1;
   buf[1] := 'c';
   while not eof(fileread) and (buf[1] <> '/') do 
         readfileline(fileread,buf);
   while not eof(fileread) and (buf[1] <> '!') do
      begin (*While*)
         emptyline := false;
         checkinterrupt;
         if (lastvarassigned = max2s) then
		begin
		twrite(output,'Maximum number of edges reached. No more edges allowed.|');
		twriteeoln(output);
		end;
         errors := errors - [ignoreall];
         head := '  '; tail := '  ';
	 if not infromtty then
 	    begin
         	readfileline(fileread,buf);
                if buf[1] = '!' then goto 999;
                if myeoln then emptyline := true;
	        if emptyline then goto 999 else
                 begin
                    readit(c);
                    if c = '{' then 
                      begin
                        readcomment(fileread);
                        while myeoln and (not eof(fileread)) do
                            readfileline(fileread,buf);
                      end
                    else curpos := curpos - 1;
                    if (fatal in errors) then goto 999;
                    readname(head,ignoreall);
                    readwhitespace;
                    readit(c);
                    readit(c);
                    readwhitespace;
                    readname(tail,ignoreall);
	 	    lineno := lineno + 1;
               end;
	    end;
         if not emptyline and (head[1] <> ' ') and (head[1] <> chr(0)) and
            (tail[1] <> ' ') and (tail[1] <> chr(0)) then
           begin
                headpos := listposition(head,true,vernames,wasinserted);
                tailpos := listposition(tail,true,vernames,wasinserted);
                if not (ignoreall in errors) then
                  if firsttime then newan^[headpos,tailpos] := 1
                  else newan^[headpos,tailpos] := 3;
           end;
     end;
999:end;  

procedure comparepags1;
var an,an1: adjtype; var pagfile: text; 
    changed: boolean; pagfilename: inbuf; i,j: vertextype;
    numan,numnotan,numanc,numnotanc,truean,truenotan: integer;
begin
   numan := 0;
   numanc := 0;
   numnotan := 0;
   numnotanc := 0;
   truean := 0;
   truenotan := 0;
   tnewadj(an,fatal);
   tnewadj(an1,fatal);
   for i := 1 to max do
     for j := 1 to max do
       begin
         an^[i,j] := notthere;
         an1^[i,j] := notthere;
       end;
   getknowledge;    (* true graph file *)
   calcancestors(adjacency,an);
   getfile(pagfile,'Output from search: |',graphf,pagfilename,changed); 
   readpag1(false,true,true,pagfile,an1);
   readpag1(false,true,false,pagfile,an1);
   for i := 1 to numob do
     for j := 1 to numob do
       if (i <> j) then
         begin
            if an^[i,j] <> notthere then truean := truean + 1
            else truenotan := truenotan + 1;
            if an1^[i,j] = 1 then 
               begin
                 numan := numan + 1;
                 if an^[i,j] <> notthere then numanc := numanc + 1
                 else
                   begin
                      write(output,' ancestor wrong ');
                      writename(i,output);
                      write(output, ' ');
                      writename(j,output);
                      writeln(output, ' ');
                   end;
               end
            else if an1^[i,j] = 3 then
               begin
                 numnotan := numnotan + 1;
                 if an^[i,j] = notthere then numnotanc := numnotanc + 1
                 else
                    begin
                       write(output,' nonancestor wrong ');
                       writename(i,output);
                       write(output, ' ');
                       writename(j,output);
                       writeln(output, ' ');
                    end;
               end;
         end;
    if numan = 0 then 
        writeln(ofilvar,' Ancestors predicted ',numan,' Ancestors right ',numanc, '% ancestors right 100')
    else writeln(ofilvar,' Ancestors predicted ',numan,' Ancestors right ',numanc, '% ancestors right ',
            100 * numanc/numan:6:2);
    if numnotan = 0 then
        writeln(ofilvar,' Non-ancestors predicted ',numnotan,' Non-ancestors right ',numnotanc,
                ' % Non-ancestors 100')
    else writeln(ofilvar,' Non-ancestors predicted ',numnotan,' Non-ancestors right ',numnotanc,
            ' % Non-ancestors ', 100 * numnotanc/numnotan:6:2);
    writeln(ofilvar,' Number of ancestors ',truean,' Number of nonancestors ',truenotan);
    writeln(ofilvar,' True ancestor percent ',100 * truean/((numob * (numob - 1))):6:2,
                    ' True nonancestor percent ',100 * truenotan/((numob * (numob - 1))):6:2);
    writeln(ofilvar,' Percent ancestor predicted ',100 * numanc/truean:6:2,
                    ' True nonancestor percent ',100 * numnotanc/truenotan:6:2);
    dispose(an);
    dispose(an1);
end;







